#version 300 es

uniform lowp sampler2D diffuseMap;
uniform lowp sampler2D normalMap;
uniform lowp sampler2D specularMap;
uniform lowp samplerCube envMap;
uniform lowp sampler2D _ShadowMap;
uniform lowp vec3 materialSpecular;
uniform highp vec3 _LightPosition;
uniform lowp vec3 _LightColor;
uniform mediump float _LightRadius;
uniform lowp float roughness;
uniform lowp float reflectivity;
uniform highp int _LightType;
uniform highp float _LightConeAngleCos;
uniform highp vec3 _LightDirection;

in highp vec2 vTexCoord;
in highp vec3 vViewVec;
in highp vec3 vLightVec;
#ifdef ENVIRONMENT_MAP
in highp vec3 vEnvTexCoord;
#endif

out lowp vec4 fragColor;

// Shadow stuff
in highp vec4 vPosition;
in highp vec4 vProjCoord;
uniform highp vec3 _LightPositionWorld;
uniform highp int _LightCastsShadow;

highp float linstep( highp float low, highp float high, highp float v )
{
    return clamp( (v - low) / (high - low), 0.0, 1.0 );
}

highp float VSM( in highp float depth )
{
    highp vec2 moments = texture( _ShadowMap, vProjCoord.xy / vProjCoord.w ).rg;
    
    highp float variance = moments.y - moments.x * moments.x;
    //variance = max( variance, 0.0005 );
    
    highp float delta = depth - moments.x;
    highp float pMax = linstep( 0.2, 1.0, variance / (variance + delta * delta) );
    
    return (depth <= moments.x) ? 1.0 : pMax;
}

void main()
{
    mediump vec3 albedoTex = texture( diffuseMap, vTexCoord ).rgb;
#ifdef SPECULAR_MAP
    mediump vec3 specularTex = texture( specularMap, vTexCoord ).rgb;
#else
    mediump vec3 specularTex = materialSpecular;
#endif
#ifdef ENVIRONMENT_MAP
    mediump vec3 envTex = texture( envMap, vEnvTexCoord ).rgb;
    albedoTex = mix( albedoTex, envTex, reflectivity ) * specularTex;
#endif

    if (_LightType == -1)
    {
        lowp vec3 ambient = vec3( 0.2, 0.2, 0.2 );
        fragColor = vec4( ambient * albedoTex, 1.0f );
        return;
    }
   
    // input vectors
    mediump vec3 N = normalize( texture( normalMap, vTexCoord ).rgb * 2.0 - 1.0 );
    mediump vec3 V = normalize( vViewVec );
    mediump vec3 L = normalize( vLightVec );
    mediump vec3 H = normalize( L + V );
    
    // dot products
    mediump float dotNL = clamp( dot( N, L ), 0.0, 1.0 );
    mediump float dotNH = clamp( dot( N, H ), 0.0, 1.0 );
    mediump float dotNV = clamp( dot( N, V ), 0.0, 1.0 );
    mediump float dotVH = clamp( dot( V, H ), 0.0, 1.0 );
    mediump float r_sq = roughness * roughness;
    
    // Geometric term.
    mediump float geoNumerator = 2.0 * dotNH;
    mediump float geoB = (geoNumerator * dotNV) / dotVH;
    mediump float geoC = (geoNumerator * dotNL) / dotVH;
    mediump float geo = min( 1.0, min( geoB, geoC ) );
    
    // Beckmann roughness.
    mediump float roughness_a = 1.0 / (4.0 * r_sq * pow( dotNH, 4.0 ) );
    mediump float roughness_b = dotNH * dotNH - 1.0;
    mediump float roughness_c = r_sq * dotNH * dotNH;
    mediump float roughness = roughness_a * exp( roughness_b / roughness_c );

    // Schlick Fresnel.
    mediump float ref_at_norm_incidence = 1.0;
    mediump float fresnel = pow( 1.0 - dotVH, 5.0 );
    fresnel *= (1.0 - ref_at_norm_incidence);
    fresnel += ref_at_norm_incidence;
    
    // Specular BRDF.
    mediump float Rs_numerator = fresnel * geo * roughness;
    mediump float Rs_denominator  = dotNV * dotNL;
    // Added max to fix negative lighting, bug somewhere earlier.
    mediump float Rs = max( 0.0, Rs_numerator / Rs_denominator );

    lowp float falloff = 1.0;
    
    if (_LightType == 1)
    {
        mediump vec3 vToLight = _LightPosition - vPosition.xyz;
        mediump float lightDistance = length( vToLight );

        if (lightDistance < _LightRadius)
        {
            mediump float x = lightDistance / _LightRadius;

            // fake inverse squared falloff:
            // -(1/k)*(1-(k+1)/(1+k*x^2))
            // k=20: -(1/20)*(1 - 21/(1+20*x^2))
            falloff = -0.05 + 1.05 / (1.0 + 20.0 * x * x);
        }
        else
        {
            fragColor = vec4( 0.0, 0.0, 0.0, 1.0 );
            return;
        }
    }
    else if (_LightType == 2)
    {
        highp vec3 vecToLight = normalize( _LightPosition - vPosition.xyz );
        highp float spotAngle = dot( -_LightDirection.xyz, vecToLight );

        // Falloff
        highp float dist = distance( vPosition.xyz, _LightPosition );
        highp float a = dist / _LightRadius + 1.0;
        highp float att = 1.0f / (a * a);
        highp vec3 color = _LightColor * att;

        fragColor = spotAngle > _LightConeAngleCos ? vec4( color, 1.0 ) : vec4( 0.0, 0.0, 0.0, 1.0 );
        
        if (_LightCastsShadow == 1)
        {
            highp float depth = vProjCoord.z / vProjCoord.w;
            highp float shadow = VSM( depth );
            fragColor *= shadow;
            fragColor.w = 1.0;
        }

        return;
    }

    mediump vec3 final = clamp( max( 0.0, dotNL ) * (specularTex * Rs + albedoTex), 0.0, 1.0 ) * falloff * _LightColor;
    // Should not be needed, but without this there is no diffuse lighting when running on PC (Windows tested). Works on iOS commented out.
    final += _LightColor * albedoTex * falloff * 2.0;

    if (_LightCastsShadow == 1)
    {
        highp float depth = vProjCoord.z / vProjCoord.w;
        highp float shadow = VSM( depth );
        final *= shadow;
    }

    fragColor = vec4( final, 1.0 );
}

