#version 150 core

uniform sampler2D diffuseMap;
uniform sampler2D normalMap;
uniform sampler2D specularMap;
uniform samplerCube envMap;
uniform sampler2D _ShadowMap;

uniform vec3 materialSpecular;
uniform float roughness;
uniform float reflectivity;
uniform vec3 _LightPosition;
uniform vec3 _LightColor;
uniform float _LightRadius;
uniform int _LightCastsShadow;
uniform float _LightConeAngleCos;
uniform vec3 _LightDirection;
uniform int _LightType;

in vec2 vTexCoord;
in vec3 vViewVec;
in vec3 vLightVec;
#ifdef ENVIRONMENT_MAP
in vec3 vEnvTexCoord;
#endif

// Shadow stuff
in vec4 vPosition;
in vec4 vProjCoord;
uniform vec3 _LightPositionWorld;

out vec4 fragColor;

float linstep(float low, float high, float v)
{
    return clamp( (v - low) / (high - low), 0.0, 1.0 );
}

float VSM( in float depth )
{
    vec2 moments = textureProj( _ShadowMap, vProjCoord ).rg;

    float variance = max( moments.y - moments.x * moments.x, -0.001 );
    
    float delta = depth - moments.x;
    float p = smoothstep( depth - 0.02, depth, moments.x );
    float pMax = linstep( 0.2, 1.0, variance / (variance + delta * delta) );
    
    return clamp( max( p, pMax ), 0.0, 1.0 );
}

// Based on http://content.gpwiki.org/D3DBook:%28Lighting%29_Cook-Torrance
void main()
{
    vec3 albedoTex = texture( diffuseMap, vTexCoord ).rgb;
#ifdef SPECULAR_MAP
    vec3 specularTex = texture( specularMap, vTexCoord ).rgb;
#else
    vec3 specularTex = materialSpecular;
#endif
#ifdef ENVIRONMENT_MAP
    vec3 envTex = texture( envMap, vEnvTexCoord ).rgb;
    albedoTex = mix( albedoTex, envTex, reflectivity ) * specularTex; // Could use a reflection mask texture or channel.
#endif
    
    if (_LightType == -1)
    {
        vec3 ambient = vec3( 0.1, 0.1, 0.1 );
        fragColor = vec4( ambient * albedoTex, 1.0f );
        return;
    }
    
    // input vectors
    vec3 N = normalize( texture( normalMap, vTexCoord ).rgb * 2.0 - 1.0 );
    vec3 V = normalize( vViewVec );
    vec3 L = normalize( vLightVec );
    vec3 H = normalize( L + V );
    
    // dot products
    float dotNL = clamp( dot( N, L ), 0.0, 1.0 );
    float dotNH = clamp( dot( N, H ), 0.0, 1.0 );
    float dotNV = clamp( dot( N, V ), 0.0, 1.0 );
    float dotVH = clamp( dot( V, H ), 0.0, 1.0 );
    float r_sq = roughness * roughness;
    
    // Geometric term.
    float geoNumerator = 2.0 * dotNH;
    float geoB = (geoNumerator * dotNV) / dotVH;
    float geoC = (geoNumerator * dotNL) / dotVH;
    float geo = min( 1.0, min( geoB, geoC ) );
    
    // Beckmann roughness.
    float roughness_a = 1.0 / (4.0 * r_sq * pow( dotNH, 4 ) );
    float roughness_b = dotNH * dotNH - 1.0;
    float roughness_c = r_sq * dotNH * dotNH;
    float roughness = roughness_a * exp( roughness_b / roughness_c );

    // Schlick Fresnel.
    float ref_at_norm_incidence = 1.0;
    float fresnel = pow( 1.0 - dotVH, 5.0 );
    fresnel *= (1.0 - ref_at_norm_incidence);
    fresnel += ref_at_norm_incidence;
    
    // Specular BRDF.
    float Rs_numerator = fresnel * geo * roughness;
    float Rs_denominator  = dotNV * dotNL;
    float Rs = max( 0.0, Rs_numerator / Rs_denominator ); // Added max() to prevent black areas.
    
    float falloff = 1.0f;

    // Point light.
    if (_LightType == 1)
    {
        vec3 vecToLight = _LightPosition - vPosition.xyz;
        float lightDistance = length( vecToLight );

        if (lightDistance < _LightRadius)
        {
            float x = lightDistance / _LightRadius;

            // fake inverse squared falloff:
            // -(1/k)*(1-(k+1)/(1+k*x^2))
            // k=20: -(1/20)*(1 - 21/(1+20*x^2))
            falloff = -0.05 + 1.05 / (1.0 + 20.0 * x * x);
        }
        else
        {
            fragColor = vec4( 0.0, 0.0, 0.0, 1.0 );
            return;
        }
    }
    else if (_LightType == 2)
    {
        // http://www.mbsoftworks.sk/index.php?page=tutorials&series=1&tutorial=20
        vec3 vecToLight = normalize( _LightPosition - vPosition.xyz );
        float spotAngle = dot( -_LightDirection.xyz, vecToLight );

        // Falloff
        float dist = distance( vPosition.xyz, _LightPosition );
        float a = dist / _LightRadius + 1.0;
        float att = 1.0f / (a * a);
        vec3 color = _LightColor * att;

        // Edge smoothing begins.
        float dif = 1.0f - _LightConeAngleCos;
        float factor = clamp( (spotAngle - _LightConeAngleCos) / dif, 0.0f, 1.0f );
        float linearFalloff = 0.04f;
        vec3 edgeColor = _LightColor * factor / (dist * linearFalloff);
        color = min( edgeColor, color ) * specularTex;
        // Edge smoothing ends.
        
        fragColor = spotAngle > _LightConeAngleCos ? vec4( color, 1.0 ) : vec4( 0.0, 0.0, 0.0, 1.0 );

        if (_LightCastsShadow == 1)
        {
            float depth = vProjCoord.z / vProjCoord.w;
            float shadow = VSM( depth );
            fragColor *= shadow;
            fragColor.w = 1.0;
        }

        return;
    }
    
    vec3 final = max( 0.0, dotNL ) * (specularTex * Rs + albedoTex) * falloff * _LightColor;
    // Should not be needed, but without this there is no diffuse lighting with point lights?!
    if (_LightType == 1)
    {
        final += _LightColor * albedoTex * falloff * 2.0;
    }
    
    if (_LightCastsShadow == 1)
    {
        float depth = vProjCoord.z / vProjCoord.w;
        float shadow = VSM( depth );
        final *= shadow;
    }
    
    fragColor = vec4( final, 1.0 );
}
