cbuffer buffer : register(b0)
{
    column_major matrix _ProjectionMatrix;
    float uTanHalfFov;
    float uAspectRatio;
    int uKernelSize;
    float3 uKernelOffsets[ 128 ];
    float uRadius;
    float uPower;
    float uLinearDepthConstant;
}

struct VS_INPUT
{
    float3 Pos : POSITION;
    float2 Tex : TEXCOORD0;
    float3 Normal : NORMAL;
    float4 Tangent : TANGENT;
    float4 Color : COLOR;
};

struct PS_INPUT
{
    float4 Pos : SV_POSITION;
    float2 Tex : TEXCOORD0;
    float3 viewDir : VIEW_DIR;
};

PS_INPUT VS( VS_INPUT input )
{
    PS_INPUT output = (PS_INPUT)0;
    float4 position = float4(input.Pos, 1);

    output.Pos = position;
    output.Tex = input.Tex;
    output.viewDir = float3( -position.x * uTanHalfFov * uAspectRatio, -position.y * uTanHalfFov, 1.0 );

    return output;
}

Texture2D depthNormalMap : register(t0);
Texture2D noiseMap : register(t1);
SamplerState SampleType1 : register(s0);
SamplerState SampleType2 : register(s1);

float ssao( float3x3 kernelBasis, float3 originPos, float radius )
{
    float occlusion = 0.0f;

    for (int i = 0; i < uKernelSize; ++i)
    {
        float3 samplePos = mul( kernelBasis, uKernelOffsets[ i ] );
        samplePos = samplePos * radius + originPos;

        // project sample position:
        float4 offset = mul( _ProjectionMatrix, float4( samplePos, 1.0f ) );
        offset.xy /= offset.w; // only need xy
        offset.xy = offset.xy * 0.5f + 0.5f; // scale/bias to texcoords

        // get sample depth:
        float sampleDepth = depthNormalMap.Sample( SampleType1, offset.xy ).r;

        float rangeCheck = smoothstep( 0.0f, 1.0f, radius / abs( originPos.z - sampleDepth ) );
        occlusion += rangeCheck * step( sampleDepth, samplePos.z );
    }

    occlusion = 1.0f - (occlusion / float( uKernelSize ));
    return pow( abs( occlusion ), uPower );
}

float4 PS( PS_INPUT input ) : SV_Target
{
    int depthWidth, depthHeight;
    depthNormalMap.GetDimensions( depthWidth, depthHeight );

    int noiseWidth, noiseHeight;
    noiseMap.GetDimensions( noiseWidth, noiseHeight );

    float2 noiseTexCoords = float2( float( depthWidth ) / float( noiseWidth ), float( depthHeight ) / float( noiseHeight ) );
    noiseTexCoords *= input.Tex;

    // get view space origin:
    float originDepth = depthNormalMap.Sample( SampleType1, input.Tex ).r;
    float3 originPos = input.viewDir * originDepth;

    // get view space normal:
    float3 normal = -normalize( depthNormalMap.Sample( SampleType1, input.Tex ).gba );

    // construct kernel basis matrix:
    float3 rvec = normalize( noiseMap.Sample( SampleType2, noiseTexCoords ).rgb );
    float3 tangent = normalize( rvec - normal * dot( rvec, normal ) );
    float3 bitangent = cross( tangent, normal );
    float3x3 kernelBasis = float3x3( tangent, bitangent, normal );

    float s = ssao( kernelBasis, originPos, uRadius );

    return float4( s, s, s, s );
}
