#ifndef AETHER3D_H
#define AETHER3D_H

/**
   \file Aether3D.hpp
   \author Timo Wiren
   \date 2017-01-21
   License: zlib License
   Website: http://twiren.kapsi.fi/aether3d.html

   ┌─┬─┬─┐
   │ │0│ │ Aether3D Game Engine header.
   ├─┼─┼─┤ 
   │ │ │0│ Contains methods to create objects like
   ├─┼─┼─┤ Models, Materials, Cameras etc.
   │0│0│0│
   └─┴─┴─┘
 */
/**
   \mainpage Aether3D Game Engine Documentation

   \section Introduction

   Aether3D is an open source game engine focusing on simplicity, robustness, portability and forward-thinking.
   <br>
   In most cases it's built as a DLL and dynamically loaded into an application allowing to switch renderers easily.

   \section Compilation

   \subsection Windows
   
   You'll need Visual Studio 2017. Build one of the renderers in Renderers folder and then build and run one of the samples.

   \subsection osx OS X
   
   You'll need at least OS X Mountain Lion.  Make sure you have SDL2.framework in /Library/Frameworks
   <br>
   Build Renderers/GL32/Aether3D_GL_3.2.xcodeproj and build and run one of the Samples/.
   <br>
   You can also build on the command-line using Makefiles located in renderer and sample folders.

   \subsection iOS

   Locate the _iOS Xcode projects in Samples folder hierarchy and run them.

   \subsection Linux

   You'll need GCC 4.8 or newer or clang.
   <br>
   Required packages: libudev-dev, libsdl2-dev, libopenal-dev, libgl1-mesa-dev
   <br>
   Navigate to Renderers/GL32 and run make. Then navigate into one of the samples and run make.

   \section bugs Open Bugs
   <ul>
   <li>Camera's transform is not affected by the scene graph.</li>
   <li>Spotlight has visual bugs in GL 4.3 and D3D11 renderers.</li>
   <li>SSAO and shadows look weird in D3D11 renderer.</li>
   <li>Samples crash when loading the engine DLL when compiled with MinGW.</li>
   <li>When models are hot-reloaded, they don't preserve their materials.</li>
   <li>MSAA breaks shadows (tested on GL/macOS)</li>
   <li>Bug: In editor the sphere is lit even if a light is inside it.</li>
   <li>Bug: convert_obj produces garbage with meshes containing quads without normals.</li>
   <li>Bug: MSAA causes OpenGL error on GL43 on OpenWindow: OpenGL: glBlitFramebuffer failed because the framebuffer configurations require that the source and destination formats and sample counts match (GL_INVALID_OPERATION) [source=API type=ERROR severity=HIGH id=3010]</li>
   </ul>
 */
#ifndef INCLUDE_STRING_H
#    include <string>
#    define INCLUDE_STRING_H
#endif
#ifndef INCLUDE_MEMORY_H
#    include <memory>
#    define INCLUDE_MEMORY_H
#endif
#ifndef AUDIO_H
#    include "Audio.hpp"
#endif
#ifndef CAMERA_H
#    include "Camera.hpp"
#endif
#ifndef KEYBOARD_INPUT_H
#    include "KeyboardInput.hpp"
#endif
#ifndef LIGHT_H
#    include "Light.hpp"
#endif
#ifndef MOUSE_INPUT_H
#    include "MouseInput.hpp"
#endif
#ifndef MATERIAL_H
#    include "Material.hpp"
#endif
#ifndef MODEL_H
#    include "Model.hpp"
#endif
#ifndef QUATERNION_H
#    include "Quaternion.hpp"
#endif
#ifndef RENDERER_H
#    include "Renderer.hpp"
#endif
#ifndef SCENE_H
#    include "Scene.hpp"
#endif
#ifndef TEXTURE_H
#    include "Texture.hpp"
#endif
#ifndef INCLUDE_VEC3_H
#    include "Vec3.hpp"
#endif

/// Handles DLL loading and engine object creation.
class Aether3D_API
{
public:
    Aether3D_API();
    virtual ~Aether3D_API();

    /**
     Converts return code to an error string.

     \param returnCode Integer representation of Error::Enum.
     \return Error string.
     */
    static const std::string& ReturnCodeToErrorString( int returnCode );

    /**
     Loads the engine DLL. Must be called before
     doing anything else with the engine.

     \param path DLL path. If extension is omitted, it's automatically added depending on the running platform.
     \return NO_ERRORS if loading DLL succeeded.
     \return DLL_NOT_FOUND if DLL could not be found or read.
     \return UNSUPPORTED_RENDERER if GPU driver doesn't support requested renderer.
     \return FACTORY_NOT_FOUND if DLL is found but could not be loaded by LoadLibraryA() on Windows or dlopen() on OSX or Linux.
     */
    Error::Enum LoadDLL( const std::string& path );

    /// \return new Audio. Caller is responsible for freeing the allocated memory.
    Audio* MakeAudio();

    /// \return new Camera. Caller is responsible for freeing the allocated memory.
    Camera* MakeCamera();

    /// \return new KeyboardInput. Caller is responsible for freeing the allocated memory.
    KeyboardInput* MakeKeyboardInput();

    /**
        \param lightType Light Type.
        \return new Light. Caller is responsible for freeing the allocated memory.
    */
    Light* MakeLight( Light::Type lightType );
    
    /// \return new Material. Caller is responsible for freeing the allocated memory.  Must be called after Renderer::OpenWindow().
    Material* MakeMaterial();

    /// \return new MouseInput. Caller is responsible for freeing the allocated memory.
    MouseInput* MakeMouseInput();

    /**
       Loads a model from .ae3d file. Caller is responsible for freeing the allocated memory.
       Must be called after Renderer::OpenWindow().

       \param path Model path.
       \param outError Error.
       \return Model.
     */
    Model* MakeModel( const std::string& path, Error::Enum& outError );

    /**
       Creates a renderer and returns its instance.  It's a singleton.  Doesn't allocate memory so must not be deleted.
     
       \return Instance to Renderer.
     */
    Renderer* MakeRenderer( const std::string& builtinAssetPath );

    /// \return new Scene. Caller is responsible for freeing the allocated memory.
    Scene* MakeScene();

    /**
       Loads a texture.  Formats supported: .png, .jpg, .tga, .bmp, .dds and .pvr depending on platform.
       Must be called after Renderer::OpenWindow().

       \param path Texture path.
       \param useMipmaps True, if mipmaps should be generated.
       \param sRGB True, if the texture is sRGB. Normal maps should use false.
       \param tiling Tiling.
       \param filter Filter.
       \return Texture.
     */
    Texture* MakeTexture( const std::string& path, bool useMipmaps, bool sRGB, Texture::Tiling tiling, Texture::Filter filter );

    /**
       Loads a cube map texture.  Formats supported: .png, .jpg, .tga, .bmp etc.
       Must be called after Renderer::OpenWindow().

       \todo Document sides also in API-friendly way: "positive z" etc.

       \param left Left.
       \param right Right.
       \param top Top.
       \param bottom Bottom.
       \param front Front.
       \param back Back.
     */
    Texture* MakeTexture( const char* left,  const char* right,
                          const char* top,   const char* bottom,
                          const char* front, const char* back );
    /**
       Creates a 2D render texture.
       Must be called after Renderer::OpenWindow().

       \param width Width in pixels.
       \param height Height in pixels.
       \return Texture.
     */
    Texture* MakeRenderTexture( int width, int height );

    /**
     Creates a cube map render texture.
     Must be called after Renderer::OpenWindow().

     \param size Width and height in pixels.
     \return Texture.
     */
    Texture* MakeRenderTextureCube( int size );

    /**
     Loads a texture from an atlas.  Formats supported: .png, .jpg, .tga, .bmp etc.
     Must be called after Renderer::OpenWindow().

     \param atlasPath Texture path.
     \param atlasMetadataPath Meta data path.  Format must be CEGUI/Ogre (eg. Texture Packer exports them).
     \param fileName Name of texture in atlas.
     \return Texture.
     */
    Texture* MakeTexture( const std::string& atlasPath, const std::string& atlasMetadataPath, const std::string& fileName );

private:
    struct Impl;
    std::unique_ptr< Impl > m;
};

#endif
