#ifndef MATH_UTIL_H
#define MATH_UTIL_H

#ifndef INCLUDE_VECTOR_H
#    include <vector>
#    define INCLUDE_VECTOR_H
#endif

#ifndef VEC3_H
#    include "Vec3.hpp"
#endif

struct Face
{
    Face() : a( 0 ), b( 0 ), c( 0 ) {}
    
    Face( unsigned short fa, unsigned short fb, unsigned short fc )
    : a( fa )
    , b( fb )
    , c( fc )
    {}
    
    unsigned short a, b, c;
};

namespace MathUtil
{
    const float deg2rad = 3.1415926535f / 180.0f;
    const float rad2deg = 180.0f / 3.1415926535f;
    const float tolerance = 0.00001f;
    const float pi = 3.1415926535f;
    
    /**
       \param f1 Value 1.
       \param f2 Value 2.
       \return True, if f1 almost equals f2.  Difference is at most epsilon.
     */
    bool AlmostEquals( float f1, float f2 );

    /**
      Converts a 32-bit float to 16-bit.
      \param valueToConvert a 32-bit float.
      \return 16-bit float.
    */
    unsigned short ConvertF32ToF16( float valueToConvert );

    /**
       Gets AABB's minimum and maximum corner.

       \param points Vector of 8 corner points of AABB.
       \param outMin Minimum of points is returned here.
       \param outMax Maximum of points is returned here.
     */
    void GetMinMax( const std::vector< Vec3 >& points, Vec3& outMin, Vec3& outMax );

    /**
       Solves corners of a cube defined by min and max corner points.
     
       \param min Minimum corner.
       \param max Maximum corner.
       \param outCorners Vector of 8 corner points.  Doesn't preserve previous contents, allocates memory.
     */
    void GetCorners( const Vec3& min, const Vec3& max, std::vector< Vec3 >& outCorners );
    
    /** \return f floored. */
    float Floor( float f );
    
    /** \return Rounds f upward, returning the smallest integral value that is not less than f. */
    float Ceil( float f );
    
    /** \return f rounded. */
    float Round( float f );
    
    /**
       Linear interpolation.
     
       \param start Range's minimum value.
       \param end Range's maximum value.
       \param amount Value in range [0, 1].
       \return start, if amount is 0, end if amount is 1, value between start and end otherwise.
     */
    float Lerp( float start, float end, float amount );

    /**
       Returns a pseudo-random number in range [aMin, aMax].
     
       \param aMin Minimum value.
       \param aMax Maximum value.
       \return Value between aMin and aMax (inclusive).
     */
    float Random( float aMin, float aMax );

    /**
       Maps a value in range [oldMin, oldMax] to range [newMin, newMax]

       \param value Value.
       \param oldMin Old range minimum.
       \param oldMax Old range maximum.
       \param newMin New range minimum.
       \param newMax New range maximum.
       \return value mapped in [newMin, newMax].
     */ 
    float Remap( float value, float oldMin, float oldMax, float newMin, float newMax );

    /**
       Tests if a value is a NaN (not-a-number)
       Limitation: Works in debug mode, but release mode correctness is not guaranteed.
     
       \param f Value to be tested.
       \return True, if f is a NaN, false otherwise.
     */
    bool IsNaN( float f );

    /**
       Tests if a value is power-of-two.
     
       \param i Value to be tested.
       \return True, if i is power-of-two, false otherwise.
     */
    bool IsPowerOfTwo( unsigned i );

    template< typename T >
    T Min2( T t1, T t2 )
    {
        return t1 < t2 ? t1 : t2;
    }

    template< typename T >
    T Max2( T t1, T t2 )
    {
        return t1 > t2 ? t1 : t2;
    }
    
    /**
     Intersects an axis-aligned bounding box with a ray.
     
     \param origin Ray's starting point.
     \param target Viewing point.
     \param min Bounding box's minimum corner.
     \param max Bounding box's maximum corner.
     \return Distance to the intersection or a negative value if there is no intersection.
     */
    float IntersectRayAABB( const Vec3& origin, const Vec3& target, const Vec3& min, const Vec3& max );

    /**
       Intersects an AABB with a sphere.

       \param sphereOrigin Sphere's origin.
       \param sphereRadius Sphere's radius.
       \param aabbMin AABB min.
       \param aabbMax AABB max.
       \return True, if the sphere intersects or is inside the AABB, false otherwise.
     */
    bool IntersectSphereAndAABB( const Vec3& sphereOrigin, float sphereRadius, const Vec3& aabbMin, const Vec3& aabbMax );

    /**
       Intersects an AABB with another AABB.

       \param aabb1Min First AABB's minimum corner.
       \param aabb1Max First AABB's maximum corner.
       \param aabb2Min Second AABB's minimum corner.
       \param aabb2Max Second AABB's maximum corner.
       \return True, if AABBs intersect, false otherwise.
     */
    bool IntersectAABBandAABB( const Vec3& aabb1Min, const Vec3& aabb1Max, const Vec3& aabb2Min, const Vec3& aabb2Max );

    /**
     Intersects a triangle list with a ray using the Moller and Trumbore method.
     
     \param origin Ray's starting point.
     \param target Viewing point.
     \param vertices Vertices.
     \param faces Faces.
    */
    float IntersectRayTriangles( const Vec3& origin, const Vec3& target, const std::vector< Vec3 >& vertices, const std::vector< Face >& faces );
}
#endif
