#ifndef RENDERER_H
#define RENDERER_H

#ifndef ERRORS_H
#    include "Errors.hpp"
#endif

class Scene;
class Texture;
struct Vec3;
struct Vec4;

/// Renders models and UI.
class Renderer
{
  public:
    /** Used to verify that the DLL version matches header version. */
    static const unsigned apiVersion = 23;

    /* Destructor. */
    virtual ~Renderer() {}

    /**
       \brief Adds a path where Material will search for shaders.
       Paths are searched in order they are added and the first matching shader is used when using Material.SetShaders().

       \param path Path. Must end with '/'.
     */
    virtual void AddShaderSearchPath( const char* path ) const = 0;
    
    /** Clears the active framebuffer using color set by SetClearColor. Implicitly called by Draw( scene ). */
    virtual void ClearScreen() const = 0;

    /** Releases resources and terminates the running application. */
    virtual void Deinit() = 0;

    /**
     Draws a texture.

     \param texture Texture.
     \param x Pixel coordinate from the left side of the window.
     \param y Pixel coordinate from the top of the window.
     \param color Color tint in range 0-1. .w is opacity.
     */
    virtual void Draw( const Texture* texture, int x, int y, const Vec4& color ) = 0;

    /**
     Draws a texture in custom dimensions.

     \param texture Texture.
     \param x Pixel coordinate from the left side of the window.
     \param y Pixel coordinate from the top of the window.
     \param width Width in pixels.
     \param height Height in pixels.
     \param color Color tint in range 0-1. .w is opacity.
     */
    virtual void Draw( const Texture* texture, int x, int y, int width, int height, const Vec4& color ) = 0;

    /**
       Draws a string of text (screen-space).
     
       \param text Text to draw.  Characters not in font will be rendered empty.
       \param fontName Font name.  Must have been added using CreateFont().  Builtin font name is "aether3d_font".
       \param scale Scale of the font.
       \param x Pixel coordinate of the left side of the text.
       \param y Pixel coordinate of the top side of the text.
       \param color Color tint in range 0-1. .w is opacity.
     */
    virtual void Draw( const char* text, const char* fontName, float scale, int x, int y, const Vec4& color ) = 0;

    /**
       Draws a scene.  Should only be called once per frame.  Render-to-texture Cameras are rendered before normal cameras.
     
       \param scene Scene.
     */
    virtual void Draw( Scene* scene ) = 0;

    /**
       Draws lines in 3D.
     
       \param lineArray Line geometry.
       \param arraySize lineArray length.
       \param color Color.
     */
    virtual void DrawLines( const Vec3* lineArray, int arraySize, const Vec4& color ) = 0;
    
    /** Draws frame time and draw call count. */
    virtual void DrawStatistics() = 0;

    /** Renders everything that was called with Draw() methods.  Call before SwapBuffers(). */
    virtual void FinishFrame() = 0;

    /** \return Time in seconds. */
    virtual double GetTime() const = 0;

    /** \return Window height in pixels. */
    virtual int Height() const = 0;

    /**
       Creates a bitmap font from a file and metadata.
     
       \param fontName Font's name.  Used when drawing text.
       \param fontBitmapPath Font's path.  Supported image formats are .png, .bmp, .tga, .jpg.
       \param fontMetadataPath Font's meta data path.  Currently supports only meta data generated by BMFont.
       \return NO_ERRORS on success, error otherwise.
     */
    virtual Error::Enum LoadFont( const char* fontName, const char* fontBitmapPath, const char* fontMetadataPath ) = 0;

    /**
       Opens a window, creating rendering context.
     
       \param width Width in pixels.  0 means desktop width.
       \param height Height in pixels.  0 means desktop height.
       \param fullScreen Should the window be fullscreen.
       \param antialiasSamples How many antialias sample buffers are used?  Pass 1 to disable antialiasing.
       \param hasWindowAndContext True, if the DLL creates the graphics window and context. False, if the DLL is loaded into some other application's window/context, like Editor.
     */
    virtual Error::Enum OpenWindow( int width, int height, bool fullScreen, int antialiasSamples, bool hasWindowAndContext ) = 0;

    /**
       \param width New width.
       \param height New height.
     */
    virtual void Resize( int width, int height ) = 0;
    
    /**
       Sets bloom.  Defaults to no bloom.
     
       \param enable True if bloom should be applied, false otherwise.
       \param factor Bloom intensity factor.  Values near 0.5 should be ok.
       \param blurIterations Blur iterations.
     */
    virtual void SetBloom( bool enable, float factor, int blurIterations ) = 0;
    
    /**
       Sets clear color.  Must be set after OpenWindow().  Defaults to black.

       \param red Red (0.0-1.0)
       \param green Green (0.0-1.0)
       \param blue Blue (0.0-1.0)
     */
    virtual void SetClearColor( float red, float green, float blue ) = 0;

    /**
       Sets SSAO.  Defaults to disabled.

       \param enable True if SSAO should be applied, false otherwise.
       \param radius Radius.
       \param power Power.
     */
    virtual void SetSSAO( bool enable, float radius, float power ) = 0;

    /** \param title Title */
    virtual void SetWindowTitle( const char* title ) const = 0;

    /** Shows the rendering on screen. */
    virtual void SwapBuffers() = 0;

    /** \return window width in pixels. */
    virtual int Width() const = 0;

    /** Begins a GPU timer query. */
    virtual void BeginQuery() = 0;

    /** \return Elapsed GPU time from BeginQuery to EndQuery in milliseconds. */
    virtual double EndQuery() = 0;

    /** \return True, if the window has focus. */
    virtual bool HasFocus() const = 0;
};

#endif
