#ifndef SCENE_H
#define SCENE_H

#ifndef INCLUDE_STRING_H
#    include <string>
#    define INCLUDE_STRING_H
#endif
#ifndef INCLUDE_LIST_H
#    include <list>
#    define INCLUDE_LIST_H
#endif

class Camera;
class Light;
class Material;
class Model;
class Texture;
class SceneGraphNode;
struct Vec3;

enum class CollisionTest
{
    AABB,
    Triangles
};

/// Stores scene in a scene graph.
class Scene
{
  public:
    /** Destructor. */
    virtual ~Scene() {}

    /** \param camera Camera. */
    virtual void Add( Camera* camera ) = 0;

    /** \param light Light. */
    virtual void Add( Light* light ) = 0;

    /** \param model Model. */
    virtual void Add( Model* model ) = 0;

    /** \param node Node to be removed from the scene graph. */
    virtual void Remove( SceneGraphNode* node ) = 0;

    /** \return The camera that is not set to render into a texture or nullptr if there is no such camera. */
    virtual Camera* GetNonRTCamera() = 0;
    
    /**
       Finds a material that was loaded from a .scene file.

       \param name Material's name.
       \return Material or null if Scene does not contain a Material named name.
     */
    virtual Material* GetMaterial( const char* name ) = 0;

    /**
       Finds a model.

       \param name Model's name.
       \return Model or null if Scene does not contain a Model named name.
     */
    virtual Model* GetModel( const char* name ) = 0;
    
    /**
     Gets colliders for a ray shot through a screen point.
     The first object's first mesh is the closest.
     
     \param screenX Screen x coordinate in pixels.
     \param screenY Screen y coordinate in pixels.
     \param collisionTest Collision test to perform.
     \param maxDepth Maximum depth where collisions are tested.
     \return List of (Model, collided mesh names)
     */
    virtual std::list< std::pair< Model*, std::list< std::string > > > GetColliders( int screenX, int screenY, CollisionTest collisionTest, float maxDepth ) const = 0;

    /**
       Gets colliders for an AABB.  Note that collision with a model's AABB does
       not necessarily mean that any of the meshes were included in the collision.

       \param aabbMin AABB's min value in world-space.
       \param aabbMax AABB's max value in world-space.
       \param collisionTest Collision test to perform.
       \return List of (Model, collided mesh names)
     */
    virtual std::list< std::pair< Model*, std::list< std::string > > > GetColliders( const Vec3& aabbMin, const Vec3& aabbMax, CollisionTest collisionTest ) const = 0;

    /** 
       Loads a scene from a text file.  Does not empty current scene contents.
     
       \param path .scene file.
       \return True on success, false on failure.  Failure reason is printed into console.
     */
    virtual bool LoadFromFile( const char* path ) = 0;

    /**
      Saves the scene into an ascii file.
      \param path .scene file.
      \return True, if succeeded, false otherwise.
      */
    virtual bool SaveToFile( const char* path ) = 0;

    /** \return Root node. */
    virtual SceneGraphNode& RootNode() = 0;
    
    /** \param cubeMap Cube map that's used for skybox cube mesh. */
    virtual void SetSkybox( const Texture* cubeMap ) = 0;

    /**
       Called automatically when a node is added into scene to update
       Scene's internal representation.
     
       \param node Node.
     */
    virtual void OnNodeAdded( SceneGraphNode* node ) = 0;
};

#endif
