#include "ComputeShader.hpp"
#include <fstream>
#include <unordered_map>
#include <d3d11.h>
#include <d3dcompiler.h>
#include "RendererD3D11State.hpp"
#include "Log.hpp"
#include "Statistics.hpp"
#include "TextureImpl.hpp"

namespace ae3d
{
    // Used to release resources.
    std::list< ID3D11ComputeShader* > computeShaders;
}

struct ComputeShader::Impl
{
    ID3D11ComputeShader* shader = nullptr;
};

ComputeShader::ComputeShader() : m( new Impl ) {}

ComputeShader::~ComputeShader()
{
}

void ComputeShader::Deinit()
{
    for (auto& shader : ae3d::computeShaders)
    {
        shader->Release();
    }

    ae3d::computeShaders.clear();
}

void ComputeShader::Load( const char* path, Error::Enum& outError )
{
    DWORD flags = D3DCOMPILE_ENABLE_STRICTNESS;
#ifndef NDEBUG
    flags |= D3DCOMPILE_DEBUG;
#endif

    std::ifstream shaderStream( path );
    std::string shaderSrc = std::string( std::istreambuf_iterator< char >( shaderStream ),
        std::istreambuf_iterator< char >() );

    ID3DBlob* pErrorBlob = nullptr;
    ID3DBlob* pBlob = nullptr;
    HRESULT hr = D3DCompile( shaderSrc.c_str(), shaderSrc.size(), nullptr, nullptr, nullptr, "CS", "cs_5_0", flags, 0, &pBlob, &pErrorBlob );

    if (FAILED( hr ))
    {
        Log::Print( "Could not compile %s: %s", path, (LPCSTR)pErrorBlob->GetBufferPointer() );
        Log::Assert( false, "Could not compile compute shader. See debug output for details." );
        pErrorBlob->Release();
        outError = Error::SHADER_LINK_FAILED;
        return;
    }

    hr = rendererD3D11State.d3dDev->CreateComputeShader( pBlob->GetBufferPointer(), pBlob->GetBufferSize(), nullptr, &m->shader );

    if (FAILED( hr ))
    {
        Log::Print( "Could not create %s.", path );
        Log::Assert( false, "Could not create compute shader." );
        outError = Error::CONTRACT_VIOLATION;
        return;
    }

    ae3d::computeShaders.push_back( m->shader );

    pBlob->Release();
}

void ComputeShader::Use()
{
    Log::Assert( m->shader != nullptr, "Compute shader not initialized." );
    rendererD3D11State.immContext->CSSetShader( m->shader, nullptr, 0 );
}

void ComputeShader::Dispatch( int numGroupsX, int numGroupsY, int numGroupsZ )
{
#if DEBUG
    rendererD3D11State.debugDevice->ValidateContextForDispatch( rendererD3D11State.immContext );
#endif
    rendererD3D11State.immContext->Dispatch( numGroupsX, numGroupsY, numGroupsZ );
}

