#include "ForwardPlus.hpp"
#include <string>
#include <vector>
#include <d3d11.h>
#include "Log.hpp"
#include "Matrix.hpp"
#include "MathUtil.hpp"
#include "Platform.hpp"
#include "RendererD3D11State.hpp"
#include "RendererTypes.hpp"
#include "TextureImpl.hpp"
#include "Shader.hpp"
#include "Vec3.hpp"

#define AE3D_SAFE_RELEASE(x) if (x) { x->Release(); x = nullptr; }

extern RendererD3D11State rendererD3D11State;

void SetDebugLabel( ID3D11DeviceChild* d3dObj, const std::string& name );

#pragma pack(push,1)
struct CB_PER_FRAME
{
    Matrix44 projection;
    Matrix44 projectionInv;
    Vec3 cameraPos;
    float alphaTest;
    unsigned numLights;
    unsigned windowWidth;
    unsigned windowHeight;
    unsigned maxNumLightsPerTile;
    Vec4 linearDepthConstant;
    Matrix44 view;
};
#pragma pack(pop)

// TODO: Merge with ForwardPlusGL's identical method.
SpotParams PackSpotParams( const Vec3& lightDir, float cosineOfConeAngle, float falloffRadius )
{
    //assert( cosineOfConeAngle > 0.0f );
    //assert( falloffRadius > 0.0f );
    SpotParams spotParams;
    spotParams.lightDirX = MathUtil::ConvertF32ToF16( lightDir.x );
    spotParams.lightDirY = MathUtil::ConvertF32ToF16( lightDir.y );
    spotParams.cosineOfConeAngleAndLightDirZSign = MathUtil::ConvertF32ToF16( cosineOfConeAngle );
    spotParams.falloffRadius = MathUtil::ConvertF32ToF16( falloffRadius );

    // put the sign bit for light dir z in the sign bit for the cone angle
    // (we can do this because we know the cone angle is always positive)
    if (lightDir.z < 0.0f)
    {
        spotParams.cosineOfConeAngleAndLightDirZSign |= 0x8000;
    }
    else
    {
        spotParams.cosineOfConeAngleAndLightDirZSign &= 0x7FFF;
    }

    return spotParams;
}

struct ForwardPlus::Impl
{
    void InitPointLights();
    void InitSpotLights();

    std::vector< Vec4 > pointLightCenterAndRadius;
    std::vector< Vec4 > pointLightColors;
    int numActivePointLights = 0;
    std::vector< Vec4 > spotLightCenterAndRadius;
    std::vector< Vec4 > spotLightColors;
    std::vector< Vec4 > spotLightParams;
    int numActiveSpotLights = 0;

    static const int MAX_NUM_LIGHTS = 2048;
    static const unsigned MAX_NUM_LIGHTS_PER_TILE = 544;
    static const int TILE_RES = 16;

    ID3D11ShaderResourceView*  pointLightBufferCenterAndRadiusSRV = nullptr;
    ID3D11Buffer*              pointLightBufferCenterAndRadius = nullptr;
    ID3D11ShaderResourceView*  pointLightBufferColorsSRV = nullptr;
    ID3D11Buffer*              pointLightBufferColors = nullptr;
    ID3D11ShaderResourceView*  spotLightBufferCenterAndRadiusSRV = nullptr;
    ID3D11Buffer*              spotLightBufferCenterAndRadius = nullptr;
    ID3D11ShaderResourceView*  spotLightBufferColorsSRV = nullptr;
    ID3D11Buffer*              spotLightBufferColors = nullptr;
    ID3D11ShaderResourceView*  spotLightBufferParamsSRV = nullptr;
    ID3D11Buffer*              spotLightBufferParams = nullptr;

    ID3D11Buffer*              cbPerFrame11 = nullptr;
    ID3D11Buffer*              lightIndexBuffer = nullptr;
    ID3D11ShaderResourceView*  lightIndexBufferSRV = nullptr;
    ID3D11UnorderedAccessView* lightIndexBufferUAV = nullptr;
};

void ForwardPlus::Impl::InitPointLights()
{
    pointLightCenterAndRadius.resize( MAX_NUM_LIGHTS );
    pointLightColors.resize( MAX_NUM_LIGHTS );

    // Create the point light buffer (center and radius)
    D3D11_BUFFER_DESC lightBufferDesc;
    ZeroMemory( &lightBufferDesc, sizeof( lightBufferDesc ) );
    lightBufferDesc.Usage = D3D11_USAGE_DYNAMIC;
    lightBufferDesc.ByteWidth = (UINT)pointLightCenterAndRadius.size() * sizeof( Vec4 );
    lightBufferDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
    lightBufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

    D3D11_SUBRESOURCE_DATA initData;
    initData.pSysMem = pointLightCenterAndRadius.data();
    initData.SysMemPitch = 0;
    initData.SysMemSlicePitch = 0;

    HRESULT hr = rendererD3D11State.d3dDev->CreateBuffer( &lightBufferDesc, &initData, &pointLightBufferCenterAndRadius );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create point light buffer!" );
    }

    SetDebugLabel( pointLightBufferCenterAndRadius, "point light buffer" );

    D3D11_SHADER_RESOURCE_VIEW_DESC SRVDesc;
    ZeroMemory( &SRVDesc, sizeof( D3D11_SHADER_RESOURCE_VIEW_DESC ) );
    SRVDesc.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
    SRVDesc.ViewDimension = D3D11_SRV_DIMENSION_BUFFER;
    SRVDesc.Buffer.ElementOffset = 0;
    SRVDesc.Buffer.ElementWidth = MAX_NUM_LIGHTS;
    hr = rendererD3D11State.d3dDev->CreateShaderResourceView( pointLightBufferCenterAndRadius, &SRVDesc, &pointLightBufferCenterAndRadiusSRV );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create point light buffer SRV!" );
    }

    SetDebugLabel( pointLightBufferCenterAndRadiusSRV, "point light buffer SRV" );

    // Create the point light buffer (color)
    D3D11_BUFFER_DESC colorBufferDesc;
    ZeroMemory( &colorBufferDesc, sizeof( colorBufferDesc ) );
    colorBufferDesc.Usage = D3D11_USAGE_DYNAMIC;
    colorBufferDesc.ByteWidth = (UINT)pointLightColors.size() * sizeof( Vec4 );
    colorBufferDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
    colorBufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

    initData.pSysMem = pointLightColors.data();
    initData.SysMemPitch = 0;
    initData.SysMemSlicePitch = 0;

    hr = rendererD3D11State.d3dDev->CreateBuffer( &colorBufferDesc, &initData, &pointLightBufferColors );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create point light color buffer!" );
    }

    SetDebugLabel( pointLightBufferColors, "point light color buffer" );

    ZeroMemory( &SRVDesc, sizeof( D3D11_SHADER_RESOURCE_VIEW_DESC ) );
    SRVDesc.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
    SRVDesc.ViewDimension = D3D11_SRV_DIMENSION_BUFFER;
    SRVDesc.Buffer.ElementOffset = 0;
    SRVDesc.Buffer.ElementWidth = MAX_NUM_LIGHTS;
    hr = rendererD3D11State.d3dDev->CreateShaderResourceView( pointLightBufferColors, &SRVDesc, &pointLightBufferColorsSRV );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create point light color buffer SRV!" );
    }

    SetDebugLabel( pointLightBufferColorsSRV, "point light color buffer SRV" );
}

void ForwardPlus::Impl::InitSpotLights()
{
    spotLightCenterAndRadius.resize( MAX_NUM_LIGHTS );
    spotLightColors.resize( MAX_NUM_LIGHTS );
    spotLightParams.resize( MAX_NUM_LIGHTS );

    // Create the spot light buffer (center and radius)
    D3D11_BUFFER_DESC lightBufferDesc;
    ZeroMemory( &lightBufferDesc, sizeof( lightBufferDesc ) );
    lightBufferDesc.Usage = D3D11_USAGE_DYNAMIC;
    lightBufferDesc.ByteWidth = (UINT)spotLightCenterAndRadius.size() * sizeof( Vec4 );
    lightBufferDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
    lightBufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

    D3D11_SUBRESOURCE_DATA initData;
    initData.pSysMem = spotLightCenterAndRadius.data();
    initData.SysMemPitch = 0;
    initData.SysMemSlicePitch = 0;

    HRESULT hr = rendererD3D11State.d3dDev->CreateBuffer( &lightBufferDesc, &initData, &spotLightBufferCenterAndRadius );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create spot light buffer!" );
    }

    SetDebugLabel( spotLightBufferCenterAndRadius, "spot light buffer" );

    D3D11_SHADER_RESOURCE_VIEW_DESC SRVDesc;
    ZeroMemory( &SRVDesc, sizeof( D3D11_SHADER_RESOURCE_VIEW_DESC ) );
    SRVDesc.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
    SRVDesc.ViewDimension = D3D11_SRV_DIMENSION_BUFFER;
    SRVDesc.Buffer.ElementOffset = 0;
    SRVDesc.Buffer.ElementWidth = MAX_NUM_LIGHTS;
    hr = rendererD3D11State.d3dDev->CreateShaderResourceView( spotLightBufferCenterAndRadius, &SRVDesc, &spotLightBufferCenterAndRadiusSRV );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create spot light buffer SRV!" );
    }

    SetDebugLabel( spotLightBufferCenterAndRadiusSRV, "spot light buffer SRV" );

    // Create the spot light buffer (color)
    D3D11_BUFFER_DESC colorBufferDesc;
    ZeroMemory( &colorBufferDesc, sizeof( colorBufferDesc ) );
    colorBufferDesc.Usage = D3D11_USAGE_DYNAMIC;
    colorBufferDesc.ByteWidth = (UINT)spotLightColors.size() * sizeof( Vec4 );
    colorBufferDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
    colorBufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

    initData.pSysMem = spotLightColors.data();
    initData.SysMemPitch = 0;
    initData.SysMemSlicePitch = 0;

    hr = rendererD3D11State.d3dDev->CreateBuffer( &colorBufferDesc, &initData, &spotLightBufferColors );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create spot light color buffer!" );
    }

    SetDebugLabel( spotLightBufferColors, "spot light color buffer" );

    ZeroMemory( &SRVDesc, sizeof( D3D11_SHADER_RESOURCE_VIEW_DESC ) );
    SRVDesc.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
    SRVDesc.ViewDimension = D3D11_SRV_DIMENSION_BUFFER;
    SRVDesc.Buffer.ElementOffset = 0;
    SRVDesc.Buffer.ElementWidth = MAX_NUM_LIGHTS;
    hr = rendererD3D11State.d3dDev->CreateShaderResourceView( spotLightBufferColors, &SRVDesc, &spotLightBufferColorsSRV );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create spot light color buffer SRV!" );
    }

    SetDebugLabel( spotLightBufferColorsSRV, "spot light color buffer SRV" );

    // Create the spot light buffer (params)
    D3D11_BUFFER_DESC paramsBufferDesc;
    ZeroMemory( &paramsBufferDesc, sizeof( paramsBufferDesc ) );
    paramsBufferDesc.Usage = D3D11_USAGE_DYNAMIC;
    paramsBufferDesc.ByteWidth = (UINT)spotLightParams.size() * sizeof( Vec4 );
    paramsBufferDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
    paramsBufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

    initData.pSysMem = spotLightParams.data();
    initData.SysMemPitch = 0;
    initData.SysMemSlicePitch = 0;

    hr = rendererD3D11State.d3dDev->CreateBuffer( &paramsBufferDesc, &initData, &spotLightBufferParams );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create spot light params buffer!" );
    }

    SetDebugLabel( spotLightBufferParams, "spot light params buffer" );

    ZeroMemory( &SRVDesc, sizeof( D3D11_SHADER_RESOURCE_VIEW_DESC ) );
    SRVDesc.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
    SRVDesc.ViewDimension = D3D11_SRV_DIMENSION_BUFFER;
    SRVDesc.Buffer.ElementOffset = 0;
    SRVDesc.Buffer.ElementWidth = MAX_NUM_LIGHTS;
    hr = rendererD3D11State.d3dDev->CreateShaderResourceView( spotLightBufferParams, &SRVDesc, &spotLightBufferParamsSRV );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create spot light params buffer SRV!" );
    }

    SetDebugLabel( spotLightBufferParamsSRV, "spot light params buffer SRV" );
}

ForwardPlus::ForwardPlus() : m( new Impl )
{
}

ForwardPlus::~ForwardPlus()
{
}

void ForwardPlus::Init()
{
    m->InitPointLights();
    m->InitSpotLights();

    // Creates constant buffers.
    D3D11_BUFFER_DESC CBDesc;
    ZeroMemory( &CBDesc, sizeof( CBDesc ) );
    CBDesc.Usage = D3D11_USAGE_DYNAMIC;
    CBDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    CBDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

    CBDesc.ByteWidth = sizeof( CB_PER_FRAME );
    HRESULT hr = rendererD3D11State.d3dDev->CreateBuffer( &CBDesc, NULL, &m->cbPerFrame11 );
    
    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create constant buffer!" );
    }

    SetDebugLabel( m->cbPerFrame11, "Forward+ per-frame CB" );

    // Light index buffer.

    const unsigned numTiles = GetNumTilesX() * GetNumTilesY();
    const unsigned maxNumLightsPerTile = GetMaxNumLightsPerTile();

    D3D11_BUFFER_DESC BufferDesc;
    ZeroMemory( &BufferDesc, sizeof( BufferDesc ) );
    BufferDesc.Usage = D3D11_USAGE_DEFAULT;
    BufferDesc.ByteWidth = 4 * maxNumLightsPerTile * numTiles;
    BufferDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_UNORDERED_ACCESS;
    hr = rendererD3D11State.d3dDev->CreateBuffer( &BufferDesc, NULL, &m->lightIndexBuffer );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create light index buffer!" );
    }

    SetDebugLabel( m->lightIndexBuffer, "LightIndexBuffer" );

    D3D11_SHADER_RESOURCE_VIEW_DESC SRVDesc;
    ZeroMemory( &SRVDesc, sizeof( D3D11_SHADER_RESOURCE_VIEW_DESC ) );
    SRVDesc.Format = DXGI_FORMAT_R32_UINT;
    SRVDesc.ViewDimension = D3D11_SRV_DIMENSION_BUFFER;
    SRVDesc.Buffer.ElementOffset = 0;
    SRVDesc.Buffer.ElementWidth = maxNumLightsPerTile * numTiles;
    hr = rendererD3D11State.d3dDev->CreateShaderResourceView( m->lightIndexBuffer, &SRVDesc, &m->lightIndexBufferSRV );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create light index SRV!" );
    }

    SetDebugLabel( m->lightIndexBufferSRV, "LightIndexBufferSRV" );

    D3D11_UNORDERED_ACCESS_VIEW_DESC UAVDesc;
    ZeroMemory( &UAVDesc, sizeof( D3D11_UNORDERED_ACCESS_VIEW_DESC ) );
    UAVDesc.Format = DXGI_FORMAT_R32_UINT;
    UAVDesc.ViewDimension = D3D11_UAV_DIMENSION_BUFFER;
    UAVDesc.Buffer.FirstElement = 0;
    UAVDesc.Buffer.NumElements = maxNumLightsPerTile * numTiles;
    hr = rendererD3D11State.d3dDev->CreateUnorderedAccessView( m->lightIndexBuffer, &UAVDesc, &m->lightIndexBufferUAV );

    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to create point light index UAV!" );
    }

    SetDebugLabel( m->lightIndexBufferUAV, "LightIndexBufferUAV" );
}

void ForwardPlus::Deinit()
{
    AE3D_SAFE_RELEASE( m->pointLightBufferCenterAndRadius )
    AE3D_SAFE_RELEASE( m->pointLightBufferCenterAndRadiusSRV )
    AE3D_SAFE_RELEASE( m->pointLightBufferColors )
    AE3D_SAFE_RELEASE( m->pointLightBufferColorsSRV )
    AE3D_SAFE_RELEASE( m->spotLightBufferCenterAndRadius )
    AE3D_SAFE_RELEASE( m->spotLightBufferCenterAndRadiusSRV )
    AE3D_SAFE_RELEASE( m->spotLightBufferColors )
    AE3D_SAFE_RELEASE( m->spotLightBufferColorsSRV )
    AE3D_SAFE_RELEASE( m->spotLightBufferParams )
    AE3D_SAFE_RELEASE( m->spotLightBufferParamsSRV )
    AE3D_SAFE_RELEASE( m->cbPerFrame11 )
    AE3D_SAFE_RELEASE( m->lightIndexBuffer )
    AE3D_SAFE_RELEASE( m->lightIndexBufferSRV )
    AE3D_SAFE_RELEASE( m->lightIndexBufferUAV )
}

//--------------------------------------------------------------------------------------
// Adjust max number of lights per tile based on screen height.
// This assumes that the demo has a constant vertical field of view (fovy).
//
// Note that the light culling tile size stays fixed as screen size changes.
// With a constant fovy, reducing the screen height shrinks the projected 
// view of the scene, and so more lights can fall into our fixed tile size.
//
// This function reduces the max lights per tile as screen height increases, 
// to save memory. It was tuned for this particular demo and is not intended 
// as a general solution for all scenes.
//--------------------------------------------------------------------------------------
unsigned ForwardPlus::GetMaxNumLightsPerTile() const
{
    const unsigned kAdjustmentMultipier = 32;

    // I haven't tested at greater than 1080p, so cap it
    const unsigned uHeight = (Screen::Height() > 1080) ? 1080 : Screen::Height();

    // adjust max lights per tile down as height increases
    return (m->MAX_NUM_LIGHTS_PER_TILE - (kAdjustmentMultipier * (uHeight / 120)));
}

int ForwardPlus::GetNextPointLightBufferIndex()
{
    if (m->numActivePointLights + m->numActiveSpotLights + 1 < m->MAX_NUM_LIGHTS)
    {
        ++m->numActivePointLights;
        return m->numActivePointLights - 1;
    }

    // Error handling is done in the caller.
    return -1;
}

int ForwardPlus::GetNextSpotLightBufferIndex()
{
    if (m->numActiveSpotLights + m->numActivePointLights + 1 < m->MAX_NUM_LIGHTS)
    {
        ++m->numActiveSpotLights;
        return m->numActiveSpotLights - 1;
    }

    // Error handling is done in the caller.
    return -1;
}

void ForwardPlus::CullLights( const RendererState& state, ComputeShader& cullShader, TextureImpl& depthNormalTarget )
{
    Matrix44 invProj = state.projection;
    Matrix44::Invert( invProj, invProj );

    D3D11_MAPPED_SUBRESOURCE MappedResource;
    HRESULT hr = rendererD3D11State.immContext->Map( m->cbPerFrame11, 0, D3D11_MAP_WRITE_DISCARD, 0, &MappedResource );
    if (hr != S_OK)
    {
        Log::Assert( false, "Failed to map light culling constant buffer!" );
        return;
    }

    CB_PER_FRAME* perFrame = static_cast< CB_PER_FRAME* >( MappedResource.pData );

    perFrame->projectionInv = invProj;
    perFrame->cameraPos = state.cameraPosition;
    perFrame->numLights = ((((unsigned)m->numActiveSpotLights & 0xFFFFu) << 16) | ((unsigned)m->numActivePointLights & 0xFFFFu));
    perFrame->windowWidth = Screen::Width();
    perFrame->windowHeight = Screen::Height();
    perFrame->maxNumLightsPerTile = GetMaxNumLightsPerTile();
    perFrame->view = state.view;

    rendererD3D11State.immContext->Unmap( m->cbPerFrame11, 0 );

    rendererD3D11State.immContext->VSSetConstantBuffers( 1, 1, &m->cbPerFrame11 );
    rendererD3D11State.immContext->PSSetConstantBuffers( 1, 1, &m->cbPerFrame11 );
    rendererD3D11State.immContext->CSSetConstantBuffers( 1, 1, &m->cbPerFrame11 );

    ID3D11ShaderResourceView* pNULLSRV = nullptr;
    ID3D11UnorderedAccessView* pNULLUAV = nullptr;

    rendererD3D11State.immContext->PSSetShaderResources( 5, 1, &pNULLSRV );
    rendererD3D11State.immContext->PSSetShaderResources( 6, 1, &pNULLSRV );
    rendererD3D11State.immContext->VSSetShader( nullptr, nullptr, 0 );
    rendererD3D11State.immContext->PSSetShader( nullptr, nullptr, 0 );

    cullShader.Use();
    rendererD3D11State.immContext->CSSetShaderResources( 0, 1, &m->pointLightBufferCenterAndRadiusSRV );
    rendererD3D11State.immContext->CSSetShaderResources( 2, 1, &m->spotLightBufferCenterAndRadiusSRV );

    auto depthSRV = depthNormalTarget.SRV();
    rendererD3D11State.immContext->CSSetShaderResources( 1, 1, &depthSRV );
    rendererD3D11State.immContext->CSSetUnorderedAccessViews( 0, 1, &m->lightIndexBufferUAV, nullptr );

    cullShader.Dispatch( GetNumTilesX(), GetNumTilesY(), 1 );

    rendererD3D11State.immContext->CSSetShader( nullptr, nullptr, 0 );
    rendererD3D11State.immContext->CSSetShaderResources( 0, 1, &pNULLSRV );
    rendererD3D11State.immContext->CSSetShaderResources( 1, 1, &pNULLSRV );
    rendererD3D11State.immContext->CSSetShaderResources( 2, 1, &pNULLSRV );
    rendererD3D11State.immContext->CSSetUnorderedAccessViews( 0, 1, &pNULLUAV, nullptr );

    rendererD3D11State.immContext->PSSetShaderResources( 5, 1, &m->pointLightBufferCenterAndRadiusSRV );
    rendererD3D11State.immContext->PSSetShaderResources( 6, 1, &m->lightIndexBufferSRV );
    rendererD3D11State.immContext->PSSetShaderResources( 7, 1, &m->pointLightBufferColorsSRV );
    rendererD3D11State.immContext->PSSetShaderResources( 8, 1, &m->spotLightBufferCenterAndRadiusSRV );
    rendererD3D11State.immContext->PSSetShaderResources( 9, 1, &m->spotLightBufferColorsSRV );
    rendererD3D11State.immContext->PSSetShaderResources( 10, 1, &m->spotLightBufferParamsSRV );
}

void ForwardPlus::SetPointLightColor( int bufferIndex, const Vec3& color )
{
    if (-1 < bufferIndex && bufferIndex < m->numActivePointLights)
    {
        m->pointLightColors[ bufferIndex ] = Vec4( color );

        // TODO: Find a way to update a single item instead of the whole buffer.
        D3D11_MAPPED_SUBRESOURCE mappedResource;
        ZeroMemory( &mappedResource, sizeof( D3D11_MAPPED_SUBRESOURCE ) );
        HRESULT hr = rendererD3D11State.immContext->Map( m->pointLightBufferColors, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource );
        
        if (hr == S_OK)
        {
            std::memcpy( mappedResource.pData, &m->pointLightColors[ 0 ].x, m->numActivePointLights * sizeof( Vec4 ) );
        }

        rendererD3D11State.immContext->Unmap( m->pointLightBufferColors, 0 );
    }
}

void ForwardPlus::SetSpotLightColor( int bufferIndex, const Vec3& color )
{
    if (-1 < bufferIndex && bufferIndex < m->numActiveSpotLights)
    {
        m->spotLightColors[ bufferIndex ] = Vec4( color );

        // TODO: Find a way to update a single item instead of the whole buffer.
        D3D11_MAPPED_SUBRESOURCE mappedResource;
        ZeroMemory( &mappedResource, sizeof( D3D11_MAPPED_SUBRESOURCE ) );
        HRESULT hr = rendererD3D11State.immContext->Map( m->spotLightBufferColors, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource );

        if (hr == S_OK)
        {
            std::memcpy( mappedResource.pData, &m->spotLightColors[ 0 ].x, m->numActiveSpotLights * sizeof( Vec4 ) );
        }

        rendererD3D11State.immContext->Unmap( m->spotLightBufferColors, 0 );
    }
}

void ForwardPlus::SetPointLightPositionAndRadius( int bufferIndex, const Vec4& positionAndRadius )
{
    if (-1 < bufferIndex && bufferIndex < m->numActivePointLights)
    {
        m->pointLightCenterAndRadius[ bufferIndex ] = positionAndRadius;

        // TODO: Find a way to update a single item instead of the whole buffer.
        D3D11_MAPPED_SUBRESOURCE mappedResource;
        ZeroMemory( &mappedResource, sizeof( D3D11_MAPPED_SUBRESOURCE ) );
        HRESULT hr = rendererD3D11State.immContext->Map( m->pointLightBufferCenterAndRadius, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource );

        if (hr == S_OK)
        {
            std::memcpy( mappedResource.pData, &m->pointLightCenterAndRadius[ 0 ].x, m->numActivePointLights * sizeof( Vec4 ) );
        }

        rendererD3D11State.immContext->Unmap( m->pointLightBufferCenterAndRadius, 0 );
    }
}

void ForwardPlus::SetSpotLightPositionAndRadius( int bufferIndex, const Vec4& positionAndRadius )
{
    if (-1 < bufferIndex && bufferIndex < m->numActiveSpotLights)
    {
        Log::Print( "SetSpotLightPositionAndRadius: position: %f %f %f, radius: %f", positionAndRadius.x, positionAndRadius.y, positionAndRadius.z, positionAndRadius.w );
        m->spotLightCenterAndRadius[ bufferIndex ] = positionAndRadius;

        // TODO: Find a way to update a single item instead of the whole buffer.
        D3D11_MAPPED_SUBRESOURCE mappedResource;
        ZeroMemory( &mappedResource, sizeof( D3D11_MAPPED_SUBRESOURCE ) );
        const HRESULT hr = rendererD3D11State.immContext->Map( m->spotLightBufferCenterAndRadius, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource );

        if (hr == S_OK)
        {
            std::memcpy( mappedResource.pData, &m->spotLightCenterAndRadius[ 0 ].x, m->numActiveSpotLights * sizeof( Vec4 ) );
        }
        else
        {
            Log::Assert( false, "Could not update spotLightBufferCenterAndRadius" );
        }

        rendererD3D11State.immContext->Unmap( m->spotLightBufferCenterAndRadius, 0 );
    }
}

void ForwardPlus::SetSpotLightParams( int bufferIndex, const Vec3& direction, float coneAngleCosRad, float falloff )
{
    if (-1 < bufferIndex && bufferIndex < m->numActiveSpotLights)
    {
        //const SpotParams params = PackSpotParams( direction, coneAngleCosRad, falloff );
        m->spotLightParams[ bufferIndex ].x = direction.x;// params.lightDirX;
        m->spotLightParams[ bufferIndex ].y = direction.y;// params.lightDirY;
        m->spotLightParams[ bufferIndex ].z = coneAngleCosRad * (direction.z > 0 ? 1 : -1);// .cosineOfConeAngleAndLightDirZSign;
        m->spotLightParams[ bufferIndex ].w = falloff;// params.falloffRadius;

        D3D11_MAPPED_SUBRESOURCE mappedResource;
        ZeroMemory( &mappedResource, sizeof( D3D11_MAPPED_SUBRESOURCE ) );
        const HRESULT hr = rendererD3D11State.immContext->Map( m->spotLightBufferParams, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource );

        if (hr == S_OK)
        {
            std::memcpy( mappedResource.pData, &m->spotLightParams[ 0 ].x, m->numActiveSpotLights * sizeof( Vec4 ) );
        }

        rendererD3D11State.immContext->Unmap( m->spotLightBufferParams, 0 );
    }
}
