#ifndef RENDERER_D3D11_STATE_H
#define RENDERER_D3D11_STATE_H
#ifndef LOG_H
#    include "Log.hpp"
#endif
#ifndef TEXTURE_IMPL_H
#    include "TextureImpl.hpp"
#endif

#define AE3D_SAFE_RELEASE(x) if (x) { x->Release(); x = nullptr; }

struct ID3D11Device;
struct IDXGISwapChain;
struct ID3D11DeviceContext;
struct ID3D11RenderTargetView;

ID3D11SamplerState* CreateSamplerState( D3D11_FILTER filter, D3D11_TEXTURE_ADDRESS_MODE tiling, int maxAnisotropy );

struct RenderTextureData
{
    RenderTextureData()
        : rtView( nullptr )
        , srView( nullptr )
        , depthStencilBuffer( nullptr )
        , depthStencilView( nullptr )
    {
        cubeRTVs[ 0 ] = cubeRTVs[ 1 ] = cubeRTVs[ 2 ] = cubeRTVs[ 3 ] = cubeRTVs[ 4 ] = cubeRTVs[ 5 ] = nullptr;
    }

    RenderTextureData( const RenderTextureData& other )
    {
        *this = other;
    }

    RenderTextureData& operator=(const RenderTextureData& other)
    {
        rtView = other.rtView;
        srView = other.srView;
        depthStencilBuffer = other.depthStencilBuffer;
        depthStencilView = other.depthStencilView;
        depthSRV = other.depthSRV;
        cubeRTVs[ 0 ] = other.cubeRTVs[ 0 ];
        cubeRTVs[ 1 ] = other.cubeRTVs[ 1 ];
        cubeRTVs[ 2 ] = other.cubeRTVs[ 2 ];
        cubeRTVs[ 3 ] = other.cubeRTVs[ 3 ];
        cubeRTVs[ 4 ] = other.cubeRTVs[ 4 ];
        cubeRTVs[ 5 ] = other.cubeRTVs[ 5 ];

        return *this;
    }

    ID3D11RenderTargetView* rtView;
    ID3D11RenderTargetView* cubeRTVs[ 6 ];
    ID3D11ShaderResourceView* srView;
    ID3D11Texture2D* depthStencilBuffer;
    ID3D11DepthStencilView* depthStencilView;
    ID3D11ShaderResourceView* depthSRV = nullptr;
};

struct Samplers
{
    ID3D11SamplerState* Get( const TextureImpl::Sampler& sampler )
    {
        Log::Assert( wasInit, "Samplers are not initted." );

        if (sampler.filter == TextureImpl::Filter::Linear &&
            sampler.tiling == TextureImpl::Tiling::Clamp)
        {
            return samplerStateClampLinear;
        }
        else if (sampler.filter == TextureImpl::Filter::Point &&
            sampler.tiling == TextureImpl::Tiling::Clamp)
        {
            return samplerStateClampPoint;
        }
        else if (sampler.filter == TextureImpl::Filter::Linear &&
            sampler.tiling == TextureImpl::Tiling::Repeat)
        {
            return samplerStateRepeatLinear;
        }
        else if (sampler.filter == TextureImpl::Filter::Point &&
            sampler.tiling == TextureImpl::Tiling::Repeat)
        {
            return samplerStateRepeatPoint;
        }
        else if (sampler.filter == TextureImpl::Filter::Anisotropic8 &&
            sampler.tiling == TextureImpl::Tiling::Repeat)
        {
            return samplerStateRepeatAnisotropic8;
        }
        else if (sampler.filter == TextureImpl::Filter::Anisotropic16 &&
            sampler.tiling == TextureImpl::Tiling::Repeat)
        {
            return samplerStateRepeatAnisotropic16;
        }
        else if (sampler.filter == TextureImpl::Filter::Anisotropic8 &&
            sampler.tiling == TextureImpl::Tiling::Clamp)
        {
            return samplerStateClampAnisotropic8;
        }
        else if (sampler.filter == TextureImpl::Filter::Anisotropic16 &&
            sampler.tiling == TextureImpl::Tiling::Clamp)
        {
            return samplerStateClampAnisotropic16;
        }
        else
        {
            Log::Assert( false, "Texture is using an unimplemented sampler!" );
            return samplerStateRepeatLinear;
        }
    }

    void Deinit()
    {
        AE3D_SAFE_RELEASE( samplerStateClampLinear )
        AE3D_SAFE_RELEASE( samplerStateClampPoint )
        AE3D_SAFE_RELEASE( samplerStateClampAnisotropic8 )
        AE3D_SAFE_RELEASE( samplerStateClampAnisotropic16 )
        AE3D_SAFE_RELEASE( samplerStateRepeatLinear )
        AE3D_SAFE_RELEASE( samplerStateRepeatPoint )
        AE3D_SAFE_RELEASE( samplerStateRepeatAnisotropic8 )
        AE3D_SAFE_RELEASE( samplerStateRepeatAnisotropic16 )
    }

    void Init()
    {
        wasInit = true;

        samplerStateClampLinear = CreateSamplerState( D3D11_FILTER_MIN_MAG_MIP_LINEAR, D3D11_TEXTURE_ADDRESS_CLAMP, 1 );
        samplerStateClampPoint = CreateSamplerState( D3D11_FILTER_MIN_MAG_MIP_POINT, D3D11_TEXTURE_ADDRESS_CLAMP, 1 );
        samplerStateRepeatLinear = CreateSamplerState( D3D11_FILTER_MIN_MAG_MIP_LINEAR, D3D11_TEXTURE_ADDRESS_WRAP, 1 );
        samplerStateRepeatPoint = CreateSamplerState( D3D11_FILTER_MIN_MAG_MIP_POINT, D3D11_TEXTURE_ADDRESS_WRAP, 1 );
        samplerStateClampAnisotropic8 = CreateSamplerState( D3D11_FILTER_ANISOTROPIC, D3D11_TEXTURE_ADDRESS_CLAMP, 8 );
        samplerStateRepeatAnisotropic8 = CreateSamplerState( D3D11_FILTER_ANISOTROPIC, D3D11_TEXTURE_ADDRESS_WRAP, 8 );
        samplerStateClampAnisotropic16 = CreateSamplerState( D3D11_FILTER_ANISOTROPIC, D3D11_TEXTURE_ADDRESS_CLAMP, 16 );
        samplerStateRepeatAnisotropic16 = CreateSamplerState( D3D11_FILTER_ANISOTROPIC, D3D11_TEXTURE_ADDRESS_WRAP, 16 );
    }

    bool wasInit = false;
    ID3D11SamplerState* samplerStateClampLinear = nullptr;
    ID3D11SamplerState* samplerStateClampPoint = nullptr;
    ID3D11SamplerState* samplerStateClampAnisotropic8 = nullptr;
    ID3D11SamplerState* samplerStateClampAnisotropic16 = nullptr;
    ID3D11SamplerState* samplerStateRepeatLinear = nullptr;
    ID3D11SamplerState* samplerStateRepeatPoint = nullptr;
    ID3D11SamplerState* samplerStateRepeatAnisotropic8 = nullptr;
    ID3D11SamplerState* samplerStateRepeatAnisotropic16 = nullptr;
};

struct RendererD3D11State
{
    void Deinit()
    {
        samplers.Deinit();
        AE3D_SAFE_RELEASE( immContext )
        AE3D_SAFE_RELEASE( swapChain )
        AE3D_SAFE_RELEASE( backBufferRenderTargetView )
        AE3D_SAFE_RELEASE( lequalDepthView )
        AE3D_SAFE_RELEASE( depthStencilBuffer )
        //ID3D11Debug* d3dDebug = nullptr;
        //d3dDev->QueryInterface(__uuidof(ID3D11Debug), reinterpret_cast<void**>(&d3dDebug));
        AE3D_SAFE_RELEASE( d3dDev )
        //d3dDebug->ReportLiveDeviceObjects( D3D11_RLDO_DETAIL );
        //AE3D_SAFE_RELEASE( d3dDebug )
    }

    ID3D11Device* d3dDev = nullptr;
    ID3D11Debug* debugDevice = nullptr;
    ID3D11DeviceContext* immContext = nullptr;
    IDXGISwapChain* swapChain = nullptr;
    ID3D11RenderTargetView* backBufferRenderTargetView = nullptr;
    ID3D11RenderTargetView* currentRenderTargetView = nullptr;
    ID3D11DepthStencilView* currentDepthStencilView = nullptr;
    ID3D11DepthStencilView* lequalDepthView = nullptr;
    ID3D11Texture2D* depthStencilBuffer = nullptr;

    const TextureImpl* currentRenderTarget = nullptr; // Not owning pointer. If nullptr, uses backbuffer.
    Samplers samplers;
};

extern RendererD3D11State rendererD3D11State;

#endif
