#include "RendererImpl.hpp"
#define WIN32_LEAN_AND_MEAN
#define VC_EXTRALEAN
#include <windows.h>
#include <d3d11_1.h>
#include <map>
#include "FileWatcher.hpp"
#include "GfxDevice.hpp"
#include "MaterialImpl.hpp"
#include "ModelImpl.hpp"
#include "Log.hpp"
#include "Statistics.hpp"
#include "RendererD3D11State.hpp"

#define AE3D_SAFE_RELEASE(x) if (x) { x->Release(); x = nullptr; }

LRESULT CALLBACK WindowProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );
std::vector< Vec4 > CreateSSAOKernel( int size );
void CreateSkybox( VertexBuffer& outSkybox );
void CreateQuadBuffer( VertexBuffer& outQuadBuffer );
std::wstring stringToWstring( const std::string& s );

void SetDebugLabel( ID3D11DeviceChild* d3dObj, const std::string& name )
{
    (void)d3dObj;
    (void)name;
#if DEBUG
    d3dObj->SetPrivateData( WKPDID_D3DDebugObjectName, (UINT)name.length(), name.c_str() );
#endif
}

HWND hwnd;
RendererD3D11State rendererD3D11State;

namespace GfxDeviceD3D11
{
    extern ID3D11Device* d3dDev;
    extern ID3D11DeviceContext* context;
}

namespace Global
{
    ID3D11Query* queryDisjoint = nullptr;
    ID3D11Query* queryBeginFrame = nullptr;
    ID3D11Query* queryEndFrame = nullptr;
    ID3DUserDefinedAnnotation* userDefinedAnnotation = nullptr;

    extern std::map< std::string, TextureImpl > pathToCachedTexture;
    extern std::map< std::string, ID3D11Texture2D* > pathToCachedTexture2D;
    extern std::map< ID3D11ShaderResourceView*, RenderTextureData > textureToRT;
    extern int maxInstanceCount;
}

void SetDebugEventBegin( const std::string& name )
{
    if (Global::userDefinedAnnotation)
    {
        Global::userDefinedAnnotation->BeginEvent( stringToWstring( name ).c_str() );
    }
}

void SetDebugEventEnd()
{
    if (Global::userDefinedAnnotation)
    {
        Global::userDefinedAnnotation->EndEvent();
    }
}

void UploadModelMatrices( Matrix44* modelMatrices, int matrixCount, Shader& depthNormalShader )
{
    depthNormalShader.SetMatrices( "instanceModelMatrices", modelMatrices, matrixCount );
}

void ErrorCheck( const std::string& )
{
}

void RendererImpl::CreateTimerQueries()
{
    D3D11_QUERY_DESC queryDesc = { D3D11_QUERY_TIMESTAMP_DISJOINT, 0 };
    rendererD3D11State.d3dDev->CreateQuery( &queryDesc, &Global::queryDisjoint );

    queryDesc.Query = D3D11_QUERY_TIMESTAMP;
    rendererD3D11State.d3dDev->CreateQuery( &queryDesc, &Global::queryBeginFrame );

    queryDesc.Query = D3D11_QUERY_TIMESTAMP;
    rendererD3D11State.d3dDev->CreateQuery( &queryDesc, &Global::queryEndFrame );
}

ID3D11SamplerState* CreateSamplerState( D3D11_FILTER filter, D3D11_TEXTURE_ADDRESS_MODE tiling, int maxAnisotropy )
{
    D3D11_SAMPLER_DESC samplerDesc;
    ZeroMemory( &samplerDesc, sizeof( samplerDesc ) );
    samplerDesc.Filter = filter;
    samplerDesc.AddressU = tiling;
    samplerDesc.AddressV = tiling;
    samplerDesc.AddressW = tiling;
    samplerDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
    samplerDesc.MaxAnisotropy = maxAnisotropy;
    samplerDesc.MinLOD = 0;
    samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;

    ID3D11SamplerState* outSamplerState = nullptr;
    HRESULT hr = rendererD3D11State.d3dDev->CreateSamplerState( &samplerDesc, &outSamplerState );

    if (FAILED( hr ))
    {
        MessageBoxA( 0, "Failed to create sampler state!", "Error", MB_OK );
    }

    return outSamplerState;
}

static void CreateAppWindow( int width, int height, bool fullScreen )
{
    const HINSTANCE hInstance = GetModuleHandle( nullptr );
    Log::Assert( hInstance != nullptr, "GetModuleHandle() failed." );

    WNDCLASSEX wc;
    ZeroMemory( &wc, sizeof( WNDCLASSEX ) );

    wc.cbSize = sizeof( WNDCLASSEX );
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = WindowProc;
    wc.hInstance = hInstance;
    wc.hCursor = LoadCursor( nullptr, IDC_ARROW );
    wc.hbrBackground = (HBRUSH)COLOR_WINDOW;
    wc.lpszClassName = "WindowClass1";

    wc.hIcon = (HICON)LoadImage( nullptr,
        "glider.ico",
        IMAGE_ICON,
        32,
        32,
        LR_LOADFROMFILE );

    RegisterClassEx( &wc );

    hwnd = CreateWindowExA( fullScreen ? WS_EX_TOOLWINDOW | WS_EX_TOPMOST : 0,
        "WindowClass1",    // name of the window class
        "Window",   // title of the window
        fullScreen ? WS_POPUP : (WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU),    // window style
        CW_USEDEFAULT,    // x-position of the window
        CW_USEDEFAULT,    // y-position of the window
        width,    // width of the window
        height,    // height of the window
        nullptr,    // we have no parent window
        nullptr,    // we aren't using menus
        hInstance,    // application handle
        nullptr );    // used with multiple windows    
    ShowWindow( hwnd, SW_SHOW );
}

static void CreateDeviceAndSwapChain( int width, int height, int antialiasSamples )
{
    DXGI_SWAP_CHAIN_DESC sd;
    ZeroMemory( &sd, sizeof( sd ) );
    sd.BufferCount = 1;
    sd.BufferDesc.Width = width;
    sd.BufferDesc.Height = height;
    sd.BufferDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM_SRGB;
    sd.BufferDesc.RefreshRate.Numerator = 60;
    sd.BufferDesc.RefreshRate.Denominator = 1;
    sd.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
    sd.OutputWindow = hwnd;
    sd.SampleDesc.Count = antialiasSamples;
    sd.SampleDesc.Quality = 0;
    sd.SwapEffect = DXGI_SWAP_EFFECT_DISCARD;
    sd.Windowed = TRUE;

    UINT flags = 0;
#ifndef NDEBUG
    flags = D3D11_CREATE_DEVICE_DEBUG;
#endif

    const HRESULT hr = D3D11CreateDeviceAndSwapChain( nullptr, D3D_DRIVER_TYPE_HARDWARE, nullptr, flags, 0, 0,
        D3D11_SDK_VERSION, &sd, &rendererD3D11State.swapChain, &rendererD3D11State.d3dDev, nullptr, &rendererD3D11State.immContext );
    if (hr == E_FAIL)
    {
        MessageBoxA( 0, "Failed to create Direct3D 11 context!  Attempted to create a device with the debug layer enabled and the layer is not installed.", "Could not create D3D device and swap chain!", MB_OK );
    }
    else if (hr == E_INVALIDARG)
    {
        MessageBoxA( 0, "An invalid parameter was passed to the returning function.", "Could not create D3D device and swap chain!", MB_OK );
    }
    else if (hr == E_OUTOFMEMORY)
    {
        MessageBoxA( 0, "Direct3D could not allocate sufficient memory to complete the call.", "Could not create D3D device and swap chain!", MB_OK );
    }
    else if (hr == D3D11_ERROR_TOO_MANY_UNIQUE_STATE_OBJECTS || hr == D3D11_ERROR_TOO_MANY_UNIQUE_VIEW_OBJECTS)
    {
        MessageBoxA( 0, "Too many state or view objects.", "Could not create D3D device and swap chain!", MB_OK );
    }
    else if (hr == DXGI_ERROR_INVALID_CALL)
    {
        MessageBoxA( 0, "Invalid parameters.", "Could not create D3D device and swap chain!", MB_OK );
    }
    else if (hr == DXGI_ERROR_UNSUPPORTED)
    {
        MessageBoxA( 0, "The requested functionality is not supported by the device or the driver.", "Could not create D3D device and swap chain!", MB_OK );
    }

    Log::Assert( rendererD3D11State.d3dDev != nullptr, "Could not create D3D device." );
    Log::Assert( rendererD3D11State.immContext != nullptr, "Could not create immediate context." );
    Log::Assert( rendererD3D11State.swapChain != nullptr, "Could not create swapchain." );

#if DEBUG
    HRESULT dhr = rendererD3D11State.d3dDev->QueryInterface( IID_PPV_ARGS( &rendererD3D11State.debugDevice ) );

    if (dhr == S_OK)
    {
        ID3D11InfoQueue* infoQueue = nullptr;
        HRESULT ihr = rendererD3D11State.debugDevice->QueryInterface( IID_PPV_ARGS( &infoQueue ) );

        if (ihr == S_OK)
        {
            infoQueue->SetBreakOnSeverity( D3D11_MESSAGE_SEVERITY_ERROR, TRUE );
        }
    }
#endif
}

static void CreateBackBufferRenderTargetView( int width, int height )
{
    Log::Assert( width > 0 && height > 0, "Invalid width or height." );

    ID3D11Texture2D* backBuffer;

    if (FAILED( rendererD3D11State.swapChain->GetBuffer( 0, __uuidof(ID3D11Texture2D), (LPVOID*)&backBuffer ) ))
    {
        MessageBoxA( 0, "Failed to create back buffer!", "Error", MB_OK );
    }

    const HRESULT hrColor = rendererD3D11State.d3dDev->CreateRenderTargetView( backBuffer, nullptr, &rendererD3D11State.backBufferRenderTargetView );
    AE3D_SAFE_RELEASE( backBuffer )

    if (FAILED( hrColor ))
    {
        MessageBoxA( 0, "Failed to create render target view!", "Error", MB_OK );
    }
}

static void CreateDepthStencilViews( int width, int height, int antialiasSamples )
{
    AE3D_SAFE_RELEASE( rendererD3D11State.depthStencilBuffer );
    AE3D_SAFE_RELEASE( rendererD3D11State.lequalDepthView );

    D3D11_TEXTURE2D_DESC depthStencilDesc;
    ZeroMemory( &depthStencilDesc, sizeof( depthStencilDesc ) );
    depthStencilDesc.Width = width;
    depthStencilDesc.Height = height;
    depthStencilDesc.MipLevels = 1;
    depthStencilDesc.ArraySize = 1;
    depthStencilDesc.Format = DXGI_FORMAT_D24_UNORM_S8_UINT;
    depthStencilDesc.SampleDesc.Count = antialiasSamples;
    depthStencilDesc.SampleDesc.Quality = 0;
    depthStencilDesc.Usage = D3D11_USAGE_DEFAULT;
    depthStencilDesc.BindFlags = D3D11_BIND_DEPTH_STENCIL;
    depthStencilDesc.CPUAccessFlags = 0;
    depthStencilDesc.MiscFlags = 0;

    rendererD3D11State.d3dDev->CreateTexture2D( &depthStencilDesc, nullptr, &rendererD3D11State.depthStencilBuffer );
    SetDebugLabel( rendererD3D11State.depthStencilBuffer, "backbuffer depth-stencil texture" );

    rendererD3D11State.d3dDev->CreateDepthStencilView( rendererD3D11State.depthStencilBuffer, nullptr, &rendererD3D11State.lequalDepthView );
}

Error::Enum RendererImpl::OpenWindow( int width, int height, bool fullScreen, int aAntialiasSamples, bool )
{
    if (width < 0 || height < 0)
    {
        Log::Print( "OpenWindow: Dimension must be a positive value." );
        return Error::CONTRACT_VIOLATION;
    }

    screenSize[ 0 ] = width == 0 ? GetSystemMetrics( SM_CXSCREEN ) : width;
    screenSize[ 1 ] = height == 0 ? GetSystemMetrics( SM_CYSCREEN ) : height;
    antialiasSamples = aAntialiasSamples;
    CreateAppWindow( screenSize[ 0 ], screenSize[ 1 ], fullScreen );
    CreateDeviceAndSwapChain( screenSize[ 0 ], screenSize[ 1 ], antialiasSamples );
    CreateBackBufferRenderTargetView( screenSize[ 0 ], screenSize[ 1 ] );
    CreateDepthStencilViews( screenSize[ 0 ], screenSize[ 1 ], antialiasSamples );
    CreateTimerQueries();
    rendererD3D11State.samplers.Init();

    CreateSkybox( skyboxBuffer );
    CreateQuadBuffer( quadBuffer );
    GfxDevice::Instance().SetRenderTarget( nullptr );

    const int bloomBlurSize = 512;

    builtinTargets.depthNormalTarget.CreateRenderTexture2D( Width(), Height(), TextureImpl::Format::Float, TextureImpl::Filter::Point, TextureImpl::Tiling::Clamp, "builtin depth-normal rt" );
    builtinTargets.hdrTarget.CreateRenderTexture2D( Width(), Height(), TextureImpl::Format::HalfFloat, TextureImpl::Filter::Point, TextureImpl::Tiling::Clamp, "builtin hdr rt" );
    builtinTargets.bloomBlurHTarget.CreateRenderTexture2D( bloomBlurSize, bloomBlurSize, TextureImpl::Format::UnsignedByte, TextureImpl::Filter::Linear, TextureImpl::Tiling::Clamp, "builtin bloom blur h rt" );
    builtinTargets.bloomBlurVTarget.CreateRenderTexture2D( bloomBlurSize, bloomBlurSize, TextureImpl::Format::UnsignedByte, TextureImpl::Filter::Linear, TextureImpl::Tiling::Clamp, "builtin bloom blur v rt" );
    builtinTargets.ssaoTarget.CreateRenderTexture2D( Width(), Height(), TextureImpl::Format::UnsignedByte, TextureImpl::Filter::Point, TextureImpl::Tiling::Clamp, "builtin ssao rt" );
    builtinTargets.ssaoBlurTarget.CreateRenderTexture2D( Width() / 2, Height() / 2, TextureImpl::Format::UnsignedByte, TextureImpl::Filter::Linear, TextureImpl::Tiling::Clamp, "builtin ssao blur rt" );
    builtinTargets.shadowBlurHTarget.CreateRenderTexture2D( 512, 512, TextureImpl::Format::Float, TextureImpl::Filter::Linear, TextureImpl::Tiling::Clamp, "builtin shadow blur h rt" );
    builtinTargets.shadowBlurVTarget.CreateRenderTexture2D( 512, 512, TextureImpl::Format::Float, TextureImpl::Filter::Linear, TextureImpl::Tiling::Clamp, "builtin shadow blur v rt" );

    const float black[] = { 0, 0, 0, 0 };
    rendererD3D11State.immContext->ClearRenderTargetView( builtinTargets.bloomBlurHTarget.RTV(), black );
    rendererD3D11State.immContext->ClearRenderTargetView( builtinTargets.bloomBlurVTarget.RTV(), black );

    const Error::Enum assetLoadStatus = LoadBuiltinAssets();

    if (assetLoadStatus == Error::NO_ERRORS)
    {
        const int kernelSize = 16;
        std::vector< Vec4 > kernel = CreateSSAOKernel( kernelSize );
        builtinShaders.ssaoShader.Use();
        builtinShaders.ssaoShader.SetInt( "uKernelSize", kernelSize );
        builtinShaders.ssaoShader.SetVectors( "uKernelOffsets", &kernel[ 0 ], (int)kernel.size() );
    }

    GfxDeviceD3D11::d3dDev = rendererD3D11State.d3dDev;
    GfxDeviceD3D11::context = rendererD3D11State.immContext;
    GfxDevice::Instance().Init();

    forwardPlus.Init();

#if DEBUG
    rendererD3D11State.immContext->QueryInterface( IID_PPV_ARGS( &Global::userDefinedAnnotation ) );
#endif
    return assetLoadStatus;
}

Error::Enum RendererImpl::LoadBuiltinAssets()
{
    const std::string fontPath( builtinAssetPath + "textures/font.png" );
    const std::string metaPath( builtinAssetPath + "textures/font.fnt" );

    Error::Enum error = fonts[ "aether3d_font" ].Load( fontPath.c_str(), metaPath.c_str() );

    AddShaderSearchPath( "../Assets/builtin/shaders/hlsl5/" );

    //double time = GetTime();

    builtinShaders.unlitShader.Load( builtinAssetPath + "shaders/hlsl5/unlit.hlsl", builtinAssetPath + "shaders/hlsl5/unlit.hlsl", "", "", error );
    builtinShaders.depthNormalShader.Load( builtinAssetPath + "shaders/hlsl5/depth.hlsl", builtinAssetPath + "shaders/hlsl5/depth.hlsl", "", "", error );
    builtinShaders.depthNormalSkinShader.Load( builtinAssetPath + "shaders/hlsl5/depth.hlsl", builtinAssetPath + "shaders/hlsl5/depth.hlsl", "", "SKIN", error );
    builtinShaders.skyboxShader.Load( builtinAssetPath + "shaders/hlsl5/cubemap.hlsl", builtinAssetPath + "shaders/hlsl5/cubemap.hlsl", "", "", error );
    builtinShaders.blurShader.Load( builtinAssetPath + "shaders/hlsl5/blur.hlsl", builtinAssetPath + "shaders/hlsl5/blur.hlsl", "", "", error );
    builtinShaders.blurSSAOShader.Load( builtinAssetPath + "shaders/hlsl5/blur_ssao.hlsl", builtinAssetPath + "shaders/hlsl5/blur_ssao.hlsl", "", "", error );
    builtinShaders.toneMapShader.Load( builtinAssetPath + "shaders/hlsl5/tonemap.hlsl", builtinAssetPath + "shaders/hlsl5/tonemap.hlsl", "", "", error );
    builtinShaders.ssaoShader.Load( builtinAssetPath + "shaders/hlsl5/ssao.hlsl", builtinAssetPath + "shaders/hlsl5/ssao.hlsl", "", "", error );
    builtinShaders.momentsShader.Load( builtinAssetPath + "shaders/hlsl5/moments.hlsl", builtinAssetPath + "shaders/hlsl5/moments.hlsl", "", "", error );
    builtinShaders.momentsSkinShader.Load( builtinAssetPath + "shaders/hlsl5/moments.hlsl", builtinAssetPath + "shaders/hlsl5/moments.hlsl", "", "SKIN", error );
    builtinShaders.cullLightShader.Load( (builtinAssetPath + "shaders/hlsl5/cull_lights.hlsl").c_str(), error );
    //Log::Print( "Shader loading time: %.2f s.", (GetTime() - time) );

    MaterialImpl::Default().SetShaders( builtinAssetPath + "shaders/hlsl5/unlit.hlsl", builtinAssetPath + "shaders/hlsl5/unlit.hlsl", "", "" );

    noiseTexture.CreateNoise( 4 );
    whiteTexture.Load( builtinAssetPath + "textures/white.png", true, true );
    return error;
}

void RendererImpl::BeginQuery()
{
    rendererD3D11State.immContext->Begin( Global::queryDisjoint );
    rendererD3D11State.immContext->End( Global::queryBeginFrame );
}

double RendererImpl::EndQuery()
{
    rendererD3D11State.immContext->End( Global::queryEndFrame );
    rendererD3D11State.immContext->End( Global::queryDisjoint );

    while (rendererD3D11State.immContext->GetData( Global::queryDisjoint, nullptr, 0, 0 ) == S_FALSE)
    {
        Sleep( 1 );
    }

    bool ok = true;

    // Check whether timestamps were disjoint during the last frame
    D3D10_QUERY_DATA_TIMESTAMP_DISJOINT tsDisjoint;
    if (rendererD3D11State.immContext->GetData( Global::queryDisjoint, &tsDisjoint, sizeof( tsDisjoint ), 0 ) == S_OK)
    {
        if (tsDisjoint.Disjoint)
        {
            ok = false;
        }
    }

    double result = 0;

    if (ok)
    {
        UINT64 beginFrame, endFrame;
        rendererD3D11State.immContext->GetData( Global::queryBeginFrame, &beginFrame, sizeof( UINT64 ), 0 );
        rendererD3D11State.immContext->GetData( Global::queryEndFrame, &endFrame, sizeof( UINT64 ), 0 );
        result = double( endFrame - beginFrame ) / double( tsDisjoint.Frequency ) * 1000.0;
    }

    return result;
}

void RendererImpl::Deinit()
{
    forwardPlus.Deinit();
    AE3D_SAFE_RELEASE( Global::userDefinedAnnotation )

    ID3D11ShaderResourceView* whiteSRV = whiteTexture.SRV();
    ID3D11UnorderedAccessView* whiteUAV = whiteTexture.UAV();
    AE3D_SAFE_RELEASE( whiteSRV )
    AE3D_SAFE_RELEASE( whiteUAV )

    // Release textures.
    for (const auto& textureImpl : Global::pathToCachedTexture)
    {
        ID3D11ShaderResourceView* srv = textureImpl.second.SRV();
        AE3D_SAFE_RELEASE( srv )

        ID3D11UnorderedAccessView* uav = textureImpl.second.UAV();
        AE3D_SAFE_RELEASE( uav )
    }

    for (const auto& texture2D : Global::pathToCachedTexture2D)
    {
        ID3D11Texture2D* texture = texture2D.second;
        AE3D_SAFE_RELEASE( texture )
    }

    for (auto& rtTexture : Global::textureToRT)
    {
        if (rtTexture.second.cubeRTVs[ 0 ] == rtTexture.second.rtView)
        {
            for (int i = 0; i < 6; ++i)
            {
                AE3D_SAFE_RELEASE( rtTexture.second.cubeRTVs[ i ] );
            }
        }
        else
        {
            AE3D_SAFE_RELEASE( rtTexture.second.rtView );
        }

        AE3D_SAFE_RELEASE( rtTexture.second.depthStencilBuffer );
        AE3D_SAFE_RELEASE( rtTexture.second.depthStencilView );
        AE3D_SAFE_RELEASE( rtTexture.second.depthSRV );
    }

    rendererD3D11State.immContext->VSSetShader( nullptr, nullptr, 0 );
    rendererD3D11State.immContext->PSSetShader( nullptr, nullptr, 0 );
    rendererD3D11State.immContext->GSSetShader( nullptr, nullptr, 0 );
    rendererD3D11State.immContext->CSSetShader( nullptr, nullptr, 0 );
    rendererD3D11State.immContext->IASetInputLayout( nullptr );

    Shader::Deinit();
    ComputeShader::Deinit();
    ModelImpl::Deinit();
    GfxDevice::Instance().Deinit();
    VertexBuffer::Deinit();

    AE3D_SAFE_RELEASE( Global::queryBeginFrame )
    AE3D_SAFE_RELEASE( Global::queryDisjoint )
    AE3D_SAFE_RELEASE( Global::queryEndFrame )

    rendererD3D11State.Deinit();
}

void RendererImpl::SetWindowTitle( const char* title ) const
{
    SetWindowTextA( hwnd, title );
}

void RendererImpl::SwapBuffers()
{
    Log::Assert( rendererD3D11State.swapChain != nullptr, "Swap chain not created." );

    MSG msg;

    if (PeekMessage( &msg, nullptr, 0, 0, PM_REMOVE ))
    {
        TranslateMessage( &msg );
        DispatchMessage( &msg );
    }

    Statistics::EndFrame();

    static int frame = 0;
    ++frame;
    if ((frame % 60) == 0)
    {
        FileWatcher::Instance().Poll();
    }

    if (rendererD3D11State.swapChain != nullptr)
    {
        rendererD3D11State.swapChain->Present( 1, 0 );
    }
}

void RendererImpl::Resize( int width, int height )
{
    // If window was minimized, do nothing.
    if (width == 0 && height == 0)
    {
        return;
    }

    if (rendererD3D11State.immContext == nullptr)
    {
        return;
    }

    screenSize[ 0 ] = width;
    screenSize[ 1 ] = height;

    rendererD3D11State.immContext->ClearState();
    AE3D_SAFE_RELEASE( rendererD3D11State.backBufferRenderTargetView );
    rendererD3D11State.immContext->OMSetRenderTargets( 0, 0, 0 );

    HRESULT hr = rendererD3D11State.swapChain->ResizeBuffers( 0, 0, 0, DXGI_FORMAT_UNKNOWN, 0 );

    if (hr != S_OK)
    {
        Log::Assert( false, "ResizeWindow: ResizeBuffers() failed." );
    }

    /*ID3D11Texture2D* pBackBuffer = nullptr;
    m->swapChain->GetBuffer( 0, __uuidof( ID3D11Texture2D ), (LPVOID*)&pBackBuffer );
    ID3D11RenderTargetView* rtView = nullptr;

    hr = m->d3dDev->CreateRenderTargetView( pBackBuffer, nullptr, &rtView );

    if (hr != S_OK)
    {
    assert( !"CreateRenderTargetView() failed." );
    }

    pBackBuffer->Release();
    AE3D_SAFE_RELEASE( rtView );
    m->renderTargetView = rtView;
    */
    CreateDepthStencilViews( width, height, antialiasSamples );
    CreateBackBufferRenderTargetView( width, height );
}
