#include "VertexBuffer.hpp"
#include <vector>
#include <d3d11.h>
#include "Log.hpp"
#include "Matrix.hpp"
#include "Mesh.hpp"
#include "RendererD3D11State.hpp"
#include "Statistics.hpp"

#define AE3D_SAFE_RELEASE(x) if (x) { x->Release(); x = nullptr; }

namespace Global
{
    // Used to keep track of resources to release on application exit.
    struct BufferObject
    {
        ID3D11Buffer* vertexBuffer = nullptr;
        ID3D11Buffer* indexBuffer = nullptr;
    };

    static std::vector< BufferObject > sBufferObjects;
}

// Vertex and Influence combined.
struct VertexWithInfluence
{
    Vec3 position;
    float u, v;
    Vec3 normal;
    // w is the handedness.
    Vec4 tangent;
    Vec4 color;
    float jointIndices[ 4 ];
    float weights[ 4 ];
};

struct VertexWith2Morphs
{
    Vec3 position;
    float u, v;
    Vec3 normal;
    // w is the handedness.
    Vec4 tangent;
    Vec4 color;
    Vec3 morphPosition1;
    Vec3 morphPosition2;
};

struct VertexBuffer::Impl
{
    void CreateIndexBuffer( const std::vector< Face >& indices );

    ID3D11Buffer* vertexBuffer = nullptr;
    ID3D11Buffer* indexBuffer = nullptr;
    int elementCount = 0;
    UINT stride = sizeof( Vertex );
};

void VertexBuffer::Impl::CreateIndexBuffer( const std::vector< Face >& indices )
{
    D3D11_BUFFER_DESC indexBufferDesc;
    ZeroMemory( &indexBufferDesc, sizeof( indexBufferDesc ) );
    indexBufferDesc.Usage = D3D11_USAGE_DEFAULT;
	indexBufferDesc.ByteWidth = (unsigned)(sizeof(unsigned short) * indices.size() * 3);
	indexBufferDesc.BindFlags = D3D11_BIND_INDEX_BUFFER;
	indexBufferDesc.CPUAccessFlags = 0;
	indexBufferDesc.MiscFlags = 0;
	indexBufferDesc.StructureByteStride = 0;

    D3D11_SUBRESOURCE_DATA indexData; 
    ZeroMemory( &indexData, sizeof( indexData ) );
    indexData.pSysMem = (D3D11_SUBRESOURCE_DATA*)(&indices[0]);
    indexData.SysMemPitch = 0;
	indexData.SysMemSlicePitch = 0;

    if (FAILED( rendererD3D11State.d3dDev->CreateBuffer( &indexBufferDesc, &indexData, &indexBuffer ) ))
    {
        MessageBoxA( 0, "Failed to create index buffer!", "CreateIndexBuffer Error", MB_OK );
    }
#if DEBUG
    static int i = 0;
    const std::string bufferName = "index_buffer_" + std::to_string( i );
    ++i;
    indexBuffer->SetPrivateData( WKPDID_D3DDebugObjectName, (UINT)bufferName.length(), bufferName.c_str() );
#endif
}

VertexBuffer::VertexBuffer() : m( new Impl ) {}

VertexBuffer::VertexBuffer( const VertexBuffer& other ) : m( new Impl )
{
    *this = other;
}

VertexBuffer::~VertexBuffer() {}

VertexBuffer& VertexBuffer::operator=( const VertexBuffer& other )
{
    *m = *other.m;
    return *this;
}

bool VertexBuffer::IsGenerated() const
{
    return m->vertexBuffer != nullptr;
}

void VertexBuffer::Bind() const
{
    const UINT offset = 0;
    rendererD3D11State.immContext->IASetIndexBuffer( m->indexBuffer, DXGI_FORMAT_R16_UINT, 0 );
    rendererD3D11State.immContext->IASetVertexBuffers( 0, 1, &m->vertexBuffer, &m->stride, &offset );
    rendererD3D11State.immContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST );
}

void VertexBuffer::Draw() const
{
    if (!m->vertexBuffer || !m->indexBuffer)
    {
        Log::Print( "Tried to draw an uninitialized vertex buffer." );
        return;
    }

#if DEBUG
    rendererD3D11State.debugDevice->ValidateContext( rendererD3D11State.immContext );
#endif
    rendererD3D11State.immContext->DrawIndexed( m->elementCount, 0, 0 );
    Statistics::IncDrawCalls();
}

void VertexBuffer::DrawInstanced( int instanceCount ) const
{
    if (!m->vertexBuffer || !m->indexBuffer)
    {
        Log::Print( "Tried to draw an uninitialized vertex buffer." );
        return;
    }

#if DEBUG
    rendererD3D11State.debugDevice->ValidateContext( rendererD3D11State.immContext );
#endif
    rendererD3D11State.immContext->DrawIndexedInstanced( m->elementCount, instanceCount, 0, 0, 0 );
    Statistics::IncDrawCalls();
}

void VertexBuffer::Deinit()
{
    for (auto buf : Global::sBufferObjects)
    {
        AE3D_SAFE_RELEASE( buf.vertexBuffer );
        AE3D_SAFE_RELEASE( buf.indexBuffer );
    }
}

void VertexBuffer::Generate( const std::vector< Face >& indices, const std::vector< Vertex >& vertices, const std::vector< Influence >& influences )
{
    m->elementCount = (int)indices.size() * 3;
    m->CreateIndexBuffer( indices );

    if (!influences.empty())
    {
        m->stride = sizeof( VertexWithInfluence );

        // Interleaves vertices with skinning data.
        
        std::vector< VertexWithInfluence > verticesWithInfluence( vertices.size() );

        for (std::size_t i = 0; i < verticesWithInfluence.size(); ++i)
        {
            verticesWithInfluence[ i ].position = vertices[ i ].position;    
            verticesWithInfluence[ i ].u = vertices[ i ].u;
            verticesWithInfluence[ i ].v = vertices[ i ].v;
            verticesWithInfluence[ i ].normal = vertices[ i ].normal;
            verticesWithInfluence[ i ].tangent = vertices[ i ].tangent;

            verticesWithInfluence[ i ].jointIndices[ 0 ] = influences[ i ].jointIndices[ 0 ];
            verticesWithInfluence[ i ].jointIndices[ 1 ] = influences[ i ].jointIndices[ 1 ];
            verticesWithInfluence[ i ].jointIndices[ 2 ] = influences[ i ].jointIndices[ 2 ];
            verticesWithInfluence[ i ].jointIndices[ 3 ] = influences[ i ].jointIndices[ 3 ];

            verticesWithInfluence[ i ].weights[ 0 ] = influences[ i ].weights[ 0 ];
            verticesWithInfluence[ i ].weights[ 1 ] = influences[ i ].weights[ 1 ];
            verticesWithInfluence[ i ].weights[ 2 ] = influences[ i ].weights[ 2 ];
            verticesWithInfluence[ i ].weights[ 3 ] = influences[ i ].weights[ 3 ];
        }

        D3D11_BUFFER_DESC boneBufferDesc;
        ZeroMemory( &boneBufferDesc, sizeof( boneBufferDesc ) );
        boneBufferDesc.Usage = D3D11_USAGE_DEFAULT;
        boneBufferDesc.ByteWidth = (unsigned)(sizeof( VertexWithInfluence ) * verticesWithInfluence.size());
        boneBufferDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
        boneBufferDesc.CPUAccessFlags = 0;
        boneBufferDesc.MiscFlags = 0;
        boneBufferDesc.StructureByteStride = 0;

        D3D11_SUBRESOURCE_DATA boneDesc; 
        ZeroMemory( &boneDesc, sizeof( boneDesc ) );
        boneDesc.pSysMem = (D3D11_SUBRESOURCE_DATA*)(&verticesWithInfluence[0]);
        boneDesc.SysMemPitch = 0;
	    boneDesc.SysMemSlicePitch = 0;

        if (FAILED( rendererD3D11State.d3dDev->CreateBuffer( &boneBufferDesc, &boneDesc, &m->vertexBuffer ) ))
        {
            MessageBoxA( 0, "Failed to create vertex buffer!", "VertexBufferD3D11 Error", MB_OK );
        }    
    }
    else
    {
        m->stride = sizeof( Vertex );

        D3D11_BUFFER_DESC vertexBufferDesc;
        ZeroMemory( &vertexBufferDesc, sizeof( vertexBufferDesc ) );
        vertexBufferDesc.Usage = D3D11_USAGE_IMMUTABLE;
        vertexBufferDesc.ByteWidth = (unsigned)(sizeof( Vertex ) * vertices.size());
        vertexBufferDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
        vertexBufferDesc.CPUAccessFlags = 0;
        vertexBufferDesc.MiscFlags = 0;
        vertexBufferDesc.StructureByteStride = 0;

        D3D11_SUBRESOURCE_DATA vertexDesc; 
        ZeroMemory( &vertexDesc, sizeof( vertexDesc ) );
        vertexDesc.pSysMem = (D3D11_SUBRESOURCE_DATA*)(&vertices[0]);
        vertexDesc.SysMemPitch = 0;
        vertexDesc.SysMemSlicePitch = 0;

        if (FAILED( rendererD3D11State.d3dDev->CreateBuffer( &vertexBufferDesc, &vertexDesc, &m->vertexBuffer ) ))
        {
            MessageBoxA( 0, "Failed to create vertex buffer!", "VertexBufferD3D11 Error", MB_OK );
        }

#if DEBUG
        static int i = 0;
        const std::string bufferName = "vertex_buffer_" + std::to_string( i );
        ++i;
        m->vertexBuffer->SetPrivateData( WKPDID_D3DDebugObjectName, (UINT)bufferName.length(), bufferName.c_str() );
#endif
    }

    Global::BufferObject bo;
    bo.indexBuffer = m->indexBuffer;
    bo.vertexBuffer = m->vertexBuffer;
    Global::sBufferObjects.push_back( bo );
}

void VertexBuffer::Generate( const std::vector< Face >& indices, const std::vector< Vertex >& vertices, const std::vector< Vec3 >& morphPositions1,
                             const std::vector< Vec3 >& morphPositions2 )
{
    Log::Assert( vertices.size() == morphPositions1.size() && vertices.size() == morphPositions2.size(), "Morph vertex counts do not match mesh vertex count!" );

    m->elementCount = (int)indices.size() * 3;
    m->stride = sizeof( VertexWith2Morphs );
    m->CreateIndexBuffer( indices );

    std::vector< VertexWith2Morphs > verticesWith2Morphs( vertices.size() );

    for (std::size_t i = 0; i < verticesWith2Morphs.size(); ++i)
    {
        verticesWith2Morphs[ i ].position = vertices[ i ].position;        
        verticesWith2Morphs[ i ].u = vertices[ i ].u;
        verticesWith2Morphs[ i ].v = vertices[ i ].v;
        verticesWith2Morphs[ i ].normal = vertices[ i ].normal;
        verticesWith2Morphs[ i ].tangent = vertices[ i ].tangent;
        verticesWith2Morphs[ i ].morphPosition1 = morphPositions1[ i ];
        verticesWith2Morphs[ i ].morphPosition2 = morphPositions2[ i ];
    }

    D3D11_BUFFER_DESC morph2BufferDesc;
    ZeroMemory( &morph2BufferDesc, sizeof( morph2BufferDesc ) );
    morph2BufferDesc.Usage = D3D11_USAGE_DEFAULT;
    morph2BufferDesc.ByteWidth = (unsigned)(sizeof( VertexWith2Morphs ) * verticesWith2Morphs.size());
    morph2BufferDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
    morph2BufferDesc.CPUAccessFlags = 0;
    morph2BufferDesc.MiscFlags = 0;
    morph2BufferDesc.StructureByteStride = 0;

    D3D11_SUBRESOURCE_DATA morph2Desc; 
    ZeroMemory( &morph2Desc, sizeof( morph2Desc ) );
    morph2Desc.pSysMem = (D3D11_SUBRESOURCE_DATA*)(&verticesWith2Morphs[0]);
    morph2Desc.SysMemPitch = 0;
	morph2Desc.SysMemSlicePitch = 0;

    if (FAILED( rendererD3D11State.d3dDev->CreateBuffer( &morph2BufferDesc, &morph2Desc, &m->vertexBuffer ) ))
    {
        MessageBoxA( 0, "Failed to create vertex buffer!", "VertexBufferD3D11 Error", MB_OK );
    }

    Global::BufferObject bo;
    bo.indexBuffer = m->indexBuffer;
    bo.vertexBuffer = m->vertexBuffer;
    Global::sBufferObjects.push_back( bo );
}
