#define VC_EXTRALEAN
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#define GL_GLEXT_PROTOTYPES
#include <EGL/egl.h>
#include <GLES3/gl3.h>
#include <GLES2/gl2ext.h>
#include <GLES2/gl2extQCOM.h>
#include <cassert>
#include "Log.hpp"
#include "FileWatcher.hpp"
#include "RendererImpl.hpp"
#include "Statistics.hpp"
LRESULT CALLBACK WindowProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );

HWND hwnd;

namespace Global
{
    int width = 800;
    int height = 600;
    EGLDisplay  eglDisplay = EGL_NO_DISPLAY;
    EGLConfig   eglConfig = nullptr;
    EGLContext  eglContextGL = nullptr;
    EGLSurface  eglSurface = nullptr;
    UINT32      redChannelBits = 8;
    UINT32      greenChannelBits = 8;
    UINT32      blueChannelBits = 8;
    UINT32      alphaChannelBits = 8;
    UINT32      depthBufferBits = 24;
    UINT32      stencilBufferBits = 8;
    UINT32      multiSampleBuffers = 0;
    UINT32      samplesPerPixel = 0;
    bool        doubleBuffer = true;
    NativeWindowType nativeWindow;
    NativeDisplayType nativeDisplay;
}

static void EGLErrorCheck( const std::string& )
{
    // Commented out because Adreno SDK doesn't find eglGetError().
    /*EGLint error = eglGetError();

    if (error != EGL_SUCCESS)
    {
        Log::Print( "%s caused an EGL error: %d", info.c_str(), error );
    }*/
}

bool CreateAppWindow( int width, int height, int* outScreenSize )
{
    HINSTANCE hInstance = GetModuleHandle( nullptr );

    // Register the window class
    WNDCLASS wc;
    wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;  // Window style
    wc.lpfnWndProc = (WNDPROC)WindowProc;                  // WndProc message handler
    wc.cbClsExtra = 0;                                   // No extra window data
    wc.cbWndExtra = 0;                                   // No extra window data
    wc.hInstance = hInstance;                           // Instance
    wc.hIcon = 0;
    wc.hCursor = LoadCursor( NULL, IDC_ARROW );       // Cursor
    wc.hbrBackground = NULL;                                // No Background
    wc.lpszMenuName = NULL;                                // No menu
    wc.lpszClassName = "Hello OpenGL ES 3.0";               // Set the class name
    wc.hIcon = (HICON)LoadImage( nullptr,
        "glider.ico",
        IMAGE_ICON,
        32,
        32,
        LR_LOADFROMFILE );

    if (RegisterClass( &wc ) == FALSE)
    {
        assert( !"Could not register window class!" );
        return false;
    }

    outScreenSize[ 0 ] = width  == 0 ? GetSystemMetrics( SM_CXSCREEN ) : width;
    outScreenSize[ 1 ] = height == 0 ? GetSystemMetrics( SM_CYSCREEN ) : height;

    // Adjust the window size to fit our rectangle
    DWORD dwWindowStyle = WS_SYSMENU | WS_MINIMIZEBOX | WS_CAPTION | WS_BORDER |
        WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
    RECT rcWindow;
    SetRect( &rcWindow, 0, 0, outScreenSize[ 0 ], outScreenSize[ 1 ] );
    AdjustWindowRect( &rcWindow, dwWindowStyle, FALSE );

    const int xOff = width  == 0 ? 0 : 50;
    const int yOff = height == 0 ? 0 : 50;

    hwnd = CreateWindowExA(
        width == 0 ? WS_EX_TOOLWINDOW : (WS_EX_APPWINDOW | WS_EX_WINDOWEDGE),    // Extended style
        "Hello OpenGL ES 3.0",                 // Class
        "Hello OpenGL ES 3.0",                 // Title
        width == 0 ? WS_POPUP : dwWindowStyle, // Style
        xOff + rcWindow.left, yOff + rcWindow.top, // Position
        (rcWindow.right - rcWindow.left),      // Width
        (rcWindow.bottom - rcWindow.top),      // Height
        nullptr,                               // No parent window
        nullptr,                               // No menu
        hInstance,                             // Instance
        nullptr );                             // Creation parameter
        
    if (hwnd == nullptr)
    {
        assert( !"Failed to create window!" );
        return false;
    }

    Global::nativeWindow = (NativeWindowType)hwnd;
    Global::nativeDisplay = (NativeDisplayType)GetDC( hwnd );

    ShowWindow( hwnd, SW_SHOW );

    return true;
}

bool CreateContext()
{
    Global::eglDisplay = eglGetDisplay( Global::nativeDisplay );

    if (Global::eglDisplay == EGL_NO_DISPLAY)
    {
        assert( !"ERROR: eglGetDisplay() failed.\n" );
        return false;
    }

    EGLint nMajorVersion, nMinorVersion;
    if (eglInitialize( Global::eglDisplay, &nMajorVersion, &nMinorVersion ) == FALSE)
    {
        assert( !"ERROR: eglInitialize failed.\n" );
        return false;
    }

    EGLint nNumConfigs;
    if (eglGetConfigs( Global::eglDisplay, nullptr, 0, &nNumConfigs ) == FALSE)
    {
        assert( !"ERROR: eglGetConfigs failed.\n" );
        return false;
    }

    EGLConfig config;
    if (eglGetConfigs( Global::eglDisplay, &config, 1, &nNumConfigs ) == FALSE)
    {
        return false;
    }

    // Get the first config's Color buffer size
    EGLint nBuffSize;
    eglGetConfigAttrib( Global::eglDisplay, config, EGL_BUFFER_SIZE, &nBuffSize );

    // Build the attibute list
    EGLint ConfigAttribs[ ] =
    {
        EGL_RENDERABLE_TYPE, EGL_OPENGL_ES2_BIT,
        EGL_RED_SIZE, Global::redChannelBits,
        EGL_GREEN_SIZE, Global::greenChannelBits,
        EGL_BLUE_SIZE, Global::blueChannelBits,
        EGL_ALPHA_SIZE, Global::alphaChannelBits,
        EGL_DEPTH_SIZE, Global::depthBufferBits,
        EGL_STENCIL_SIZE, Global::stencilBufferBits,
        EGL_SAMPLE_BUFFERS, Global::multiSampleBuffers,
        EGL_SAMPLES, Global::samplesPerPixel,
        EGL_NONE
    };

    // Create space for the config list
    EGLConfig* pEglConfigList = new EGLConfig[ nNumConfigs ];

    // Choose config based on the requested attributes
    if (eglChooseConfig( Global::eglDisplay, ConfigAttribs, pEglConfigList, nNumConfigs, &nNumConfigs ) == FALSE)
    {
        assert( !"ERROR: eglChooseConfig failed.\n" );
        return false;
    }

    INT32 nConfig = 0;
    bool bFoundConfig = false;

    // Try each configuration to see if it matches our attribute list
    while (nConfig < nNumConfigs && !bFoundConfig)
    {
        EGLint nAttribValue = 0;
        UINT32 nAttrib = 0;

        // First check if this configuration supports our surface type
        if (eglGetConfigAttrib( Global::eglDisplay, pEglConfigList[ nConfig ], EGL_SURFACE_TYPE, &nAttribValue ) == EGL_FALSE)
        {
            assert( !"ERROR: eglGetConfigAttrib(1) failed.\n" );
        }

        bFoundConfig = (nAttribValue & EGL_WINDOW_BIT) != 0;

        // Then check the remaining requested attributes
        while (ConfigAttribs[ nAttrib ] != EGL_NONE && bFoundConfig)
        {
            if (eglGetConfigAttrib( Global::eglDisplay, pEglConfigList[ nConfig ], ConfigAttribs[ nAttrib ], &nAttribValue ) == EGL_FALSE)
            {
                assert( !"ERROR: eglGetConfigAttrib(2) failed.\n" );
            }

            if (ConfigAttribs[ nAttrib + 1 ] != nAttribValue)
            {
                bFoundConfig = false;
            }

            nAttrib += 2;
        }

        if (!bFoundConfig)
        {
            ++nConfig;
        }
    }

    if (!bFoundConfig)
    {
        // None of the configs match our requested attribute list
        //assert( !"ERROR: Can't find a matching EGL configuration.\n" );

        nConfig = 0;
        if (eglChooseConfig( Global::eglDisplay, ConfigAttribs, pEglConfigList, 1, &nNumConfigs ) == EGL_FALSE)
        {
            delete[ ] pEglConfigList;
            assert( !"ERROR: eglChooseConfig(2) failed.\n" );
            return false;
        }
    }

    // Create a window surface
    Global::eglSurface = eglCreateWindowSurface( Global::eglDisplay, pEglConfigList[ nConfig ], hwnd, nullptr );
    if (EGL_NO_SURFACE == Global::eglSurface)
    {
        delete[ ] pEglConfigList;
        assert( !"ERROR: eglCreateWindowSurface failed.\n" );
        return false;
    }

    EGLint ContextAttribs[ ] =
    {
        EGL_CONTEXT_CLIENT_VERSION, 3,
        EGL_NONE
    };

    Global::eglContextGL = eglCreateContext( Global::eglDisplay, pEglConfigList[ nConfig ], EGL_NO_CONTEXT, ContextAttribs );

    delete[ ] pEglConfigList;

    eglBindAPI( EGL_OPENGL_ES_API );
    eglMakeCurrent( Global::eglDisplay, Global::eglSurface, Global::eglSurface, Global::eglContextGL );
    eglSwapInterval( Global::eglDisplay, 1 );
    EGLErrorCheck( "eglMakeCurrent" );

    return true;
}

void EGLContextSwapBuffers()
{
    eglSwapBuffers( Global::eglDisplay, Global::eglSurface );
}

void EGLDeinit()
{
    eglMakeCurrent( Global::eglDisplay, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT );
    eglTerminate( Global::eglDisplay );
}

void RendererImpl::SwapBuffers()
{
    Statistics::EndFrame();

    static int frame = 0;
    ++frame;
    if ((frame % 120) == 0)
    {
        FileWatcher::Instance().Poll();
    }

    MSG msg;

    if (PeekMessage( &msg, nullptr, 0, 0, PM_REMOVE ))
    {
        TranslateMessage( &msg );
        DispatchMessage( &msg );
    }

    EGLContextSwapBuffers();
}
