#include "AudioImpl.hpp"
#include <cstdlib>
#include <cassert>
#include <fstream>
#include <map>
#include <string>
#include "Log.hpp"

#ifdef AETHER3D_WINDOWS
#define AETHER3D_API __declspec(dllexport)
#else
#define AETHER3D_API 
#endif

static_assert( sizeof( unsigned int   ) == 4, "" );
static_assert( sizeof( unsigned short ) == 2, "" );

struct Impl
{
    /**
     Loads a .wav file. Called by addSound().
     
     \param soundName Sound's name.
     \param fileName File name of .wav file.
     */
    void LoadWav( const std::string& soundName, const std::string& fileName );
    
    /**
     Loads a .ogg file. Called by addSound().
     
     \param soundName Sound's name.
     \param fileName File name of .ogg file.
     */
    void LoadOgg( const std::string& soundName, const std::string& fileName );
    
    bool muted;
};

namespace Global
{
    Impl impl;
}

/**
 * WAVE structure which is used by .wav files.
 * Follows the specification at the following URI:
 * <http://ccrma.stanford.edu/courses/422/projects/WaveFormat/>
 *
 * Assumes that unsigned int is 4 bytes and unsigned short is 2 bytes.
 */
struct WAVE
{
    unsigned char chunkID[ 4 ];
    unsigned int chunkSize;
    unsigned char format[4];

    // Format sub chunk
    unsigned char subchunk1ID[ 4 ];
    unsigned int subchunk1Size;
    unsigned short audioFormat;
    unsigned short numChannels;
    unsigned int sampleRate;
    unsigned int byteRate;
    unsigned short blockAlign;
    unsigned short bitsPerSample;

    // Data sub chunk
    unsigned char subchunk2ID[ 4 ]; ///< Data sub chunk ID.
    unsigned int subchunk2Size; ///< Data sub chunk size.
    unsigned char* data; ///< Chunk data.
};

extern "C" AETHER3D_API Audio* CreateAudio()
{
    return new AudioImpl();
}

AudioImpl::AudioImpl()
{
    Global::impl.muted = false;
}

AudioImpl::~AudioImpl()
{
}

void AudioImpl::Reload( const char* )
{

}

void AudioImpl::AddSound( const char* aSoundName, const char* aSoundFile )
{
    const std::string aSoundFileStr( aSoundFile );
    const std::string extension = aSoundFileStr.substr( aSoundFileStr.length() - 3, aSoundFileStr.length() );

    if (extension == "wav" || extension == "WAV")
    {
        Global::impl.LoadWav( aSoundName, aSoundFile );
    }
    else
    {
        assert( !"Unknown audio file extension!" );
        Log::Print( "Unknown audio file extension in file %s", aSoundFileStr.c_str() );
    }
}

void Impl::LoadWav( const std::string&, const std::string& fileName )
{
    const std::string systemFileName = fileName;
    std::ifstream ifs( systemFileName.c_str(), std::ios::binary );

    if (!ifs)
    {
        assert( !"Audio file not found!" );
        Log::Print( "%s not found!", systemFileName.c_str() );
        return;
    }

    WAVE wav;

    ifs.read( (char*)&wav.chunkID, 4 );
    if (wav.chunkID[0] != 'R' && wav.chunkID[1] != 'I' &&
        wav.chunkID[2] != 'F' && wav.chunkID[3] != 'F')
    {
        Log::Print( "%s is not a valid .wav file!", systemFileName.c_str() );
        assert( !"Audio file corrupted!" );
        return;
    }

    ifs.read( (char*)&wav.chunkSize, 4 );

    ifs.read( (char*)&wav.format, 4 );
    if (wav.format[0] != 'W' && wav.format[1] != 'A' &&
        wav.format[2] != 'V' && wav.format[3] != 'E')
    {
        Log::Print( "%s is not a valid .wav file!", systemFileName.c_str() );
        assert( !"Audio file corrupted!" );
        return;
    }
    
    ifs.read( (char*)&wav.subchunk1ID, 4 );
    if (wav.subchunk1ID[0] != 'f' && wav.subchunk1ID[1] != 'm' &&
        wav.subchunk1ID[2] != 't' && wav.subchunk1ID[3] != ' ')
    {
        Log::Print( "%s is not a valid .wav file!", systemFileName.c_str() );
        assert( !"Audio file corrupted!" );
        return;
    }
    
    ifs.read( (char*)&wav.subchunk1Size, 4 );
    if (wav.subchunk1Size != 16)
    {
        Log::Print( "%s does not contain PCM audio data!", systemFileName.c_str() );
        assert( !"Audio file doesn't contain PCM audio data!" );
        return;
    }
    
    ifs.read( (char*)&wav.audioFormat, 2 );
    if (wav.audioFormat != 1)
    {
        Log::Print( "%s does not contain uncompressed PCM audio data!", systemFileName.c_str() );
        assert( !"Audio file doesn't contain uncompressed PCM audio data!" );
        return;
    }
    
    ifs.read( (char*)&wav.numChannels,   2 );
    ifs.read( (char*)&wav.sampleRate,    4 );
    ifs.read( (char*)&wav.byteRate,      4 );
    ifs.read( (char*)&wav.blockAlign,    2 );
    ifs.read( (char*)&wav.bitsPerSample, 2 );
    
    ifs.read( (char*)&wav.subchunk2ID, 4 );
    if (wav.subchunk2ID[0] != 'd' && wav.subchunk2ID[1] != 'a' &&
        wav.subchunk2ID[2] != 't' && wav.subchunk2ID[3] != 'a')
    {
        Log::Print( "%s does not contain data chunk!", systemFileName.c_str() );
        return;
    }

    ifs.read( (char*)&wav.subchunk2Size, 4 );

    // Data size is file size minus header size.
    // We get it by saving the current position, moving
    // to the EOF, checking EOF's position and moving back to
    // the saved location for data reading.
    std::streampos dataPos = ifs.tellg();
    ifs.seekg( 0, std::ios::end );
    std::streampos fileSize = ifs.tellg();
    const std::size_t dataSize = (std::size_t)fileSize - 44; // Header is 44 bytes.

    ifs.seekg( dataPos );

    wav.data = new unsigned char[ dataSize ];

    ifs.read( (char*)wav.data, dataSize );

    if (wav.numChannels == 1 && wav.bitsPerSample == 8)
    {
    }
    else if (wav.numChannels == 2 && wav.bitsPerSample == 8)
    {
    }
    else if (wav.numChannels == 1 && wav.bitsPerSample == 16)
    {
    }
    else if (wav.numChannels == 2 && wav.bitsPerSample == 16)
    {
    }
    else
    {
        assert( !"Unknown format!" );
    }
    
    delete[] wav.data;
}

void AudioImpl::Play( const char* )
{
}

void AudioImpl::SetDopplerFactor( float )
{
}

void AudioImpl::SetDopplerVelocity( float )
{
}

void AudioImpl::SetGain( float )
{
}

void AudioImpl::SetMute( bool mute )
{
    Global::impl.muted = mute;
}

bool AudioImpl::IsMuted() const
{
    return Global::impl.muted;
}
