#include "ForwardPlus.hpp"
#include <string>
#include <vector>
#include "GL/glxw.h"
#include "ComputeShader.hpp"
#include "MathUtil.hpp"
#include "RendererTypes.hpp"
#include "Shader.hpp"
#include "TextureImpl.hpp"

void ErrorCheck( const std::string& info );

SpotParams PackSpotParams( const Vec3& vLightDir, float cosineOfConeAngle, float falloffRadius )
{
    //assert( cosineOfConeAngle > 0.0f );
    //assert( falloffRadius > 0.0f );

    SpotParams PackedParams;
    PackedParams.lightDirX = MathUtil::ConvertF32ToF16( vLightDir.x );
    PackedParams.lightDirY = MathUtil::ConvertF32ToF16( vLightDir.y );
    PackedParams.cosineOfConeAngleAndLightDirZSign = MathUtil::ConvertF32ToF16( cosineOfConeAngle );
    PackedParams.falloffRadius = MathUtil::ConvertF32ToF16( falloffRadius );

    // put the sign bit for light dir z in the sign bit for the cone angle
    // (we can do this because we know the cone angle is always positive)
    if (vLightDir.z < 0.0f)
    {
        PackedParams.cosineOfConeAngleAndLightDirZSign |= 0x8000;
    }
    else
    {
        PackedParams.cosineOfConeAngleAndLightDirZSign &= 0x7FFF;
    }

    return PackedParams;
}

struct ForwardPlus::Impl
{
    void InitPointLights();
    void InitSpotLights();

    std::vector< Vec4 > pointLightCenterAndRadius;
    std::vector< Vec4 > pointLightColors;
    int numActivePointLights = 0;
    std::vector< Vec4 > spotLightCenterAndRadius;
    std::vector< Vec4 > spotLightColors;
    std::vector< Vec4 > spotLightParams;
    int numActiveSpotLights = 0;
    static const int MAX_NUM_LIGHTS = 2048;
    static const unsigned MAX_NUM_LIGHTS_PER_TILE = 544;

    GLuint pointLightCenterAndRadiusTextureBufferId = 0;
    GLuint pointLightCenterAndRadiusTextureId = 0;
    GLuint pointLightColorTextureBufferId = 0;
    GLuint pointLightColorTextureId = 0;

    GLuint spotLightCenterAndRadiusTextureBufferId = 0;
    GLuint spotLightCenterAndRadiusTextureId = 0;
    GLuint spotLightColorTextureBufferId = 0;
    GLuint spotLightColorTextureId = 0;
    GLuint spotLightParamsTextureBufferId = 0;
    GLuint spotLightParamsTextureId = 0;

    GLuint lightIndexTextureBufferId = 0;
    GLuint lightIndexTextureId = 0;
};

void ForwardPlus::Impl::InitPointLights()
{
    pointLightCenterAndRadius.resize( MAX_NUM_LIGHTS );
    pointLightColors.resize( MAX_NUM_LIGHTS );

    // Point light positions.
    glGenBuffers( 1, &pointLightCenterAndRadiusTextureBufferId );
    glBindBuffer( GL_TEXTURE_BUFFER, pointLightCenterAndRadiusTextureBufferId );
    glBufferData( GL_TEXTURE_BUFFER, sizeof( Vec4 ) * pointLightCenterAndRadius.size(), pointLightCenterAndRadius.data(), GL_DYNAMIC_DRAW );

    glGenTextures( 1, &pointLightCenterAndRadiusTextureId );
    glBindTexture( GL_TEXTURE_BUFFER, pointLightCenterAndRadiusTextureId );
    glTexBuffer( GL_TEXTURE_BUFFER, GL_RGBA32F, pointLightCenterAndRadiusTextureBufferId );

    // Point light colors.
    glGenBuffers( 1, &pointLightColorTextureBufferId );
    glBindBuffer( GL_TEXTURE_BUFFER, pointLightColorTextureBufferId );
    glBufferData( GL_TEXTURE_BUFFER, sizeof( Vec4 ) * pointLightColors.size(), pointLightColors.data(), GL_DYNAMIC_DRAW );

    glGenTextures( 1, &pointLightColorTextureId );
    glBindTexture( GL_TEXTURE_BUFFER, pointLightColorTextureId );
    glTexBuffer( GL_TEXTURE_BUFFER, GL_RGBA32F, pointLightColorTextureBufferId );
}

void ForwardPlus::Impl::InitSpotLights()
{
    // Spot lights.
    spotLightCenterAndRadius.resize( MAX_NUM_LIGHTS );
    spotLightColors.resize( MAX_NUM_LIGHTS );
    spotLightParams.resize( MAX_NUM_LIGHTS );

    // Spot light positions.
    glGenBuffers( 1, &spotLightCenterAndRadiusTextureBufferId );
    glBindBuffer( GL_TEXTURE_BUFFER, spotLightCenterAndRadiusTextureBufferId );
    glBufferData( GL_TEXTURE_BUFFER, sizeof( Vec4 ) * spotLightCenterAndRadius.size(), spotLightCenterAndRadius.data(), GL_DYNAMIC_DRAW );

    glGenTextures( 1, &spotLightCenterAndRadiusTextureId );
    glBindTexture( GL_TEXTURE_BUFFER, spotLightCenterAndRadiusTextureId );
    glTexBuffer( GL_TEXTURE_BUFFER, GL_RGBA32F, spotLightCenterAndRadiusTextureBufferId );

    // Spot light colors.
    glGenBuffers( 1, &spotLightColorTextureBufferId );
    glBindBuffer( GL_TEXTURE_BUFFER, spotLightColorTextureBufferId );
    glBufferData( GL_TEXTURE_BUFFER, sizeof( Vec4 ) * spotLightColors.size(), spotLightColors.data(), GL_DYNAMIC_DRAW );

    glGenTextures( 1, &spotLightColorTextureId );
    glBindTexture( GL_TEXTURE_BUFFER, spotLightColorTextureId );
    glTexBuffer( GL_TEXTURE_BUFFER, GL_RGBA32F, spotLightColorTextureBufferId );

    // Spot light params.
    glGenBuffers( 1, &spotLightParamsTextureBufferId );
    glBindBuffer( GL_TEXTURE_BUFFER, spotLightParamsTextureBufferId );
    glBufferData( GL_TEXTURE_BUFFER, sizeof( Vec4 ) * spotLightParams.size(), spotLightParams.data(), GL_DYNAMIC_DRAW );

    glGenTextures( 1, &spotLightParamsTextureId );
    glBindTexture( GL_TEXTURE_BUFFER, spotLightParamsTextureId );
    glTexBuffer( GL_TEXTURE_BUFFER, GL_RGBA32F, spotLightParamsTextureBufferId );
}

ForwardPlus::ForwardPlus() : m( new Impl )
{
}

ForwardPlus::~ForwardPlus()
{
}

void ForwardPlus::Init()
{
    ErrorCheck( "Forward+ Init begin" );

    m->InitPointLights();
    m->InitSpotLights();

    // Light indices.
    const unsigned numTiles = GetNumTilesX() * GetNumTilesY();
    const unsigned maxNumLightsPerTile = GetMaxNumLightsPerTile();

    glGenBuffers( 1, &m->lightIndexTextureBufferId );
    glBindBuffer( GL_TEXTURE_BUFFER, m->lightIndexTextureBufferId );
    glBufferData( GL_TEXTURE_BUFFER, 4 * maxNumLightsPerTile * numTiles, nullptr, GL_DYNAMIC_DRAW );

    glGenTextures( 1, &m->lightIndexTextureId );
    glBindTexture( GL_TEXTURE_BUFFER, m->lightIndexTextureId );
    glTexBuffer( GL_TEXTURE_BUFFER, GL_R32UI, m->lightIndexTextureBufferId );

    ErrorCheck( "Forward+ Init end" );
}

void ForwardPlus::Deinit()
{
    glDeleteBuffers( 1, &m->pointLightCenterAndRadiusTextureBufferId );
    glDeleteTextures( 1, &m->pointLightCenterAndRadiusTextureId );

    glDeleteBuffers( 1, &m->pointLightColorTextureBufferId );
    glDeleteTextures( 1, &m->pointLightColorTextureId );

    glDeleteBuffers( 1, &m->spotLightCenterAndRadiusTextureBufferId );
    glDeleteTextures( 1, &m->spotLightCenterAndRadiusTextureId );

    glDeleteBuffers( 1, &m->spotLightColorTextureBufferId );
    glDeleteTextures( 1, &m->spotLightColorTextureId );

    glDeleteBuffers( 1, &m->lightIndexTextureBufferId );
    glDeleteTextures( 1, &m->lightIndexTextureId );
}

int ForwardPlus::GetNextPointLightBufferIndex()
{
    if (m->numActivePointLights + m->numActiveSpotLights < m->MAX_NUM_LIGHTS)
    {
        ++m->numActivePointLights;
        return m->numActivePointLights - 1;
    }

    // Error handling is done in the caller.
    return -1;
}

int ForwardPlus::GetNextSpotLightBufferIndex()
{
    if (m->numActiveSpotLights + m->numActiveSpotLights + 1 < m->MAX_NUM_LIGHTS)
    {
        ++m->numActiveSpotLights;
        return m->numActiveSpotLights - 1;
    }

    // Error handling is done in the caller.
    return -1;
}

unsigned ForwardPlus::GetMaxNumLightsPerTile() const
{
    const unsigned adjustmentMultipier = 32;

    // I haven't tested at greater than 1080p, so cap it
    const unsigned height = (Screen::Height() > 1080) ? 1080 : Screen::Height();

    // adjust max lights per tile down as height increases
    return (m->MAX_NUM_LIGHTS_PER_TILE - (adjustmentMultipier * (height / 120)));
}

void ForwardPlus::CullLights( const RendererState& state, ComputeShader& cullShader, TextureImpl& depthNormalTarget )
{
    ErrorCheck( "Forward+ set uniforms begin" );

    cullShader.Use();
    Matrix44 invProj;
    Matrix44::Invert( state.projection, invProj );
    cullShader.SetMatrix( "invProjection", invProj );
    cullShader.SetMatrix( "viewMatrix", state.view );
    cullShader.SetUInt( "windowWidth", Screen::Width() );
    cullShader.SetUInt( "windowHeight", Screen::Height() );
    //cullShader.SetUInt( "numLights", static_cast< unsigned  >(m->pointLightCenterAndRadius.size()) & 0xFFFFu );
    cullShader.SetUInt( "numLights", (((unsigned)m->numActiveSpotLights & 0xFFFFu) << 16) | ((unsigned)m->numActivePointLights & 0xFFFFu));
    cullShader.SetInt( "maxNumLightsPerTile", GetMaxNumLightsPerTile() );
    cullShader.SetVector( "_CameraPosition", state.cameraPosition );

    ErrorCheck( "Forward+ set uniforms end" );

    glBindImageTexture( 0, m->pointLightCenterAndRadiusTextureId, 0, GL_FALSE, 0, GL_READ_ONLY, GL_RGBA32F );
    glBindImageTexture( 1, depthNormalTarget.Id(), 0, GL_FALSE, 0, GL_READ_ONLY, GL_RGBA32F );
    glBindImageTexture( 2, m->lightIndexTextureId, 0, GL_FALSE, 0, GL_READ_WRITE, GL_R32UI );
    glBindImageTexture( 4, m->spotLightCenterAndRadiusTextureId, 0, GL_FALSE, 0, GL_READ_ONLY, GL_RGBA32F );

    cullShader.Dispatch( GetNumTilesX(), GetNumTilesY(), 1 );
    glMemoryBarrier( GL_SHADER_IMAGE_ACCESS_BARRIER_BIT );

    // Unbind
    glBindImageTexture( 1, 0, 0, GL_FALSE, 0, GL_READ_WRITE, GL_R32UI );

    ErrorCheck( "Forward+ cull end" );
}

void ForwardPlus::SetUniformsForShading( Shader& shader )
{
    shader.SetUInt( "windowWidth", Screen::Width() );
    shader.SetUInt( "windowHeight", Screen::Height() );
    shader.SetUInt( "maxNumLightsPerTile", GetMaxNumLightsPerTile() );
    shader.SetInt( "_LightType", -1 );
    glBindImageTexture( 3, m->pointLightColorTextureId, 0, GL_FALSE, 0, GL_READ_ONLY, GL_RGBA32F );
    glBindImageTexture( 5, m->spotLightColorTextureId, 0, GL_FALSE, 0, GL_READ_ONLY, GL_RGBA32F );
    glBindImageTexture( 6, m->spotLightParamsTextureId, 0, GL_FALSE, 0, GL_READ_ONLY, GL_RGBA32F );
}

void ForwardPlus::SetPointLightColor( int bufferIndex, const Vec3& color )
{
    if (-1 < bufferIndex && bufferIndex < m->numActivePointLights)
    {
        m->pointLightColors[ bufferIndex ] = Vec4( color );
        glBindBuffer( GL_TEXTURE_BUFFER, m->pointLightColorTextureBufferId );
        glBufferSubData( GL_TEXTURE_BUFFER, bufferIndex * sizeof( Vec4 ), sizeof( Vec4 ), &m->pointLightColors[ bufferIndex ].x );
        glBindBuffer( GL_TEXTURE_BUFFER, 0 );
    }
}

void ForwardPlus::SetSpotLightColor( int bufferIndex, const Vec3& color )
{
    if (-1 < bufferIndex && bufferIndex < m->numActiveSpotLights)
    {
        m->spotLightColors[ bufferIndex ] = Vec4( color );
        glBindBuffer( GL_TEXTURE_BUFFER, m->spotLightColorTextureBufferId );
        glBufferSubData( GL_TEXTURE_BUFFER, bufferIndex * sizeof( Vec4 ), sizeof( Vec4 ), &m->spotLightColors[ bufferIndex ].x );
        glBindBuffer( GL_TEXTURE_BUFFER, 0 );
    }
}

void ForwardPlus::SetPointLightPositionAndRadius( int bufferIndex, const Vec4& positionAndRadius )
{
    if (-1 < bufferIndex && bufferIndex < m->numActivePointLights)
    {
        m->pointLightCenterAndRadius[ bufferIndex ] = positionAndRadius;
        glBindBuffer( GL_TEXTURE_BUFFER, m->pointLightCenterAndRadiusTextureBufferId );
        glBufferSubData( GL_TEXTURE_BUFFER, bufferIndex * sizeof( Vec4 ), sizeof( Vec4 ), &m->pointLightCenterAndRadius[ bufferIndex ].x );
        glBindBuffer( GL_TEXTURE_BUFFER, 0 );
    }
}

void ForwardPlus::SetSpotLightPositionAndRadius( int bufferIndex, const Vec4& positionAndRadius )
{
    if (-1 < bufferIndex && bufferIndex < m->numActiveSpotLights)
    {
        m->spotLightCenterAndRadius[ bufferIndex ] = positionAndRadius;
        glBindBuffer( GL_TEXTURE_BUFFER, m->spotLightCenterAndRadiusTextureBufferId );
        glBufferSubData( GL_TEXTURE_BUFFER, bufferIndex * sizeof( Vec4 ), sizeof( Vec4 ), &m->spotLightCenterAndRadius[ bufferIndex ].x );
        glBindBuffer( GL_TEXTURE_BUFFER, 0 );
    }
}

void ForwardPlus::SetSpotLightParams( int bufferIndex, const Vec3& direction, float coneAngleCosRad, float falloff )
{
    if (-1 < bufferIndex && bufferIndex < m->numActiveSpotLights)
    {
        //const SpotParams params = PackSpotParams( direction, coneAngleCosRad, falloff );
        m->spotLightParams[ bufferIndex ].x = direction.x;// params.lightDirX;
        m->spotLightParams[ bufferIndex ].y = direction.y;// params.lightDirY;
        m->spotLightParams[ bufferIndex ].z = coneAngleCosRad * (direction.z > 0 ? 1 : -1);// .cosineOfConeAngleAndLightDirZSign;
        m->spotLightParams[ bufferIndex ].w = falloff;// params.falloffRadius;

        glBindBuffer( GL_TEXTURE_BUFFER, m->spotLightParamsTextureBufferId );
        glBufferSubData( GL_TEXTURE_BUFFER, bufferIndex * sizeof( Vec4 ), sizeof( Vec4 ), &m->spotLightParams[ bufferIndex ].x );
        glBindBuffer( GL_TEXTURE_BUFFER, 0 );
    }
}
