#include "GfxDevice.hpp"
#if AETHER3D_IOS
#    include <OpenGLES/ES3/glext.h>
#elif AETHER3D_GLES3
#    include <GLES3/gl3.h>
#else
#    include "GL/glxw.h"
#endif
#include "Log.hpp"
#include "TextureImpl.hpp"
#include "Platform.hpp"
#include "Statistics.hpp"

void BindDrawable();
void ErrorCheckFBO();
void ErrorCheck( const std::string& info );

namespace Global
{
    extern GLuint systemFBO;
}

struct GfxDevice::Impl
{
    const TextureImpl* currentRenderTarget = nullptr;
};

GfxDevice::GfxDevice() : m( new Impl() )
{
}

GfxDevice::~GfxDevice()
{
}

GfxDevice& GfxDevice::Instance()
{
    static GfxDevice gfxDevice;
    return gfxDevice;
}

const TextureImpl* GfxDevice::CurrentRenderTarget()
{
    return m->currentRenderTarget;
}

void GfxDevice::ClearScreen()
{
    glClear( GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT );
}

void GfxDevice::SetBlendMode( BlendMode blendMode )
{
    if (blendMode == BlendMode::Off)
    {
        glDisable( GL_BLEND );
    }
    else if (blendMode == BlendMode::AlphaBlend)
    {
        glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
        glEnable( GL_BLEND );
    }
    else if (blendMode == BlendMode::Additive)
    {
        glBlendFunc( GL_ONE, GL_ONE );
        glEnable( GL_BLEND );
    }
    else
    {
        Log::Assert( false, "Unhandled blend mode." );
    }
}

void GfxDevice::SetClearColor( float red, float green, float blue )
{
    glClearColor( red, green, blue, 1 );
}

void GfxDevice::SetDepthFunc( DepthFunc test )
{
    if (test == DepthFunc::None)
    {
        glDepthMask( GL_FALSE );
        glDisable( GL_DEPTH_TEST );
    }
    else if (test == DepthFunc::EqualWriteOn)
    {
        glDepthMask( GL_TRUE );
        glEnable( GL_DEPTH_TEST );
        glDepthFunc( GL_EQUAL );
    }
    else if (test == DepthFunc::LessWriteOn)
    {
        glDepthMask( GL_TRUE );
        glEnable( GL_DEPTH_TEST );
        glDepthFunc( GL_LESS );
    }
    else if (test == DepthFunc::LessOrEqualWriteOn)
    {
        glDepthMask( GL_TRUE );
        glEnable( GL_DEPTH_TEST );
        glDepthFunc( GL_LEQUAL );
    }
    else if (test == DepthFunc::EqualWriteOff)
    {
        glDepthMask( GL_FALSE );
        glEnable( GL_DEPTH_TEST );
        glDepthFunc( GL_EQUAL );
    }
    else if (test == DepthFunc::LessWriteOff)
    {
        glDepthMask( GL_FALSE );
        glEnable( GL_DEPTH_TEST );
        glDepthFunc( GL_LESS );
    }
    else if (test == DepthFunc::LessOrEqualWriteOff)
    {
        glDepthMask( GL_FALSE );
        glEnable( GL_DEPTH_TEST );
        glDepthFunc( GL_LEQUAL );
    }
    else
    {
        Log::Assert( false, "Unimplemented depth test." );
    }
}

void GfxDevice::SetViewport( int left, int right, int top, int bottom )
{
    glViewport( left, top, right, bottom );
}

void GfxDevice::SetRenderTarget( const TextureImpl* target, unsigned cubeMapFace )
{
    ErrorCheck( "SetRenderTarget begin" );
    
    static unsigned sCurrentTarget = 0;
    m->currentRenderTarget = target;

    const bool isTargetTexture = target != nullptr;
    
    if (isTargetTexture)
    {
        glBindFramebuffer( GL_FRAMEBUFFER, target->FBO() );
        Statistics::IncRenderTargetSwitches();
        
        sCurrentTarget = target->FBO();
        
        if (target->IsCubemap())
        {
            const unsigned glCubeMapFace = GL_TEXTURE_CUBE_MAP_POSITIVE_X + cubeMapFace;
            Log::Assert( glCubeMapFace >= GL_TEXTURE_CUBE_MAP_POSITIVE_X && glCubeMapFace <= GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, "Invalid cube map face." );
            glFramebufferTexture2D( GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, glCubeMapFace, target->Id(), 0 );
        }
#if DEBUG  
        if (target->Width() == 0 || target->Height() == 0)
        {
            Log::Print( "Warning: Render texture size is 0." );
        }
#endif
        glViewport( 0, 0, target->Width(), target->Height() );
        glScissor ( 0, 0, target->Width(), target->Height() );
    }
    else
    {
        if (sCurrentTarget != Global::systemFBO)
        {
            glBindFramebuffer( GL_FRAMEBUFFER, Global::systemFBO );
            Statistics::IncRenderTargetSwitches();

#if AETHER3D_IOS
            BindDrawable();
#endif
            sCurrentTarget = Global::systemFBO;
        }
#if DEBUG
        if (Screen::Width() == 0 || Screen::Height() == 0)
        {
            Log::Print( "Warning: Backbuffer size is 0." );
        }
#endif
        glViewport( 0, 0, Screen::Width(), Screen::Height() );
        glScissor ( 0, 0, Screen::Width(), Screen::Height() );
    }
    
    ErrorCheckFBO();    
    ErrorCheck( "SetRenderTarget end" );
}

void GfxDevice::SetWireframe( bool enable )
{
#if !AETHER3D_GLES3
    glPolygonMode( GL_FRONT_AND_BACK, enable ? GL_LINE : GL_FILL );
    glLineWidth(1);
#endif
}

void GfxDevice::SetRasterizerState( RasterizerState rasterizerState )
{
    if (rasterizerState == RasterizerState::CullBackCCW)
    {
        glEnable( GL_CULL_FACE );
        glCullFace( GL_BACK );
        glFrontFace( GL_CCW );
    }
    else if (rasterizerState == RasterizerState::CullNoneCCW)
    {
        glDisable( GL_CULL_FACE );
        glFrontFace( GL_CCW );
    }
    else if (rasterizerState == RasterizerState::CullNoneCW)
    {
        glDisable( GL_CULL_FACE );
        glFrontFace( GL_CW );
    }
    else
    {
        Log::Assert( false, "Unhandled rasterizer state." );
    }
}

void GfxDevice::Set_sRGB_Writes( bool enable )
{
    (void)enable;
#if !AETHER3D_GLES3
    if (enable)
    {
        glEnable( GL_FRAMEBUFFER_SRGB );
    }
    else
    {
        glDisable( GL_FRAMEBUFFER_SRGB );
    }
#endif
}
