#if AETHER3D_GL43 || AETHER3D_D3D11
#include "ForwardPlus.hpp"
#endif
#include "LightImpl.hpp"
#include <cmath>
#include "Log.hpp"
#include "CameraImpl.hpp"
#include "RendererImpl.hpp"
#include "TextureImpl.hpp"
#include "Vec3.hpp"

#ifdef AETHER3D_WINDOWS
#    define AETHER3D_API __declspec(dllexport)
#else
#    define AETHER3D_API
#endif

extern "C" AETHER3D_API Light* CreateLight( Light::Type lightType )
{
    return new LightImpl( lightType );
}

const int notInScene = -1;

struct LightImpl::Impl
{
    Vec3 color = Vec3( 0.95f, 0.95f, 0.95f );
    Vec3 direction = Vec3( 0, 0, -1 );
    Type lightType = Type::Directional;
    bool castShadow = false;
    float radius = 5;
    float coneAngleCos = 0.5f;
    TextureImpl shadowMap;
    CameraImpl shadowCamera;
    int bufferIndex = notInScene; // If using Forward+, this indexes the light array that's used for culling/shading.
};

LightImpl::LightImpl( Light::Type lightType )// : m( new Impl )
{
    static_assert(sizeof( Impl ) <= StorageSize, "LightImpl storage too small");

    SceneGraphNode::type = SceneGraphNode::Type::Light;
    
    new(&_storage)Impl();
    m().lightType = lightType;
}

LightImpl::~LightImpl()
{
    m().~Impl();
}

void LightImpl::AddPointLightToTileCuller()
{
    Log::Assert( m().lightType == Light::Type::Point, "Wrong light type!" );

#if AETHER3D_GL43 || AETHER3D_D3D11
    if (m().bufferIndex != -1)
    {
        return;
    }

    m().bufferIndex = RendererImpl::Instance().GetForwardPlus().GetNextPointLightBufferIndex();

    if (m().bufferIndex == -1)
    {
        Log::Print( "Light: Too many point lights!  Maximum is 2048." );
    }

    // Tile culler needs to refresh these, so setting here.
    SetPosition( GetPosition() );
    SetRadius( Radius() );
    SetColor( GetColor() );
#endif
}

void LightImpl::AddSpotLightToTileCuller()
{
    Log::Assert( m().lightType == Light::Type::Spot, "Wrong light type!" );

#if AETHER3D_GL43 || AETHER3D_D3D11
    if (m().bufferIndex != -1)
    {
        return;
    }

    m().bufferIndex = RendererImpl::Instance().GetForwardPlus().GetNextSpotLightBufferIndex();

    if (m().bufferIndex == -1)
    {
        Log::Print( "Light: Too many spot lights!  Maximum is 2048." );
    }

    // Tile culler needs to refresh these, so setting here.
    SetPosition( GetPosition() );
    SetRadius( Radius() );
    SetColor( GetColor() );
    SetDirection( Direction() );
#endif
}

Light::Type LightImpl::GetLightType() const
{
    return m().lightType;
}

void LightImpl::SetCastShadow( bool enable, int shadowMapSize )
{
    if (m().lightType == Light::Type::Point && enable)
    { 
        Log::Print( "Light::SetCastShadow:  Point lights currently do not support shadow maps." );
        return;
    }

    m().castShadow = enable;

    if (m().castShadow)
    {
        m().shadowMap.CreateRenderTexture2D( shadowMapSize, shadowMapSize, TextureImpl::Format::Float,
                                            TextureImpl::Filter::Linear, TextureImpl::Tiling::Clamp, "shadow map" );
    }
}

void LightImpl::SetColor( const Vec3& newColor )
{
    m().color = newColor;

#if AETHER3D_GL43 || AETHER3D_D3D11
    if (m().lightType == Light::Type::Point && m().bufferIndex != notInScene)
    {
        RendererImpl::Instance().GetForwardPlus().SetPointLightColor( m().bufferIndex, newColor );
    }
    else if (m().lightType == Light::Type::Spot && m().bufferIndex != notInScene)
    {
        RendererImpl::Instance().GetForwardPlus().SetSpotLightColor( m().bufferIndex, newColor );
    }
#endif
}

Vec3 LightImpl::GetColor() const
{
    return m().color;
}

float LightImpl::GetConeAngleCos() const
{
    return m().coneAngleCos;
}

void LightImpl::SetConeAngle( float degrees )
{
    m().coneAngleCos = std::cos( degrees * (3.14159265f / 180.0f) );
}

void LightImpl::SetDirection( const Vec3& newDirection )
{
    m().direction = newDirection;
    
    // TODO: Move into UpdateTransform if possible.
#if AETHER3D_GL43 || AETHER3D_D3D11
    if (m().lightType == Light::Type::Spot && m().bufferIndex != notInScene)
    {
        // TODO: worldDirection instead of direction.
        RendererImpl::Instance().GetForwardPlus().SetSpotLightParams( m().bufferIndex, newDirection, m().coneAngleCos, m().radius /* TODO: Separate falloff from radius */);
    }
#endif
}

void LightImpl::SetPosition( const Vec3& aPosition )
{
    position = aPosition;
    dirtyTransform = true;

#if AETHER3D_GL43 || AETHER3D_D3D11
    if (m().lightType == Light::Type::Point && m().bufferIndex != notInScene)
    {
        RendererImpl::Instance().GetForwardPlus().SetPointLightPositionAndRadius( m().bufferIndex, Vec4( aPosition, m().radius ) );
    }
    else if (m().lightType == Light::Type::Spot && m().bufferIndex != notInScene)
    {
        RendererImpl::Instance().GetForwardPlus().SetSpotLightPositionAndRadius( m().bufferIndex, Vec4( aPosition, m().radius ) );
    }
#endif
}

void LightImpl::SetRotation( const Quaternion& aRotation )
{
    rotation = aRotation;
    dirtyTransform = true;
}

void LightImpl::SetRadius( float radius )
{
    if (radius < 0)
    {
        Log::Print( "Light::SetRadius: Radius cannot be less than 0." );
        m().radius = 5;
        return;
    }

    m().radius = radius;

#if AETHER3D_GL43 || AETHER3D_D3D11
    if (m().lightType == Light::Type::Point && m().bufferIndex != notInScene)
    {
        RendererImpl::Instance().GetForwardPlus().SetPointLightPositionAndRadius( m().bufferIndex, Vec4( position, radius ) );
    }
    else if (m().lightType == Light::Type::Spot && m().bufferIndex != notInScene)
    {
        RendererImpl::Instance().GetForwardPlus().SetSpotLightPositionAndRadius( m().bufferIndex, Vec4( position, radius ) );
    }
#endif
}

bool LightImpl::CastsShadow() const
{
    return m().castShadow;
}

const Vec3& LightImpl::Color() const
{
    return m().color;
}

const Vec3& LightImpl::Direction() const
{
    return m().direction;
}

float LightImpl::Radius() const
{
    return m().radius;
}

TextureImpl& LightImpl::ShadowMap()
{
    return m().shadowMap;
}

CameraImpl& LightImpl::ShadowCamera()
{
    return m().shadowCamera;
}

void LightImpl::UpdateTransform()
{
    Matrix44::TransformPoint( Vec3( 0, 0, 0 ), modelMatrix, &worldPosition );
    Matrix44::TransformDirection( m().direction, modelMatrix, &worldDirection );

#if AETHER3D_GL43 || AETHER3D_D3D11
    if (m().lightType == Light::Type::Point && m().bufferIndex != notInScene)
    {
        RendererImpl::Instance().GetForwardPlus().SetPointLightPositionAndRadius( m().bufferIndex, Vec4( worldPosition, m().radius ) );
    }
    else if (m().lightType == Light::Type::Spot && m().bufferIndex != notInScene)
    {
        // TODO: Set Spot light direction multiplied by modelMatrix.
        RendererImpl::Instance().GetForwardPlus().SetSpotLightPositionAndRadius( m().bufferIndex, Vec4( worldPosition, m().radius ) );
    }
#endif
}
