#ifndef LIGHT_IMPL_H
#define LIGHT_IMPL_H

#include <type_traits>
#ifndef LIGHT_H
#    include "Light.hpp"
#endif

struct Vec3;
class CameraImpl;
class TextureImpl;

class LightImpl : public Light
{
public:
    explicit LightImpl( Light::Type lightType );
    virtual ~LightImpl();

    /** \return Type. */
    Type GetLightType() const override;

    /**
       \param enable True, if the light casts shadow.  Disabled by default.  Currently only supported for Directional and Spot Light.
       \param shadowMapSize Shadow map's size in pixels.  Width and height will be set to shadowMapSize.
     */
    void SetCastShadow( bool enable, int shadowMapSize ) override;
    
    /** \param newColor Color.  Values are in range 0-1. */
    void SetColor( const Vec3& newColor ) override;
    
    /** \return Color in range 0-1. */
    Vec3 GetColor() const override;

    /** \param value Spot light's cone angle in degrees. */
    void SetConeAngle( float degrees ) override;

    /** \paran newDirection Normalized direction vector.  Doesn't affect point lights. */
    void SetDirection( const Vec3& newDirection ) override;

    /** \param position Position.  Only meaningful for Point Light. */
    void SetPosition( const Vec3& position ) override;

    /** \param aRotation Absolute rotation. */
    void SetRotation( const Quaternion& rotation ) override;
    
    /** \param radius Radius  Only meaningful for Point Light. */ 
    void SetRadius( float radius ) override;

    /** \return Spot light's cone angle cos. */
    float GetConeAngleCos() const;
    
    /** Adds the light to tile culler.  Called when light is added to the scene. */
    void AddPointLightToTileCuller();

    /** Adds the light to tile culler.  Called when light is added to the scene. */
    void AddSpotLightToTileCuller();

    /** \return True, if this light casts shadow. */
    bool CastsShadow() const;

    /** \return color. */
    const Vec3& Color() const;
    
    /** \return direction. */
    const Vec3& Direction() const;

    /** \return radius. */
    float Radius() const;

    /** \return Shadow map. */
    TextureImpl& ShadowMap();
    
    /** \return Shadow camera. */
    CameraImpl& ShadowCamera();
    
    void UpdateTransform();

    /** Local position transformed by scene graph hierarchy.  Currently only used for Lights, so not in SceneGraphNode. */
    Vec3 worldPosition;

    /** Local direction transformed by scene graph hierarchy.  Currently only used for Lights, so not in SceneGraphNode. */
    Vec3 worldDirection;
    
private:
    struct Impl;
    Impl& m() { return reinterpret_cast<Impl&>(_storage); }
    Impl const& m() const { return reinterpret_cast<Impl const&>(_storage); }

    static const size_t StorageSize = 384;
    static const size_t StorageAlign = 16;

    std::aligned_storage<StorageSize, StorageAlign>::type _storage;
};

#endif
