#include "ModelImpl.hpp"
#include <cstdint>
#include <fstream>
#include <map>
#include <list>
#include <sstream>
#include <vector>
#include "Errors.hpp"
#include "FileWatcher.hpp"
#include "Log.hpp"
#include "MathUtil.hpp"
#include "Matrix.hpp"
#include "Mesh.hpp"
#include "Quaternion.hpp"
#include "Vec3.hpp"
#include "VertexBuffer.hpp"

#ifdef AETHER3D_WINDOWS
#    define AETHER3D_API __declspec(dllexport)
#else
#    define AETHER3D_API
#endif

const char* GetFullPath( const char* fileName );

extern "C" AETHER3D_API Model* CreateModel( const char* path, Error::Enum* outError )
{
    ModelImpl* outModel = new ModelImpl;
    *outError = outModel->LoadFromFile( path );
    return outModel;
}

struct ModelImpl::Impl
{
    Impl() {}
    /**
       Combines meshes for depth-pass batching.
       Must be called when mesh's material is changed.
    */
    void CombineOpaqueMeshes();

    Vec3 aabbMin;
    Vec3 aabbMax;
    Vec3 aabbMinWorld;
    Vec3 aabbMaxWorld;
    VertexBuffer combinedOpaqueMeshes; // For depth-pass.  All meshes combined into one.
    std::vector< Mesh > meshes;
    std::string path;
    bool wireFrame = false;
};

void ModelImpl::Impl::CombineOpaqueMeshes()
{
    // Finds out # of indices/vertices.
    std::size_t faceCount  = 0;
    std::size_t vertexCount = 0;
    
    for (auto& mesh : meshes)
    {
        // Cannot combine skinned meshes.
        if (mesh.HasSkin())
        {
            return;
        }

        faceCount   += mesh.Indices().size();
        vertexCount += mesh.Vertices().size();
    }

    if (vertexCount > 65536)
    {
        return;
    }

    std::vector< Face > faces( faceCount );
    std::vector< Vertex > vertices( vertexCount );
    
    unsigned short accumIndices = 0;
    unsigned short faceIndex    = 0;
    unsigned short vertexIndex  = 0;
    
    for (auto& mesh : meshes)
    {
        if (mesh.IsOpaque())
        {
            const std::vector< Face >& meshFaces = mesh.Indices();

            for (std::size_t i = 0; i < meshFaces.size(); ++i)
            {
                faces[ faceIndex ] = meshFaces[ i ];
                faces[ faceIndex ].a += accumIndices;
                faces[ faceIndex ].b += accumIndices;
                faces[ faceIndex ].c += accumIndices;
                ++faceIndex;
            }

            const std::vector< Vertex >& meshVertices = mesh.Vertices();
            
            accumIndices += (unsigned short)meshVertices.size();
            
            for (std::size_t v = 0; v < meshVertices.size(); ++v)
            {
                vertices[ vertexIndex ] = meshVertices[ v ];
                ++vertexIndex;
            }
        }
    }

    std::vector< Influence > emptyInfluences;
    combinedOpaqueMeshes.Generate( faces, vertices, emptyInfluences );
}

namespace Global
{
    static std::map< std::string, ModelImpl > sPathToModelCache;
    static std::map< std::string, std::list< ModelImpl* > > sPathToModelInstances;
}

void ModelImpl::GetAABB( Vec3& outMin, Vec3& outMax ) const
{
    outMin = AABBMinWorld();
    outMax = AABBMaxWorld();
}

const std::string& ModelImpl::GetPath() const
{
    return m->path;
}

const VertexBuffer& ModelImpl::CombinedOpaqueMeshes() const
{
    return m->combinedOpaqueMeshes;
}

void ModelImpl::Deinit()
{
    for (auto modelType : Global::sPathToModelInstances)
    {
        for (auto modelInstance : modelType.second)
        {
            for (auto mesh : modelInstance->Meshes())
            {
                mesh.Deinit();
            }
        }
    }
}

void ModelImpl::EnableWireframe( bool enable )
{
    m->wireFrame = enable;
}

bool ModelImpl::IsWireframe() const
{
    return m->wireFrame;
}

void ModelImpl::Reload( const std::string& path )
{
    if (Global::sPathToModelCache.find( path ) == Global::sPathToModelCache.end())
    {
        Log::Assert( false, "Tried to reload a model that wasnt loaded." );
        return;
    }

    Global::sPathToModelCache[ path ].Deinit(); // \todo Review and test with multiple instances.
    Global::sPathToModelCache[ path ].LoadFromFile( path );
    
    for (auto model : Global::sPathToModelInstances[ path ])
    {
        Log::Print( "Reloading model instance." );
        model->LoadFromFile( path );
    }
}

ModelImpl::ModelImpl() : m( new Impl )
{
    SceneGraphNode::type = Type::Model;
}

ModelImpl::ModelImpl( const ModelImpl& other ) : m( new Impl )
{
    *this = other;
}

ModelImpl::~ModelImpl() 
{
}

ModelImpl& ModelImpl::operator=( const ModelImpl& other )
{
    *m = *other.m;

    for (auto mesh : m->meshes)
    {
        mesh.SetModel( this );
    }

    type = Type::Model;

    return *this;
}

std::vector< std::string > ModelImpl::GetMeshNames() const
{
    std::vector< std::string > outNames( m->meshes.size() );

    for (std::size_t i = 0; i < m->meshes.size(); ++i)
    {
        outNames[ i ] = m->meshes[ i ].Name();
    }

    return outNames;
}

std::vector< std::string > ModelImpl::GetMeshJointNames( const std::string& meshName ) const
{
    std::vector< std::string > outNames;
    
    for (auto mesh : m->meshes)
    {
        if (mesh.Name() == meshName)
        {
            for (auto jointName : mesh.JointNames())
            {
                outNames.emplace_back( jointName );
            }
        }
    }
    
    return outNames;
}

void ModelImpl::ApplyAnimation( int firstKeyFrame, int lastKeyFrame )
{
    for (auto& mesh : m->meshes)
    {
        mesh.StartAnimations( firstKeyFrame, lastKeyFrame );
    }
}

Material* ModelImpl::GetMeshMaterial( const char* meshName )
{
    for (auto& mesh : m->meshes)
    {
        if (mesh.Name() == meshName)
        {
            return (Material*)mesh.Material();
        }
    }

    Log::Print( "GetMeshMaterial: Mesh %s was not found.", meshName );
    return nullptr;
}

void ModelImpl::SetMaterialForMesh( Material* material, const std::string& meshName )
{
    for (auto& mesh : m->meshes)
    {
        if (mesh.Name() == meshName)
        {
            mesh.SetMaterial( reinterpret_cast< MaterialImpl* >(material) );
            m->CombineOpaqueMeshes();
            return;
        }
    }

    Log::Print( "SetMaterialForMesh: %s was not found.", meshName.c_str() );
}

void ModelImpl::SetMaterialForMeshSubstring( Material* material, const std::string& meshNameSubstring )
{
    bool found = false;

    for (auto& mesh : m->meshes)
    {
        if (mesh.Name().find( meshNameSubstring ) != std::string::npos )
        {
            mesh.SetMaterial( reinterpret_cast< MaterialImpl* >(material) );
            found = true;
        }
    }

    if (!found)
    {
        Log::Print( "SetMaterialForMeshSubstring: None of the mesh names in model %s contain %s", GetPath().c_str(), meshNameSubstring.c_str() );
    }

    m->CombineOpaqueMeshes();
}

const Vec3& ModelImpl::AABBMax() const
{
    return m->aabbMax;
}

const Vec3& ModelImpl::AABBMin() const
{
    return m->aabbMin;
}

const Vec3& ModelImpl::AABBMaxWorld() const
{
    return m->aabbMaxWorld;
}

const Vec3& ModelImpl::AABBMinWorld() const
{
    return m->aabbMinWorld;
}

Error::Enum ModelImpl::LoadFromFile( const std::string& path )
{
    static_assert( sizeof( Vertex ) == 64, "" );
    static_assert( sizeof( Vec3   ) == 12, "" );
    static_assert( sizeof( Face   ) ==  6, "" );

    m->path = path;

    const bool firstTimeLoad = m->meshes.empty();

    if (firstTimeLoad && Global::sPathToModelCache.find( path ) != Global::sPathToModelCache.end())
    {
        *this = Global::sPathToModelCache[ path ];
        m->CombineOpaqueMeshes();
        return Error::NO_ERRORS;
    }

    std::ifstream ifs( GetFullPath( path.c_str() ), std::ios::binary );

    if (!ifs)
    {
        Log::Print( "MakeModel: %s not found!", GetFullPath( path.c_str() ) );
        return Error::FILE_NOT_FOUND;
    }

    std::vector< char > fileData;
    fileData.assign( std::istreambuf_iterator< char >( ifs ), std::istreambuf_iterator< char >() );
    
    uint8_t magic[ 2 ];

    std::istringstream is( std::string( &fileData[ 0 ], fileData.size() ) );

    is.read( (char*)magic, sizeof( magic ) );

    if (magic[ 0 ] != 'a' || magic[ 1 ] != '9')
    {
        Log::Print( "MakeModel: %s is corrupted or old format: Wrong magic number!", path.c_str() );
        return Error::CORRUPTED_DATA;
    }

    is.read( (char*)&m->aabbMin, sizeof( m->aabbMin ) );
    is.read( (char*)&m->aabbMax, sizeof( m->aabbMax ) );

    uint16_t meshCount;
    is.read( (char*)&meshCount, sizeof( meshCount ) );

    m->meshes.clear();
    m->meshes.resize( meshCount );

    Mesh* baseMesh = nullptr;

    for (auto& mesh : m->meshes)
    {
        mesh.SetModel( this );

        Error::Enum error = mesh.ReadGeometry( is );

        if (error != Error::NO_ERRORS)
        {
            return error;
        }

        // Skin bit.  0 if there is no skin, 1 otherwise.
        unsigned char hasSkin;
        is.read( (char*)&hasSkin, 1 );

        if (hasSkin != 0 && hasSkin != 1)
        {
            Log::Print( "Model data corrupted: %s", path.c_str() );
            return Error::CORRUPTED_DATA;
        }
        
        if (hasSkin == 1)
        {
            error = mesh.ReadSkin( is );

            if (error != Error::NO_ERRORS)
            {
                Log::Print( "Skin data corrupted: %s", path.c_str() );
                return error;
            }
        }

        if (mesh.IsMorphBaseMesh())
        {
            baseMesh = &mesh;
        }

        error = mesh.ReadAnimations( is );

        if (error != Error::NO_ERRORS)
        {
            Log::Print( "Animation data corrupted: %s", path.c_str() );
            return error;
        }
    }
    
    unsigned char terminator;
    is.read( (char*)&terminator, 1 );

    if (terminator != 100)
    {
        Log::Print( "Corrupted or incomplete model (terminator not found): %s", path.c_str() );
        return Error::CORRUPTED_DATA;
    }

    // Sets morph positions for morph base mesh, if any.
    if (baseMesh != nullptr)
    {
        int targetIndex = 0;

        for (auto morphTargetName : baseMesh->MorphTargetNames())
        {
            for (const auto& mesh : m->meshes)
            {
                if (targetIndex == 0 && mesh.Name() == morphTargetName)
                {
                    ++targetIndex;
                    baseMesh->SetMorphPositions1( mesh.GetVertexPositions() );
                }
                else if (targetIndex == 1 && mesh.Name() == morphTargetName)
                {
                    baseMesh->SetMorphPositions2( mesh.GetVertexPositions() );
                }
            }
        }
    }

    for (auto& mesh : m->meshes)
    {
        mesh.GenerateVertexBuffer();
    }

    if (firstTimeLoad)
    {
        Global::sPathToModelCache[ path ] = *this;
        FileWatcher::Instance().AddModel( path.c_str() );
        Global::sPathToModelInstances[ path ].push_front( this );
    }

    return Error::NO_ERRORS;
}

const Matrix44& ModelImpl::ModelMatrix() const
{
    return modelMatrix;
}

std::vector< Mesh >& ModelImpl::Meshes()
{
    return m->meshes;
}

void ModelImpl::SolveWorldAABB()
{
    std::vector< Vec3 > aabbWorld;
    MathUtil::GetCorners( m->aabbMin, m->aabbMax, aabbWorld );
        
    for (std::size_t v = 0; v < aabbWorld.size(); ++v)
    {
        Matrix44::TransformPoint( aabbWorld[ v ], modelMatrix, &aabbWorld[ v ] );
    }

    MathUtil::GetMinMax( aabbWorld, m->aabbMinWorld, m->aabbMaxWorld );

    for (auto& mesh : Meshes())
    {
        Vec3 meshAabbMinWorld = mesh.AABBMin();
        Vec3 meshAabbMaxWorld = mesh.AABBMax();

        std::vector< Vec3 > meshAabbWorld;
        MathUtil::GetCorners( meshAabbMinWorld, meshAabbMaxWorld, meshAabbWorld );
                    
        for (std::size_t v = 0; v < meshAabbWorld.size(); ++v)
        {
            Matrix44::TransformPoint( meshAabbWorld[ v ], modelMatrix, &meshAabbWorld[ v ] );
        }
                    
        MathUtil::GetMinMax( meshAabbWorld, meshAabbMinWorld, meshAabbMaxWorld );
        mesh.SetWorldAABB( meshAabbMinWorld, meshAabbMaxWorld );
        mesh.SetModelMatrix( modelMatrix );
    }
}
