#ifndef RENDERQUEUE_H
#define RENDERQUEUE_H

#ifndef INCLUDE_MEMORY_H
#    include <memory>
#    define INCLUDE_MEMORY_H
#endif
#ifndef INCLUDE_VECTOR_H
#    include <vector>
#    define INCLUDE_VECTOR_H
#endif

struct Matrix44;
class TextureImpl;
class Shader;
struct Vec4;
struct Vec3;

/// RenderQueue stores/batches Renderer's Draw() calls and draws them using unlit shader in Renderer::FinishFrame().
class RenderQueue
{
public:
    struct Vec3uv
    {
        Vec3uv();
        Vec3uv( float x, float y, float z, float u, float v );
        Vec3uv operator+( const Vec3& vec ) const;

        float x, y, z, u, v;
    };

    enum class PrimitiveType { Triangles, Lines };

    struct RenderQueueEntry
    {
        RenderQueueEntry() :
            sortingValue( 9999999 ) // \fixme Sorting code relies on large default sorting value.
            , texture( nullptr )
            , red( 1 )
            , green( 1 )
            , blue( 1 )
            , opacity( 1 )
            , primitiveType( RenderQueue::PrimitiveType::Triangles )
        {}

        RenderQueueEntry( int aSortingValue,
            const TextureImpl* aTexture,
            const std::vector< RenderQueue::Vec3uv >& aVertices,
            float aRed,
            float aGreen,
            float aBlue,
            float aOpacity,
            RenderQueue::PrimitiveType aPrimitive ) :
            sortingValue( aSortingValue )
            , texture( aTexture )
            , vertices( aVertices )
            , red( aRed )
            , green( aGreen )
            , blue( aBlue )
            , opacity( aOpacity )
            , primitiveType( aPrimitive )
        {}

    /*
     Commented destructor out because of this clang warning:
     
     warning: definition of implicit copy assignment
        operator for 'Impl' is deprecated because it has a user-declared
            destructor [-Wdeprecated]*/
     //virtual ~RenderQueueEntry() {}

        int sortingValue;
        const TextureImpl* texture;
        std::vector< RenderQueue::Vec3uv > vertices;
        float red, green, blue, opacity;
        RenderQueue::PrimitiveType primitiveType;
    };

    explicit RenderQueue( Shader& shader );
    virtual ~RenderQueue();
    
    /// If added entry differs only by vertices, it's appended to an already existing entry.
    /// Entries having smaller sorting values are rendered before larger values.
    void Add( PrimitiveType primitive, const std::vector< Vec3uv >& vertices, const TextureImpl* texture,
              int sortingValue, const Vec4& color );
    void Deinit() const;
    /**
       Renders all the items in the queue.
     
       \param width Render target's width.
       \param height Render target's height.
       \param viewProjection View-Projection matrix used for line drawing.
     */
    void Flush( float width, float height, const Matrix44& viewProjection );

private:
    struct Impl;
    std::unique_ptr< Impl > m;
};
#endif
