#include "RendererImpl.hpp"
#include <string>
#if AETHER3D_GL32 || AETHER3D_GL43
#    include "GL/glxw.h"
#    include <SDL/SDL.h>
#endif
#if defined AETHER3D_GLES3 && !defined AETHER3D_IOS
#    include <GLES3/gl3.h>
#    include <GLES2/gl2ext.h>
#    include <GLES2/gl2extQCOM.h>
#endif
#if AETHER3D_IOS
#    include <OpenGLES/ES3/glext.h>
#endif
#ifdef AETHER3D_WINDOWS
#    define AETHER3D_API __declspec(dllexport)
#else
#    define AETHER3D_API 
#endif

#if !defined AETHER3D_UNIX && !defined AETHER3D_WINDOWS && !defined AETHER3D_IOS
#    error "Platform not defined! Define either AETHER3D_UNIX or AETHER3D_WINDOWS."
#endif

#if !defined AETHER3D_GL32 && !defined AETHER3D_GL43 && !defined AETHER3D_GLES3
#    error "GL Renderer not defined! It must be either AETHER3D_GL32, AETHER3D_GL43 or AETHER3D_GLES3."
#endif

#include "FileWatcher.hpp"
#include "GfxDevice.hpp"
#include "MaterialImpl.hpp"
#include "Log.hpp"
#include "Statistics.hpp"
#if AETHER3D_GL43
#    include "ForwardPlus.hpp"
#endif

std::vector< Vec4 > CreateSSAOKernel( int size );
void CreateQuadBuffer( VertexBuffer& outQuadBuffer );
void CreateSkybox( VertexBuffer& outSkybox );
bool CreateAppWindow( int width, int height, int* outScreenSize );
bool CreateContext();
void EGLContextSwapBuffers();
void ErrorCheck( const std::string& info );
bool HasExtension( const std::string& glExtension );

#if AETHER3D_GL32 || AETHER3D_GL43
SDL_Window* mainWindow = nullptr;
SDL_GLContext mainContext2;

void ErrorCheckSDL()
{
    const char *error = SDL_GetError();

    if (*error != '\0')
    {
        SDL_ShowSimpleMessageBox( SDL_MESSAGEBOX_ERROR, "Simple DirectMedia Layer Error", SDL_GetError(), nullptr );
        SDL_Quit();
        exit( 1 );
    }
}
#endif

void SetDebugEventBegin( const std::string& name )
{
    (void)name;
#if AETHER3D_GL32 || AETHER3D_GL43
    if (HasExtension("GL_KHR_debug"))
    {
        glPushDebugGroup( GL_DEBUG_SOURCE_APPLICATION, 0, (GLsizei)name.length(), name.c_str() );
    }
#endif
#if AETHER3D_IOS
    glPushGroupMarkerEXT( 0, name.c_str() );
#endif
}

void SetDebugEventEnd()
{
#if AETHER3D_GL32 || AETHER3D_GL43
    if (HasExtension("GL_KHR_debug"))
    {
        glPopDebugGroup();
    }
#endif
#if AETHER3D_IOS
    glPopGroupMarkerEXT();
#endif
}

namespace Global
{
    std::unordered_map< unsigned, const char* > glErrors;
    GLuint queryIds[ 2 ] = { 0 };
    extern int maxInstanceCount;
    GLuint systemFBO;
}

void ErrorCheck( const std::string& info )
{
    (void)info;
#if defined _DEBUG || defined DEBUG
    GLenum errorCode = GL_INVALID_ENUM;

    while ((errorCode = glGetError()) != GL_NO_ERROR)
    {
        Log::Print( "%s caused an OpenGL error: %s", info.c_str(), Global::glErrors[ (unsigned)errorCode ] );
        Log::Assert( false, "OpenGL error!" );
    }
#endif
}

void ErrorCheckFBO()
{
#if defined _DEBUG || defined DEBUG
    const GLenum error = glCheckFramebufferStatus( GL_FRAMEBUFFER );

    if (error == GL_FRAMEBUFFER_COMPLETE)
    {
        return;
    }
    else if (error == GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT)
    {
        Log::Assert( false, "GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT: Not all framebuffer attachment points are framebuffer attachment complete." );
    }
    else if (error == GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT)
    {
        Log::Assert( false, "GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: No images are attached to the framebuffer." );
    }
    else if (error == GL_FRAMEBUFFER_UNSUPPORTED)
    {
        Log::Assert( false, "GL_FRAMEBUFFER_UNSUPPORTED: The combination of internal formats of the attached images violates an implementation-dependent set of restrictions." );
    }
    else if (error == GL_INVALID_ENUM)
    {
        Log::Assert( false, "GL_INVALID_ENUM: Invalid FBO target type." );
    }
    else if (error == GL_FRAMEBUFFER_UNDEFINED)
    {
        Log::Assert( false, "GL_FRAMEBUFFER_UNDEFINED: FBO is undefined." );
    }
    else if (error == GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE)
    {
        Log::Assert( false, "GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE" );
    }
    else
    {
        Log::Assert( false, "FBO: Unhandled error!" );
    }
#endif
}

bool HasExtension( const std::string& glExtension )
{
    static std::vector< std::string > sExtensions;

    if (sExtensions.empty())
    {
        int count;
        glGetIntegerv( GL_NUM_EXTENSIONS, &count );
        sExtensions.resize( (std::size_t)count );

        for (int i = 0; i < count; ++i)
        {
            sExtensions[ i ] = std::string( (const char*)glGetStringi( GL_EXTENSIONS, i ) );
        }
    }

    return std::find( std::begin( sExtensions ), std::end( sExtensions ), glExtension ) != std::end( sExtensions );
}

#if AETHER3D_GL43
#if AETHER3D_WINDOWS
#    define snprintf _snprintf
#endif

void FormatDebugOutputARB( char outStr[], size_t outStrSize, GLenum source, GLenum type,
    GLuint id, GLenum severity, const char *msg )
{
    char sourceStr[ 32 ];
    const char *sourceFmt = "UNDEFINED(0x%04X)";

    switch (source)
    {
    case GL_DEBUG_SOURCE_API_ARB:             sourceFmt = "API"; break;
    case GL_DEBUG_SOURCE_WINDOW_SYSTEM_ARB:   sourceFmt = "WINDOW_SYSTEM"; break;
    case GL_DEBUG_SOURCE_SHADER_COMPILER_ARB: sourceFmt = "SHADER_COMPILER"; break;
    case GL_DEBUG_SOURCE_THIRD_PARTY_ARB:     sourceFmt = "THIRD_PARTY"; break;
    case GL_DEBUG_SOURCE_APPLICATION_ARB:     sourceFmt = "APPLICATION"; break;
    case GL_DEBUG_SOURCE_OTHER_ARB:           sourceFmt = "OTHER"; break;
    }

    snprintf( sourceStr, 32, sourceFmt, source );

    char typeStr[ 32 ];
    const char *typeFmt = "UNDEFINED(0x%04X)";

    switch (type)
    {
    case GL_DEBUG_TYPE_ERROR_ARB:               typeFmt = "ERROR"; break;
    case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR_ARB: typeFmt = "DEPRECATED_BEHAVIOR"; break;
    case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR_ARB:  typeFmt = "UNDEFINED_BEHAVIOR"; break;
    case GL_DEBUG_TYPE_PORTABILITY_ARB:         typeFmt = "PORTABILITY"; break;
    case GL_DEBUG_TYPE_PERFORMANCE_ARB:         typeFmt = "PERFORMANCE"; break;
    case GL_DEBUG_TYPE_OTHER_ARB:               typeFmt = "OTHER"; break;
    }

    snprintf( typeStr, 32, typeFmt, type );

    char severityStr[ 32 ];
    const char *severityFmt = "UNDEFINED";

    switch (severity)
    {
    case GL_DEBUG_SEVERITY_HIGH_ARB:   severityFmt = "HIGH";   break;
    case GL_DEBUG_SEVERITY_MEDIUM_ARB: severityFmt = "MEDIUM"; break;
    case GL_DEBUG_SEVERITY_LOW_ARB:    severityFmt = "LOW"; break;
    }

    snprintf( severityStr, 32, severityFmt, severity );

    snprintf( outStr, outStrSize, "OpenGL: %s [source=%s type=%s severity=%s id=%u]",
        msg, sourceStr, typeStr, severityStr, id );
}

void APIENTRY DebugCallbackARB( GLenum source, GLenum type, GLuint id, GLenum severity,
    GLsizei length, const GLchar *message, const GLvoid *userParam )
{
    const int undefinedSeverity = 33387;
    
    // NVIDIA driver spams buffer creation messages at 'undefined' severity, this silences them.
    if (severity != undefinedSeverity)
    {
        Log::Print( "severity: %d", severity );
        (void)length;
        (void)userParam;
        char finalMessage[ 256 ];
        FormatDebugOutputARB( finalMessage, 256, source, type, id, severity, message );
        Log::Print( "%s", finalMessage );

        const int performanceSeverity = 37191;
        
        if (severity != performanceSeverity)
        {
            Log::Assert( false, "OpenGL error" );
        }
    }
}
#endif

void UploadModelMatrices( Matrix44* modelMatrices, int matrixCount, Shader& depthNormalShader )
{
    static GLuint depthNormalUBO = 0;
    static GLuint momentsUBO = 0;

    GLuint* ubo = nullptr;

    if (&depthNormalShader == &RendererImpl::Instance().BuiltinShader().depthNormalShader)
    {
        ubo = &depthNormalUBO;
    }
    else if (&depthNormalShader == &RendererImpl::Instance().BuiltinShader().momentsShader)
    {
        ubo = &momentsUBO;
    }
    else
    {
        Log::Assert( false, "UploadModelMatrices: Unhandled shader!" );
        ubo = &momentsUBO;
    }

    const GLuint bindingPoint = 1;

    if (*ubo == 0)
    {
        GLuint blockIndex = glGetUniformBlockIndex( depthNormalShader.GLSLProgram(), "InstanceBuffer" );
        glUniformBlockBinding( depthNormalShader.GLSLProgram(), blockIndex, bindingPoint );

        glGenBuffers( 1, ubo );
        glBindBuffer( GL_UNIFORM_BUFFER, *ubo );

        glBufferData( GL_UNIFORM_BUFFER, sizeof( Matrix44 ) * Global::maxInstanceCount, nullptr, GL_DYNAMIC_DRAW );
    }

    glBindBuffer( GL_UNIFORM_BUFFER, *ubo );
    glBindBufferBase( GL_UNIFORM_BUFFER, bindingPoint, *ubo );
    glBufferSubData( GL_UNIFORM_BUFFER, 0, sizeof( Matrix44 ) * matrixCount, modelMatrices );
}

#if AETHER3D_GL32 || AETHER3D_GL43
Error::Enum CreateAppWindowSDL( int width, int height, bool fullScreen, int antialiasSamples, int* outScreenSize )
{
#if AETHER3D_GL32
    const int major = 3;
    const int minor = 2;
#endif
#if AETHER3D_GL43
    const int major = 4;
    const int minor = 3;
#endif
    if (SDL_Init( SDL_INIT_VIDEO ) == -1)
    {
        Log::Print( "Unable to init SDL: %s", SDL_GetError() );
        return Error::UNSUPPORTED_RENDERER;
    }

    SDL_DisplayMode displayMode;
    SDL_GetDesktopDisplayMode( 0, &displayMode );

    outScreenSize[ 0 ] = width == 0 ? displayMode.w : width;
    outScreenSize[ 1 ] = height == 0 ? displayMode.h : height;

    SDL_GL_SetAttribute( SDL_GL_CONTEXT_MAJOR_VERSION, major );
    SDL_GL_SetAttribute( SDL_GL_CONTEXT_MINOR_VERSION, minor );
    SDL_GL_SetAttribute( SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE );

#if SDL_VERSION_ATLEAST( 2, 0, 1 )
    SDL_GL_SetAttribute( SDL_GL_FRAMEBUFFER_SRGB_CAPABLE, 1 );
#endif

    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
    SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 24 );
#if DEBUG
    SDL_GL_SetAttribute( SDL_GL_CONTEXT_FLAGS, SDL_GL_CONTEXT_DEBUG_FLAG | SDL_GL_CONTEXT_FORWARD_COMPATIBLE_FLAG );
#endif

    if (antialiasSamples > 1)
    {
        SDL_GL_SetAttribute( SDL_GL_MULTISAMPLEBUFFERS, 1 );
        SDL_GL_SetAttribute( SDL_GL_MULTISAMPLESAMPLES, antialiasSamples );
    }

    const unsigned createFlags = SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN | (fullScreen ? SDL_WINDOW_FULLSCREEN_DESKTOP : 0);

    mainWindow = SDL_CreateWindow( "Aether3D App", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
        outScreenSize[ 0 ], outScreenSize[ 1 ], createFlags );
    if (!mainWindow)
    {
#if SDL_VERSION_ATLEAST( 2, 0, 1 )
        SDL_GL_SetAttribute( SDL_GL_FRAMEBUFFER_SRGB_CAPABLE, 0 );
#endif
        SDL_ClearError();

        mainWindow = SDL_CreateWindow( "Aether3D App", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
            outScreenSize[ 0 ], outScreenSize[ 1 ], createFlags );
        if (mainWindow)
        {
            Log::Print( "OpenWindow: sRGB framebuffer not supported, disabling support." );
        }
        else
        {
            Log::Print( "OpenWindow: Could not create window!" );
            return Error::UNSUPPORTED_RENDERER;
        }
    }

    mainContext2 = SDL_GL_CreateContext( mainWindow );
    ErrorCheckSDL();

    SDL_ShowWindow( mainWindow );
    SDL_GL_SetSwapInterval( 1 );

    if (glxwInit())
    {
        Log::Print( "OpenWindow: Could not init glxw!" );
        return Error::UNSUPPORTED_RENDERER;
    }

#if AETHER3D_GL43 && DEBUG
    GLint v;
    glGetIntegerv( GL_CONTEXT_FLAGS, &v );
    if (v & GL_CONTEXT_FLAG_DEBUG_BIT)
    {
        glDebugMessageCallback( DebugCallbackARB, nullptr );
        glEnable( GL_DEBUG_OUTPUT_SYNCHRONOUS );
    }
#endif

    return Error::NO_ERRORS;
}
#endif

void RendererImpl::Resize( int width, int height )
{
    screenSize[ 0 ] = width;
    screenSize[ 1 ] = height;
    
    glViewport( 0, 0, width, height );
    builtinTargets.depthNormalTarget.CreateRenderTexture2D( Width(), Height(), TextureImpl::Format::HalfFloat, TextureImpl::Filter::Point, TextureImpl::Tiling::Clamp, "builtin depth-normal rt" );
    builtinTargets.hdrTarget.CreateRenderTexture2D( Width(), Height(), TextureImpl::Format::HalfFloat, TextureImpl::Filter::Point, TextureImpl::Tiling::Clamp, "builtin hdr rt" );
}

Error::Enum BuiltinShaders::Load( const std::string& builtinAssetShaderPath )
{
    Error::Enum error = Error::NO_ERRORS;

    unlitShader.Load( builtinAssetShaderPath + "unlit.vsh", builtinAssetShaderPath + "unlit.fsh", "", "", error );
    skyboxShader.Load( builtinAssetShaderPath + "cubemap.vsh", builtinAssetShaderPath + "cubemap.fsh", "", "", error );
    depthNormalShader.Load( builtinAssetShaderPath + "depth.vsh", builtinAssetShaderPath + "depth.fsh", "", "DEPTH", error );
    depthNormalSkinShader.Load( builtinAssetShaderPath + "depth.vsh", builtinAssetShaderPath + "depth.fsh", "", "DEPTH SKIN", error );
    toneMapShader.Load( builtinAssetShaderPath + "unlit.vsh", builtinAssetShaderPath + "tonemap.fsh", "", "", error );
    blurShader.Load( builtinAssetShaderPath + "blur.vsh", builtinAssetShaderPath + "blur.fsh", "", "", error );
    blurSSAOShader.Load( builtinAssetShaderPath + "unlit.vsh", builtinAssetShaderPath + "blur_ssao.fsh", "", "", error );
    ssaoShader.Load( builtinAssetShaderPath + "ssao.vsh", builtinAssetShaderPath + "ssao.fsh", "", "", error );
    momentsShader.Load( builtinAssetShaderPath + "depth.vsh", builtinAssetShaderPath + "moments.fsh", "", "", error );
    momentsSkinShader.Load( builtinAssetShaderPath + "depth.vsh", builtinAssetShaderPath + "moments.fsh", "", "SKIN", error );
#if AETHER3D_GL43
    cullLightShader.Load( (RendererImpl::Instance().BuiltinAssetPath() + "shaders/glsl430/cull_lights.comp").c_str(), error );
#endif
    return error;
}

void RendererImpl::Deinit()
{
#if AETHER3D_GL43
    forwardPlus.Deinit();
#endif
}

const char* GetFullPath( const char* fileName );

Error::Enum RendererImpl::LoadBuiltinAssets()
{
    ErrorCheck( "LoadBuiltinAssets begin" );

    Global::glErrors[ GL_OUT_OF_MEMORY ] = "GL_OUT_OF_MEMORY";
    Global::glErrors[ GL_INVALID_ENUM ] = "GL_INVALID_ENUM";
    Global::glErrors[ GL_INVALID_VALUE ] = "GL_INVALID_VALUE";
    Global::glErrors[ GL_INVALID_OPERATION ] = "GL_INVALID_OPERATION";
    Global::glErrors[ GL_INVALID_FRAMEBUFFER_OPERATION ] = "GL_INVALID_FRAMEBUFFER_OPERATION";

#if AETHER3D_GL32 || AETHER3D_GL43
    const std::string dir = "/shaders/glsl150/";
#endif
#if AETHER3D_GLES3
    const std::string dir = "/shaders/gles3/";
#endif
    std::string newPath = builtinAssetPath + dir;
    AddShaderSearchPath( newPath.c_str() );
#if AETHER3D_GL43
    std::string newPath2 = builtinAssetPath;
    newPath2 += "/shaders/glsl430";
    AddShaderSearchPath( newPath2.c_str() );
#endif
    Error::Enum error = builtinShaders.Load( builtinAssetPath + dir );

    if (error != Error::NO_ERRORS)
    {
        return error;
    }

    MaterialImpl::Default().SetShaders( builtinAssetPath + dir + "unlit.vsh", builtinAssetPath + dir + "unlit.fsh", "", "" );

    const std::string fontPath( builtinAssetPath + "textures/font.png" );
    const std::string metaPath( builtinAssetPath + "textures/font.fnt" );
    
    error = fonts[ "aether3d_font" ].Load( fontPath.c_str(), metaPath.c_str() );

    if (error != Error::NO_ERRORS)
    {
        return error;
    }

    noiseTexture.CreateNoise( 4 );
    const std::string whitePath = builtinAssetPath + "textures/white.png";
    whiteTexture.Load( GetFullPath( whitePath.c_str() ), true, true );
    
    ErrorCheck( "LoadBuiltinAssets end" );
    return error;
}

Error::Enum RendererImpl::OpenWindow( int width, int height, bool fullScreen, int aAntialiasSamples, bool createWindowAndContext )
{
    (void)fullScreen;

    if (width < 0 || height < 0)
    {
        Log::Print( "OpenWindow: Invalid width or height, defaulting to 640x400." );
        width = 640;
        height = 400;
    }

    antialiasSamples = aAntialiasSamples;
    screenSize[ 0 ] = width;
    screenSize[ 1 ] = height;
    hasWindowAndContext = createWindowAndContext;
    
    if (createWindowAndContext)
    {
#if AETHER3D_GLES3 && !AETHER3D_IOS
        if (!CreateAppWindow( width, height, screenSize ))
        {
            return Error::UNSUPPORTED_RENDERER;
        }

        if (!CreateContext())
        {
            return Error::UNSUPPORTED_RENDERER;
        }
#endif
#if AETHER3D_GL32 || AETHER3D_GL43
        Error::Enum ok = CreateAppWindowSDL( width, height, fullScreen, antialiasSamples, screenSize );
        if (ok != Error::NO_ERRORS)
        {
            return ok;
        }
#endif
    }
#if AETHER3D_GL32 || AETHER3D_GL43
    else if (glxwInit())
    {
        Log::Print( "OpenWindow: Could not init glxw!" );
        return Error::UNSUPPORTED_RENDERER;
    }

    int contextMajor, contextMinor;
    glGetIntegerv( GL_MAJOR_VERSION, &contextMajor );
    glGetIntegerv( GL_MINOR_VERSION, &contextMinor );
    Log::Print( "Created OpenGL %d.%d context", contextMajor, contextMinor );

    if (!HasExtension( "GL_ARB_explicit_attrib_location" ))
    {
        Log::Print( "Required OpenGL extension not found: GL_ARB_explicit_attrib_location" );
        return Error::UNSUPPORTED_RENDERER;
    }

    glEnable( GL_FRAMEBUFFER_SRGB );
    glEnable( GL_TEXTURE_CUBE_MAP_SEAMLESS );
#endif
    glEnable( GL_CULL_FACE );
    glEnable( GL_DEPTH_TEST );

    int fboId = 0;
    glGetIntegerv( GL_FRAMEBUFFER_BINDING, &fboId );
    Global::systemFBO = static_cast< unsigned >(fboId);

#if !AETHER3D_IOS
    if (antialiasSamples > 1)
    {
        glEnable( GL_MULTISAMPLE );
    }
#endif
    
    const GLubyte* renderer = glGetString( GL_RENDERER );
    Log::Print( "Renderer: %s", renderer );

    CreateSkybox( skyboxBuffer );
    CreateQuadBuffer( quadBuffer );
    CreateTimerQueries();
    CreateRenderTargets();

    const Error::Enum success = LoadBuiltinAssets();

    if (success != Error::NO_ERRORS)
    {
        Log::Print( "LoadBuiltinAssets failed!" );
        return success;
    }

    const int kernelSize = 16;
    std::vector< Vec4 > kernel = CreateSSAOKernel( kernelSize );
    ErrorCheck( "SSAO uniform set begin." );
    builtinShaders.ssaoShader.Use();
    builtinShaders.ssaoShader.SetInt( "uKernelSize", kernelSize );
    builtinShaders.ssaoShader.SetVectors( "uKernelOffsets", &kernel[ 0 ], (int)kernel.size() );
    ErrorCheck( "SSAO uniform set end." );

#if AETHER3D_GL43
    forwardPlus.Init();
#endif

    return success;
}

void RendererImpl::ResolveMSAA()
{
#if !AETHER3D_GLES3
    ErrorCheck( "ResolveMSAA begin " );
    glBindFramebuffer( GL_DRAW_FRAMEBUFFER, Global::systemFBO );   // Make sure no FBO is set as the draw framebuffer
    glBindFramebuffer( GL_READ_FRAMEBUFFER, builtinTargets.hdrTarget.FBO() ); // Make sure your multisampled FBO is the read framebuffer
    glDrawBuffer( GL_BACK );                       // Set the back buffer as the draw buffer
    glBlitFramebuffer( 0, 0, screenSize[ 0 ], screenSize[ 1 ], 0, 0, screenSize[ 0 ], screenSize[ 1 ], GL_COLOR_BUFFER_BIT, GL_NEAREST );
    GfxDevice::Instance().SetRenderTarget( nullptr, 0 );
    ErrorCheck( "ResolveMSAA end " );
#endif
}

void RendererImpl::SetWindowTitle( const char* title ) const
{
    (void)title;
#if !AETHER3D_GLES3
    if (hasWindowAndContext)
    {
        SDL_SetWindowTitle( mainWindow, title );
    }
#endif
}

#if AETHER3D_GL32 || AETHER3D_GL43 || (AETHER3D_GLES3 && !AETHER3D_IOS && !AETHER3D_WINDOWS)
void RendererImpl::SwapBuffers()
{
    Statistics::EndFrame();

    static int frame = 0;
    ++frame;

    if ((frame % 60) == 0)
    {
        FileWatcher::Instance().Poll();
    }

#if AETHER3D_GLES3
    EGLContextSwapBuffers();
#else
    //CheckSDLErrorAndDie2();
    SDL_Event e;

    while (SDL_PollEvent( &e ))
    {
        if (e.type == SDL_WINDOWEVENT)
        {
            if (e.window.event == SDL_WINDOWEVENT_CLOSE)
            {
                SDL_Quit();
                exit( 0 );
            }
            else if (e.window.event == SDL_WINDOWEVENT_ENTER)
            {
                hasFocus = true;
            }	
            else if (e.window.event == SDL_WINDOWEVENT_LEAVE)
            {
                hasFocus = false;
            }
        }
    }

    if (hasWindowAndContext)
    {
        if (antialiasSamples > 1)
        {
            ResolveMSAA();
        }

        SDL_GL_SwapWindow( mainWindow );
    }
#endif
}
#endif
#if AETHER3D_IOS
void RendererImpl::SwapBuffers()
{
    Statistics::EndFrame();
    
    static int frame = 0;
    ++frame;
}
#endif

void RendererImpl::CreateRenderTargets()
{
    const int bloomBlurSize = 512;
    
    builtinTargets.depthNormalTarget.CreateRenderTexture2D( Width(), Height(), TextureImpl::Format::Float, TextureImpl::Filter::Point, TextureImpl::Tiling::Clamp, "builtin depth-normal rt" );
    builtinTargets.hdrTarget.CreateRenderTexture2D( Width(), Height(), TextureImpl::Format::HalfFloat, TextureImpl::Filter::Point, TextureImpl::Tiling::Clamp, "builtin hdr rt" );
    GfxDevice::Instance().SetClearColor( 0, 0, 0 );
    builtinTargets.bloomBlurHTarget.CreateRenderTexture2D( bloomBlurSize, bloomBlurSize, TextureImpl::Format::UnsignedByte, TextureImpl::Filter::Linear, TextureImpl::Tiling::Clamp, "builtin bloom blur h rt" );
    builtinTargets.bloomBlurVTarget.CreateRenderTexture2D( bloomBlurSize, bloomBlurSize, TextureImpl::Format::UnsignedByte, TextureImpl::Filter::Linear, TextureImpl::Tiling::Clamp, "builtin bloom blur v rt" );
    GfxDevice::Instance().SetClearColor( 1, 1, 1 );
    builtinTargets.ssaoTarget.CreateRenderTexture2D( Width(), Height(), TextureImpl::Format::UnsignedByte, TextureImpl::Filter::Point, TextureImpl::Tiling::Clamp, "builtin ssao rt" );
    builtinTargets.ssaoBlurTarget.CreateRenderTexture2D( Width() / 2, Height() / 2, TextureImpl::Format::UnsignedByte, TextureImpl::Filter::Linear, TextureImpl::Tiling::Clamp, "builtin ssao blur rt" );
    builtinTargets.shadowBlurHTarget.CreateRenderTexture2D( 512, 512, TextureImpl::Format::Float, TextureImpl::Filter::Linear, TextureImpl::Tiling::Clamp, "builtin shadow blur h rt" );
    builtinTargets.shadowBlurVTarget.CreateRenderTexture2D( 512, 512, TextureImpl::Format::Float, TextureImpl::Filter::Linear, TextureImpl::Tiling::Clamp, "builtin shadow blur v rt" );
}

void RendererImpl::CreateTimerQueries()
{
#if !AETHER3D_GLES3
    if (Global::queryIds[ 0 ] == 0)
    {
        glGenQueries( 2, Global::queryIds );
    }
#endif
}

void RendererImpl::BeginQuery()
{
#if !AETHER3D_GLES3
    glQueryCounter( Global::queryIds[ 0 ], GL_TIMESTAMP );
#endif
}

double RendererImpl::EndQuery()
{
#if !AETHER3D_GLES3
    int done = 0;

    glQueryCounter( Global::queryIds[ 1 ], GL_TIMESTAMP );

    while (!done)
    {
        glGetQueryObjectiv( Global::queryIds[ 1 ], GL_QUERY_RESULT_AVAILABLE, &done );
    }

    GLuint64 timerStart, timerEnd;
    glGetQueryObjectui64v( Global::queryIds[ 0 ], GL_QUERY_RESULT, &timerStart );
    glGetQueryObjectui64v( Global::queryIds[ 1 ], GL_QUERY_RESULT, &timerEnd );

    return (timerEnd - timerStart) / 1000000.0;
#else
    return 0;
#endif
}
