#include "Shader.hpp"
#include <fstream>
#include <sstream>
#include <unordered_map>
#if AETHER3D_IOS
#    include <OpenGLES/ES3/glext.h>
#    define GL_GEOMETRY_SHADER 0
#elif AETHER3D_GLES3
#    include <GLES3/gl3.h>
#    define GL_GEOMETRY_SHADER 0
#else
#    include "GL/glxw.h"
#endif
#include "Log.hpp"
#include "FileWatcher.hpp"
#include "Matrix.hpp"
#include "Statistics.hpp"
#include "TextureImpl.hpp"
#include "Vec3.hpp"

const char* GetFullPath( const char* fileName );
bool HasExtension( const std::string& glExtension );
void ErrorCheck( const std::string& info );

namespace ae3d
{
    GLuint sCachedProgram = 0;
}

// Checks that uniform type is as expected.
// FIXME: Causes GLES3/Linux to not render anything but the background color!
//        Works in GLES3/Windows, but disabled in GLES3 just to be sure [2014-07-26]
bool IsUniformType( GLuint program, const std::string& name, GLenum aType )
{
    (void)program;
    (void)name;
    (void)aType;
#if !AETHER3D_GLES3
    GLsizei length;
    GLint size;
    GLenum type;
    GLuint index;
    std::string nameStr = name;
    GLchar* nam = const_cast< GLchar* >( nameStr.c_str() );
    glGetUniformIndices( program, 1, &nam, &index );
    
    if (static_cast<GLint>(index) != -1)
    {
        glGetActiveUniform( program, index, 1, &length, &size, &type, nam );
        
        if (type != aType)
        {
            return false;
        }
    }
#endif
    return true;
}

struct Shader::Impl
{
    /// Id is shader's source paths appended together.
    static std::unordered_map< std::string, Shader > sIdToShaderCache;
    int UniformLocation( const std::string& name );

    GLuint program = 0;
    // Used for caching and hot-loading.
    std::string vertexPath;
    std::string fragmentPath;
    std::string geometryPath;
    std::string defines;
    std::unordered_map< std::string, int > uniformNameToLocation;
};

std::unordered_map< std::string, Shader > Shader::Impl::sIdToShaderCache;

void PrintInfoLog( GLuint shader, bool isProgram )
{
    GLint logLength = 0;

    if (isProgram)
    {
        glGetProgramiv( shader, GL_INFO_LOG_LENGTH, &logLength );
    }
    else
    {
        glGetShaderiv( shader, GL_INFO_LOG_LENGTH, &logLength );
    }
    if (logLength > 0)
    {
        GLchar* log = new GLchar[ (std::size_t)logLength + 1 ];
        GLsizei charsWritten = 0;

        if (isProgram)
        {
            glGetProgramInfoLog( shader, logLength, &charsWritten, log );
        }
        else
        {
            glGetShaderInfoLog( shader, logLength, &charsWritten, log );
        }

        if (charsWritten > 0)
        {
            Log::Print( log );
        }

        delete[] log;
    }
}

/**
   \param path Shader's path
   \param shaderType Shader type
   \param defines Defines separated by space
 */
GLuint Compile( const char* path, GLenum shaderType, const std::string& defines )
{
    ErrorCheck( "Compile begin" );

    std::ifstream stream( GetFullPath( path ) );
    if (!stream.is_open())
    {
        Log::Print( "Shader::Compile: Could not find %s", GetFullPath( path ) );
        return 0;
    }

    const std::string source = std::string( std::istreambuf_iterator< char >( stream ),
                                      std::istreambuf_iterator< char >() );
    std::stringstream sourceStm( source );
    std::string sourceWithDefines;

    // Adds #defines into the source code.
    while (!sourceStm.eof())
    {
        std::string line;
        std::getline( sourceStm, line );
        sourceWithDefines += line + "\n";

        if (!defines.empty() && line.find( "#version" ) != std::string::npos)
        {
            std::stringstream defineStm( defines );

            while (!defineStm.eof())
            {
                std::string define;
                defineStm >> define;
                define = std::string( "#define " ) + define;
                sourceWithDefines += define + "\n";
            }
        }
    }

    const char* finalSource = sourceWithDefines.c_str();
    const GLuint shader = glCreateShader( shaderType );
    glShaderSource( shader, 1, &finalSource, nullptr );
    glCompileShader( shader );

    // Checks for and reports errors.
    GLint compiled;

    glGetShaderiv( shader, GL_COMPILE_STATUS, &compiled );

    if (!compiled)
    {
        Log::Print( "Shader compile error in %s", path );
        PrintInfoLog( shader, false );
        return 0;
    }

    ErrorCheck( "Compile end" );
    return shader;
}

unsigned Shader::ActiveGLProgram()
{
    return ae3d::sCachedProgram;
}

int Shader::Impl::UniformLocation( const std::string& name )
{
    const auto it = uniformNameToLocation.find( name );
    
    if (it == uniformNameToLocation.end())
    {
        uniformNameToLocation[ name ] = glGetUniformLocation( program, name.c_str() );
        return uniformNameToLocation[ name ];
    }

    return it->second;
}

Shader::Shader() : m( new Impl ) {}

Shader::Shader( const Shader& other ) : m( new Impl )
{
    *this = other;
}

Shader::~Shader() {}

Shader& Shader::operator=( const Shader& other )
{
    m->program = other.m->program;
    m->vertexPath = other.m->vertexPath;
    m->fragmentPath = other.m->fragmentPath;
    m->geometryPath = other.m->geometryPath;
    m->defines = other.m->defines;
    return *this;
}

bool Shader::Equals( const Shader& other ) const
{
    return m->program == other.m->program;
}

unsigned Shader::GLSLProgram() const
{
    return m->program;
}

void Shader::Deinit()
{
}

void Shader::Load( const std::string& vertexShaderPath,
                   const std::string& fragmentShaderPath,
                   const std::string& geometryShaderPath,
                   const std::string& defines, Error::Enum& outError )
{
    ErrorCheck( "Load begin" );

    if (m->program == 0 &&
        Shader::Impl::sIdToShaderCache.find( vertexShaderPath + fragmentShaderPath + geometryShaderPath + defines ) != Shader::Impl::sIdToShaderCache.end())
    {
        *this = Shader::Impl::sIdToShaderCache[ vertexShaderPath + fragmentShaderPath + geometryShaderPath + defines ];
        outError = Error::NO_ERRORS;
        return;
    }

    const GLuint vertexShader = Compile( vertexShaderPath.c_str(), GL_VERTEX_SHADER, defines );
    if (vertexShader == 0)
    {
        outError = Error::VERTEX_SHADER_COMPILE_FAILED;
        return;
    }

    const GLuint fragmentShader = Compile( fragmentShaderPath.c_str(), GL_FRAGMENT_SHADER, defines );
    if (fragmentShader == 0)
    {
        outError = Error::FRAGMENT_SHADER_COMPILE_FAILED;
        return;
    }

    GLuint program = glCreateProgram();
    glAttachShader( program, vertexShader );
    glAttachShader( program, fragmentShader );

#if !AETHER3D_GLES3
    if (HasExtension( "GL_KHR_debug" ))
    {
        glObjectLabel( GL_PROGRAM, program, -1, vertexShaderPath.c_str() );
    }
#endif

    if (!geometryShaderPath.empty())
    {
        const GLuint geometryShader = Compile( geometryShaderPath.c_str(), GL_GEOMETRY_SHADER, defines );
        if (geometryShader == 0)
        {
            outError = Error::GEOMETRY_SHADER_COMPILE_FAILED;
            return;
        }

        glAttachShader( program, geometryShader );
    }

    glLinkProgram( program );
    GLint wasLinked;
    glGetProgramiv( program, GL_LINK_STATUS, &wasLinked );

    if (!wasLinked)
    {
        PrintInfoLog( program, true );
        outError = Error::SHADER_LINK_FAILED;
        return;
    }

    ErrorCheck( "Load end" );

    const bool firstTimeLoad = m->program == 0;
    if (!firstTimeLoad)
    {
        glDeleteProgram( m->program );
    }

    m->vertexPath = vertexShaderPath;
    m->fragmentPath = fragmentShaderPath;
    m->geometryPath = geometryShaderPath;
    m->program = program;
    m->defines = defines;

    if (firstTimeLoad)
    {
        FileWatcher::Instance().AddShader( vertexShaderPath.c_str(), fragmentShaderPath.c_str(), geometryShaderPath.c_str(), defines.c_str() );
        Shader::Impl::sIdToShaderCache[ vertexShaderPath + fragmentShaderPath + geometryShaderPath + defines ] = *this;
    }

    ErrorCheck( "Load end" );
    return;
}

Error::Enum Shader::Reload( const std::string& vertexShaderPath,
                            const std::string& fragmentShaderPath,
                            const std::string& geometryShaderPath,
                            const std::string& defines )
{
    bool found = false;
    Shader* original = nullptr;

    if (Shader::Impl::sIdToShaderCache.find( vertexShaderPath + fragmentShaderPath + geometryShaderPath + defines ) != Shader::Impl::sIdToShaderCache.end())
    {
        original = &(Shader::Impl::sIdToShaderCache[ vertexShaderPath + fragmentShaderPath + geometryShaderPath + defines ]);
        found = true;
    }

    if (!found)
    {
        Log::Print( "Warning: Trying to reload a shader that is not loaded." );
        return Error::NO_ERRORS;
    }

    Shader reloaded = *original;
    Error::Enum error = Error::NO_ERRORS;
    reloaded.Load( vertexShaderPath, fragmentShaderPath, geometryShaderPath, defines, error );
    if (error != Error::NO_ERRORS)
    {
        Log::Print( "Warning: Tried to reload shader %s or %s but compilation failed.", vertexShaderPath.c_str(), fragmentShaderPath.c_str() );
        return Error::VERTEX_SHADER_COMPILE_FAILED;
    }

    *original = reloaded;

    return Error::NO_ERRORS;
}

void Shader::SetConstantBuffers()
{
    // Constant buffers are not used in GL renderer.
}

void Shader::Validate()
{
    // FIXME: Enable for validating renderer state before drawing. Disabled by default because on AMD it spams "Validation successful" messages.

    /*GLint isValidated = 0;
    glValidateProgram( Shader::ActiveGLProgram() );
    glGetProgramiv( Shader::ActiveGLProgram(), GL_VALIDATE_STATUS, &isValidated );

    if (isValidated != 0)
    {
        GLint logLength = 0;
        glGetProgramiv( Shader::ActiveGLProgram(), GL_INFO_LOG_LENGTH, &logLength );

        if (logLength > 0)
        {
            GLchar* log = new GLchar[ (std::size_t)logLength + 1 ];
            GLsizei charsWritten = 0;

            glGetProgramInfoLog( Shader::ActiveGLProgram(), logLength, &charsWritten, log );

            if (charsWritten > 0)
            {
                Log::Print( log );
            }

            delete[] log;
        }
    }*/
}

void Shader::SetFloat( const std::string& name, float value )
{
#if DEBUG
    if (!IsUniformType( m->program, name, GL_FLOAT ))
    {
        Log::Print( "%s: Tried to set float %s, but uniform type is not float!", m->fragmentPath.c_str(), name.c_str() );
        return;
    }
#endif
    glUniform1f( m->UniformLocation( name ), value );
}

void Shader::SetInt( const std::string& name, int value )
{
    // Commented out because setting texture causes this check to fail. [2014-07-21]
/*#if DEBUG
    if (!IsUniformType( m->program, name, GL_INT ))
    {
        Log::Print( "%s: Tried to set int %s, but uniform type is not int!", m->fragmentPath.c_str(), name.c_str() );
        return;
    }
#endif*/
    glUniform1i( m->UniformLocation( name ), value );
}

void Shader::SetUInt( const std::string& name, unsigned value )
{
#if DEBUG
    if (!IsUniformType( m->program, name, GL_UNSIGNED_INT ))
    {
        Log::Print( "%s: Tried to set unsigned int %s, but uniform type is not unsigned int!", m->fragmentPath.c_str(), name.c_str() );
        return;
    }
#endif

    glUniform1ui( m->UniformLocation( name ), value );
}

void Shader::SetMatrix( const std::string& name, const Matrix44& matrix )
{
#if DEBUG
    if (!IsUniformType( m->program, name, GL_FLOAT_MAT4 ))
    {
        Log::Print( "%s: Tried to set matrix %s, but uniform type is not mat4!", m->fragmentPath.c_str(), name.c_str() );
        return;
    }
#endif
    glUniformMatrix4fv( m->UniformLocation( name ), 1, GL_FALSE, matrix.m );
}

void Shader::SetMatrices( const std::string& name, const Matrix44* matrices, int count )
{
    glUniformMatrix4fv( m->UniformLocation( name ), count, GL_FALSE, matrices[0].m );
}

void Shader::SetTexture( const std::string& textureName, const TextureImpl* texture, int textureUnit )
{
    if (!texture)
    {
        Log::Print( "%s: Tried to set texture into shader slot %s, but the texture is null!", FragmentPath().c_str(), textureName.c_str() );
        return;
    }
    
    // unit 666 is used in D3D11 renderer to unbind textures.
    if (textureUnit == 666)
    {
        return;
    }
#if DEBUG
    if (texture->IsCubemap() && !IsUniformType( m->program, textureName, GL_SAMPLER_CUBE ))
    {
        Log::Print( "Texture %s is a cubemap but the sampler it's trying to use in shader %s is not!", textureName.c_str(), m->fragmentPath.c_str() );
    }
    else if (!texture->IsCubemap() && !IsUniformType( m->program, textureName, GL_SAMPLER_2D ))
    {
        Log::Print( "Texture %s is not a cubemap but the sampler it's trying to use in shader %s is!", textureName.c_str(), m->fragmentPath.c_str() );
    }
#endif
    ErrorCheck( "SetTexture begin" );
    texture->Bind( textureUnit );
    SetVector( textureName + "_ST", texture->ScaleOffset() );
    SetInt( textureName, textureUnit );

    ErrorCheck( "SetTexture end" );
}

void Shader::SetVector( const std::string& name, const Vec3& value )
{
#if DEBUG
    if (!IsUniformType( m->program, name, GL_FLOAT_VEC3 ))
    {
        Log::Print( "%s: Tried to set vec3 %s, but uniform type is not vec3!", m->fragmentPath.c_str(), name.c_str() );
        return;
    }
#endif
    glUniform3fv( m->UniformLocation( name ), 1, &value.x );
}

void Shader::SetVector( const std::string& name, const Vec4& value )
{
#if DEBUG
    if (!IsUniformType( m->program, name, GL_FLOAT_VEC4 ))
    {
        Log::Print( "%s: Tried to set vec4 %s, but uniform type is not vec4!", m->fragmentPath.c_str(), name.c_str() );
        return;
    }
#endif
    glUniform4fv( m->UniformLocation( name ), 1, &value.x );
}

void Shader::SetVectors( const std::string& name, const Vec3* values, int count )
{
    glUniform3fv( m->UniformLocation( name ), count, &values[ 0 ].x );
}

void Shader::SetVectors( const std::string& name, const Vec4* values, int count )
{
    glUniform4fv( m->UniformLocation( name ), count, &values[ 0 ].x );
}

void Shader::UnbindTexture( const char* /*name*/ )
{
    // Not needed in GL renderer.
}

void Shader::UnbindTextures()
{
    // Not needed for GL renderer. TODO: Check with a GL profiler if RTT feedback reduces performance/produces warnings.
}

void Shader::Use()
{
    GLuint program = Shader::Impl::sIdToShaderCache[ m->vertexPath + m->fragmentPath + m->geometryPath + m->defines ].m->program;
    Log::Assert( program != 0, "Uninitialized shader." );

    if (m->program != program)
    {
        m->program = program;
        m->uniformNameToLocation.clear();
    }

    if (m->program != ae3d::sCachedProgram)
    {
        ae3d::sCachedProgram = m->program;
        glUseProgram( ae3d::sCachedProgram );

        Statistics::IncShaderSwitches();
    }
}

const std::string& Shader::VertexPath() const
{
    return m->vertexPath;
}

const std::string& Shader::FragmentPath() const
{
    return m->fragmentPath;
}

const std::string& Shader::GeometryPath() const
{
    return m->geometryPath;
}
