#ifndef TEXTURE_IMPL_H
#define TEXTURE_IMPL_H

#ifndef TEXTURE_H
#    include "Texture.hpp"
#endif
#ifndef VEC3_H
#    include "Vec3.hpp"
#endif
#ifndef INCLUDE_STRING_H
#    include <string>
#    define INCLUDE_STRING_H
#endif

#if AETHER3D_D3D11
struct ID3D11ShaderResourceView;
struct ID3D11RenderTargetView;
struct ID3D11UnorderedAccessView;
struct ID3D11DepthStencilView;
#endif

class TextureImpl : public Texture
{
  public:
    static const int sMaxTextureUnit = 16;
    enum class Format { UnsignedByte, HalfFloat, HalfFloatRG, Float };
    
    struct Sampler
    {
        Filter filter;
        Tiling tiling;
    };

    /**
       Reloads a texture if it exists.
     
       \param path Path.
     */
    static void Reload( const std::string& path );
    TextureImpl();
    TextureImpl( const char* path, bool useMipmaps, bool sRGB, Tiling tiling, Filter filter );
    TextureImpl( const std::string& atlasPath, const std::string& atlasMetadataPath, const std::string& fileName, bool sRGB );
    TextureImpl( const std::string& left,  const std::string& right,
                 const std::string& top,   const std::string& bottom,
                 const std::string& front, const std::string& back );
    virtual ~TextureImpl();

    int Height() const override { return height; }
    int Width() const override { return width; }

    void Bind( int targetUnit ) const;
    void CreateRenderTexture2D( int width, int height, Format format, Filter filter, Tiling tiling, const std::string& debugName );
    void CreateRenderTextureCube( int size );
    void CreateNoise( int size );
    bool HasMipmaps() const { return hasMipmaps; }
#if AETHER3D_D3D11
    ID3D11ShaderResourceView* SRV() const;
    ID3D11RenderTargetView* RTV() const;
    ID3D11UnorderedAccessView* UAV() const;
    ID3D11ShaderResourceView* DepthSRV() { return depthSrv; }
    ID3D11DepthStencilView* DepthStencilView() { return depthStencilView; }
#else
    unsigned FBO() const { return fboId; }
    unsigned Id() const { return id; }
#endif
    bool IsCubemap() const { return cubeMap; }
    bool IsSRGB() const { return sRGB; }
    bool Opaque() const { return opaque; }
    const Sampler& GetSampler() const { return sampler; }
    const Vec4 ScaleOffset() const { return scaleOffset; }
    bool Load( const std::string& path, bool useMipmaps, bool sRGB );
    bool Load( const std::string& left,  const std::string& right,
              const std::string& top,   const std::string& bottom,
              const std::string& front, const std::string& back,
              bool useMipmaps, bool usesRGB );

  private:

    int width = 0, height = 0;
#if !AETHER3D_D3D11
    unsigned id = 0;
    unsigned fboId = 0; // FBO id for render texture.
    unsigned rboId = 0; // RBO id for render texture.
#endif
    // Used to get coordinates from an atlas.  If no atlas, degenerates to (1, 1, 0, 0).
    Vec4 scaleOffset;
    bool opaque;
    bool sRGB;
    bool hasMipmaps;
    bool cubeMap; // True if the texture is a cube map, false if it's 2D.
    Sampler sampler;
#if AETHER3D_D3D11
    ID3D11ShaderResourceView* srv = nullptr;
    ID3D11RenderTargetView* rtv = nullptr;
    ID3D11UnorderedAccessView* uav = nullptr;
    ID3D11ShaderResourceView* depthSrv = nullptr;
    ID3D11DepthStencilView* depthStencilView = nullptr;
#endif
};

#endif
