#include "VertexBuffer.hpp"
#include <vector>
#if AETHER3D_IOS
#    include <OpenGLES/ES3/glext.h>
#elif AETHER3D_GLES3
#    include <GLES3/gl3.h>
#else
#    include <GL/glxw.h>
#endif
#include "Log.hpp"
#include "Mesh.hpp"
#include "Shader.hpp"
#include "Statistics.hpp"

void PrintInfoLog( GLuint shader, bool isProgram );
void ErrorCheck( const std::string& info );
bool HasExtension( const std::string& glExtension );

struct VertexBuffer::Impl
{
    void GenBaseMeshBuffer( const std::vector< Face >& indices,
                            const std::vector< Vertex >& vertices );
    
    GLuint vboId = 0;
    GLuint vaoId = 0;
    GLuint iboId = 0;
    int elementCount = 0;
};

void VertexBuffer::Impl::GenBaseMeshBuffer( const std::vector< Face >& indices,
                                            const std::vector< Vertex >& vertices )
{
    elementCount = (int)indices.size() * 3;

    glDeleteBuffers( 1, &iboId );
    glDeleteBuffers( 1, &vboId );
    glDeleteVertexArrays( 1, &vaoId );

    glGenVertexArrays( 1, &vaoId );
    glBindVertexArray( vaoId );

    glGenBuffers( 1, &vboId );
    glBindBuffer( GL_ARRAY_BUFFER, vboId );
    glBufferData( GL_ARRAY_BUFFER, (GLsizei)vertices.size() *
                 sizeof( Vertex ), &vertices[0], GL_STATIC_DRAW );

    glGenBuffers( 1, &iboId );
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, iboId );
    glBufferData( GL_ELEMENT_ARRAY_BUFFER, (GLsizei)indices.size() *
                 sizeof(unsigned short) * 3, &indices[0], GL_STATIC_DRAW );

    // Vertex
    glEnableVertexAttribArray( Shader::Channel::POSITION );
    glVertexAttribPointer( Shader::Channel::POSITION, 3, GL_FLOAT, GL_FALSE, sizeof( Vertex ), nullptr/*(const GLvoid*)(0)*/ );

    // TexCoord.
    glEnableVertexAttribArray( Shader::Channel::TEXCOORD );
    glVertexAttribPointer( Shader::Channel::TEXCOORD, 2, GL_FLOAT, GL_FALSE, sizeof( Vertex ), (const GLvoid*)(3 * 4) );

    // Normal.
    glEnableVertexAttribArray( Shader::Channel::NORMAL );
    glVertexAttribPointer( Shader::Channel::NORMAL, 3, GL_FLOAT, GL_FALSE, sizeof( Vertex ), (const GLvoid*)(5 * 4) );

    // Tangent.
    glEnableVertexAttribArray( Shader::Channel::TANGENT );
    glVertexAttribPointer( Shader::Channel::TANGENT, 4, GL_FLOAT, GL_FALSE, sizeof( Vertex ), (const GLvoid*)(8 * 4) );

    // Color.
    glEnableVertexAttribArray( Shader::Channel::COLOR );
    glVertexAttribPointer( Shader::Channel::COLOR, 4, GL_FLOAT, GL_FALSE, sizeof( Vertex ), (const GLvoid*)(12 * 4) );
    
    Statistics::IncBufferSwitches();
    Statistics::IncBufferSwitches();
}

VertexBuffer::VertexBuffer() : m( new Impl ) {}

VertexBuffer::VertexBuffer( const VertexBuffer& other ) : m( new Impl )
{
    *this = other;
}

VertexBuffer::~VertexBuffer() {}

VertexBuffer& VertexBuffer::operator=( const VertexBuffer& other )
{
    *m = *other.m;
    return *this;
}

void VertexBuffer::SetName( const char* name )
{
#if !AETHER3D_GLES3
    if (HasExtension( "GL_KHR_debug" ))
    {
        glObjectLabel( GL_BUFFER, m->vaoId, -1, name );
    }
#endif
}

bool VertexBuffer::IsGenerated() const
{
    return m->vaoId != 0;
}

void VertexBuffer::Bind() const
{
    if (m->vaoId == 0)
    {
        Log::Print( "Tried to bind an uninitialized vertex buffer." );
    }

    glBindVertexArray( m->vaoId );
    Statistics::IncBufferSwitches();
}

void VertexBuffer::Deinit()
{
}

void VertexBuffer::Draw() const
{
    if (m->vaoId == 0)
    {
        Log::Print( "Tried to draw uninitialized vertex buffer." );
        return;
    }

#if DEBUG
    Shader::Validate();
#endif

    glDrawElements( GL_TRIANGLES, m->elementCount, GL_UNSIGNED_SHORT, nullptr );
    Statistics::IncDrawCalls();
}

void VertexBuffer::DrawInstanced( int instanceCount ) const
{
#if DEBUG
    Shader::Validate();
#endif

    glDrawElementsInstanced( GL_TRIANGLES, m->elementCount, GL_UNSIGNED_SHORT, nullptr, instanceCount );
    Statistics::IncDrawCalls();
}

void VertexBuffer::Generate( const std::vector< Face >& indices,
                             const std::vector< Vertex >& vertices,
                             const std::vector< Influence >& influences )
{
    ErrorCheck( "VertexBuffer::Generate begin" );

    m->GenBaseMeshBuffer( indices, vertices );

    if (!influences.empty())
    {
        GLuint weightVBO;
        glGenBuffers( 1, &weightVBO );
        glBindBuffer( GL_ARRAY_BUFFER, weightVBO );
        glBufferData( GL_ARRAY_BUFFER, (GLsizei)vertices.size() * sizeof( Influence ), &influences[ 0 ], GL_STATIC_DRAW );

        // Joint indices.
        glEnableVertexAttribArray( Shader::Channel::BONE_INDEX );
        glVertexAttribPointer( Shader::Channel::BONE_INDEX, 4, GL_FLOAT, GL_FALSE, sizeof( Influence ), nullptr );

        // Joint weights.
        glEnableVertexAttribArray( Shader::Channel::BONE_WEIGHT );
        glVertexAttribPointer( Shader::Channel::BONE_WEIGHT, 4, GL_FLOAT, GL_FALSE, sizeof( Influence ), (const GLvoid*)(4 * 4) );
    }
    else
    {
        glDisableVertexAttribArray( Shader::Channel::BONE_INDEX );
        glDisableVertexAttribArray( Shader::Channel::BONE_WEIGHT );
    }

    glBindVertexArray( 0 );
    glBindBuffer( GL_ARRAY_BUFFER, 0 );
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
    
    ErrorCheck( "VertexBuffer::Generate end" );
}

void VertexBuffer::Generate( const std::vector< Face >& indices, const std::vector< Vertex >& vertices,
                             const std::vector< Vec3 >& morphPositions1, const std::vector< Vec3 >& morphPositions2 )
{
    ErrorCheck( "VertexBuffer::Generate morph begin" );

    m->GenBaseMeshBuffer( indices, vertices );

    GLuint morphPos1VBO;
    glGenBuffers( 1, &morphPos1VBO );
    glBindBuffer( GL_ARRAY_BUFFER, morphPos1VBO );
    glBufferData( GL_ARRAY_BUFFER, (GLsizei)morphPositions1.size() * sizeof( Vec3 ), &morphPositions1[ 0 ], GL_STATIC_DRAW );

    glEnableVertexAttribArray( Shader::Channel::MORPH_POS1 );
    glVertexAttribPointer( Shader::Channel::MORPH_POS1, 3, GL_FLOAT, GL_FALSE, sizeof( Vec3 ), nullptr );

    GLuint morphPos2VBO;
    glGenBuffers( 1, &morphPos2VBO );
    glBindBuffer( GL_ARRAY_BUFFER, morphPos2VBO );
    glBufferData( GL_ARRAY_BUFFER, (GLsizei)morphPositions2.size() * sizeof( Vec3 ), &morphPositions2[ 0 ], GL_STATIC_DRAW );

    glEnableVertexAttribArray( Shader::Channel::MORPH_POS2 );
    glVertexAttribPointer( Shader::Channel::MORPH_POS2, 3, GL_FLOAT, GL_FALSE, sizeof( Vec3 ), nullptr );

    glBindVertexArray( 0 );
    glBindBuffer( GL_ARRAY_BUFFER, 0 );
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );

    ErrorCheck( "VertexBuffer::Generate morph end" );
}
