
#include <cassert>
#include <iostream>
#include "../common.hpp"

const int gWidth = 1280;
const int gHeight = 720;

int main()
{
    Aether3D_API dll;
    Error::Enum error = dll.LoadDLL( "../Renderers/libaether3d_gl32" );

    if (error == Error::DLL_NOT_FOUND)
    {
        assert( !"Could not find engine DLL!" );
        return (int)error;
    }
    else if (error == Error::FACTORY_NOT_FOUND)
    {
        assert( !"No entry point in DLL!" );
        return (int)error;
    }
    else if (error == Error::DLL_VERSION_MISMATCH)
    {
        assert( !"DLL header version does not match client header version!" );
        return (int)error;
    }

    Renderer* renderer = dll.MakeRenderer( "../Assets/builtin/" );
    assert( renderer );

    error = renderer->OpenWindow( gWidth, gHeight, false /* fullscreen */, 1, true );
    //error = renderer->OpenWindow( 0, 0, true, 4, true );

    if (error != Error::NO_ERRORS)
    {
        std::cerr << "Could not create rendering context!" << std::endl;
        return (int)error;
    }

    std::unique_ptr< KeyboardInput > keyboardInput( dll.MakeKeyboardInput() );
    std::unique_ptr< MouseInput > mouseInput( dll.MakeMouseInput() );
    mouseInput->SetRelative( true );

    std::unique_ptr< Texture > white( dll.MakeTexture( "../Assets/builtin/textures/white.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );
    std::unique_ptr< Texture > headTexture( dll.MakeTexture( "Assets/Textures/table.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );

    std::unique_ptr< Texture > barrelD( dll.MakeTexture( "Assets/Textures/barrel.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );
    std::unique_ptr< Texture > barrelN( dll.MakeTexture( "Assets/Textures/barrel_n.png", true, false, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );

    std::unique_ptr< Model > barrel( dll.MakeModel( "Assets/Models/barrel.ae3d", error ) );
    barrel->SetPosition( Vec3( 35, -7, 0 ) );
    barrel->SetScale( 2.5f );
    
    std::unique_ptr< Model > skinnedModel( dll.MakeModel( "Assets/Models/skinnedMesh.ae3d", error ) );
    skinnedModel->SetPosition( Vec3( 25, -0.5f, 0 ) );

    std::unique_ptr< Model > morphTest( dll.MakeModel( "Assets/Models/morph.ae3d", error ) );
    morphTest->SetPosition( Vec3( 10, -2, 5 ) );

    std::unique_ptr< Material > material( dll.MakeMaterial() );
    material->SetShaders( "unlit", "unlit", "", "" );
    material->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    material->SetTexture( "textureMap", white.get() );

    std::unique_ptr< Material > skinMaterial( dll.MakeMaterial() );
    skinMaterial->SetShaders( "unlit", "unlit", "", "SKIN" );
    skinMaterial->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    skinMaterial->SetTexture( "textureMap", white.get() );

    std::unique_ptr< Material > morphMaterial( dll.MakeMaterial() );
    morphMaterial->SetShaders( "morph_test", "unlit", "", "" );
    morphMaterial->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    morphMaterial->SetTexture( "textureMap", headTexture.get() );

    skinnedModel->SetMaterialForMesh( skinMaterial.get(), "body" );
    morphTest->SetMaterialForMesh( morphMaterial.get(), "polySurfaceShape1" );

    std::unique_ptr< Scene > scene( dll.MakeScene() );
    scene->Add( skinnedModel.get() );
    scene->Add( morphTest.get() );
    scene->Add( barrel.get() );

    std::unique_ptr< Light > sunLight( dll.MakeLight( Light::Type::Directional ) );
    Vec3 lightDir = Vec3( 0.5f, -0.5f, 0.5f ).Normalized();
    sunLight->SetDirection( lightDir );
    sunLight->SetCastShadow( true, 1024 );
    scene->Add( sunLight.get() );

    std::unique_ptr< Texture > gridTexture( dll.MakeTexture( "Assets/Textures/metal_grid.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );
    std::unique_ptr< Texture > gridTextureN( dll.MakeTexture( "Assets/Textures/metal_grid_n2.png", true, false, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );
    std::unique_ptr< Texture > gridTextureS( dll.MakeTexture( "Assets/Textures/metal_grid_s.png", true, false, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );

    std::unique_ptr< Material > gridMaterial( dll.MakeMaterial() );
    gridMaterial->SetShaders( "cook_torrance", "cook_torrance", "", "SPECULAR_MAP" );
    gridMaterial->SetVector( "diffuseMap_ST", Vec4( 10, 10, 0, 0 ) );
    gridMaterial->SetTexture( "diffuseMap", gridTexture.get() );
    gridMaterial->SetTexture( "normalMap", gridTextureN.get() );
    gridMaterial->SetTexture( "specularMap", gridTextureS.get() );
    gridMaterial->SetVector( "materialSpecular", Vec3( 0.2f, 0.2f, 0.2f ) );
    gridMaterial->SetFloat( "roughness", 0.3f );

    std::unique_ptr< Material > barrelMaterial( dll.MakeMaterial() );
    barrelMaterial->SetShaders( "cook_torrance", "cook_torrance", "", "" );
    barrelMaterial->SetTexture( "diffuseMap", barrelD.get() );
    barrelMaterial->SetTexture( "normalMap", barrelN.get() );
    barrelMaterial->SetVector( "materialSpecular", Vec3( 0.2f, 0.2f, 0.2f ) );
    barrelMaterial->SetFloat( "roughness", 0.1f );

    barrel->SetMaterialForMesh( barrelMaterial.get(), "barrel_Cylinder" );
    barrel->EnableWireframe( true );

    std::unique_ptr< Texture > skybox( dll.MakeTexture( "../Assets/textures/skybox/skybox_left.jpg",
        "../Assets/textures/skybox/skybox_right.jpg",
        "../Assets/textures/skybox/skybox_top.jpg",
        "../Assets/textures/skybox/skybox_bottom.jpg",
        "../Assets/textures/skybox/skybox_front.jpg",
        "../Assets/textures/skybox/skybox_back.jpg" ) );

    std::unique_ptr< Model > cubeChild( dll.MakeModel( "../Tests/assets/models/textured_cube.ae3d", error ) );
    cubeChild->SetMaterialForMesh( gridMaterial.get(), "Cube" );
    cubeChild->SetPosition( Vec3( -3, 2, 0 ) );
    barrel->AddChild( cubeChild.get() );

    std::unique_ptr< Model > floor( dll.MakeModel( "Assets/Models/plane.ae3d", error ) );
    floor->SetScale( 10 );
    floor->SetPosition( Vec3( 0, 1, 0 ) );
    floor->SetMaterialForMesh( gridMaterial.get(), "Cube" );
    scene->Add( floor.get() );

    scene->SetSkybox( skybox.get() );

    std::unique_ptr< Camera > camera( dll.MakeCamera() );
    camera->SetProjection( 45, (float)renderer->Width() / (float)renderer->Height(), 1, 200 );
    //camera->SetViewport( 0, renderer->Width() / 2, 0, renderer->Height() );
    camera->LookAt( Vec3( 0, 0, -20 ), Vec3( 2000, -2000, 0 ), Vec3( 0, 1, 0 ) );
    scene->Add( camera.get() );

    skinnedModel->ApplyAnimation( 0, 0 );

    renderer->SetWindowTitle( "Aether3D Example: Model Loading" );
    renderer->SetClearColor( 0.2f, 0.2f, 0.5f );
    renderer->SetBloom( false, 0.25f, 2 );
    renderer->SetSSAO( true, 2.2f, 1.8f );
    
    float angle = 0;
    bool morphPlaying = false;

    Quaternion morphRotation;
    morphRotation.FromAxisAngle( Vec3( 0, 1, 0 ), 180 );
    morphTest->SetRotation( morphRotation );

    while (!keyboardInput->IsDown( ae3d::keyboardinput::KEY_ESC ))
    {
        if (!morphPlaying && keyboardInput->IsDown( ae3d::keyboardinput::KEY_r ))
        {
            morphTest->ApplyAnimation( 0, 0 );
            morphPlaying = true;
        }
        else if (morphPlaying)
        {
            morphPlaying = false;
        }

        ++angle;
        Quaternion rotation;
        const Vec3 axis( 0, 1, 0 );
        rotation.FromAxisAngle( axis, angle );
        barrel->SetRotation( rotation );

        mouseInput->SetRelative( renderer->HasFocus() );

        HandleInput( keyboardInput.get(), camera.get() );
        HandleInput( mouseInput.get(), camera.get() );

        renderer->Draw( scene.get() );
        renderer->DrawStatistics();
        renderer->Draw( "press r for blend shape animation", "aether3d_font", 1, 0, 150, Vec4( 1, 1, 1, 1 ) );
        renderer->FinishFrame();
        renderer->SwapBuffers();
    }

    renderer->Deinit();
    return 0;
}
