#import "ViewController.h"
#include "CameraImpl.hpp"
#include "MaterialImpl.hpp"
#include "ModelImpl.hpp"
#include "RendererImpl.hpp"
#include "SceneImpl.hpp"
#include "TextureImpl.hpp"

const char* GetFullPath( const char* fileName );
GLKView *globalView = NULL;

void BindDrawable()
{
    [globalView bindDrawable];
}

@interface ViewController () {
    CameraImpl camera;
    CameraImpl cubeCamera;
    CameraImpl rtCamera;
    MaterialImpl material;
    MaterialImpl cubeMaterial;
    MaterialImpl rtMaterial;
    ModelImpl model;
    ModelImpl model2;
    ModelImpl model3;
    SceneImpl scene;
    TextureImpl* texture;
    TextureImpl* renderTextureCube;
    TextureImpl* renderTexture;
}
@property (strong, nonatomic) EAGLContext *context;

- (void)setupGL;
- (void)tearDownGL;

@end

@implementation ViewController

- (void)viewDidLoad
{
    [super viewDidLoad];
    
    self.context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES3];
    
    if (!self.context)
    {
        NSLog(@"Failed to create ES context");
    }
    
    GLKView *view = (GLKView *)self.view;
    view.context = self.context;
    view.drawableDepthFormat = GLKViewDrawableDepthFormat24;
    globalView = view;
    
    [self setupGL];
}

- (void)dealloc
{
    [self tearDownGL];
    
    if ([EAGLContext currentContext] == self.context) {
        [EAGLContext setCurrentContext:nil];
    }
}

- (void)didReceiveMemoryWarning
{
    [super didReceiveMemoryWarning];
    
    if ([self isViewLoaded] && ([[self view] window] == nil)) {
        self.view = nil;
        
        [self tearDownGL];
        
        if ([EAGLContext currentContext] == self.context) {
            [EAGLContext setCurrentContext:nil];
        }
        self.context = nil;
    }
}

- (void)setupGL
{
    [EAGLContext setCurrentContext:self.context];
    
    glEnable(GL_DEPTH_TEST);
    
    RendererImpl::Instance().SetBuiltinAssetPath( "/Assets/builtin/" );
    RendererImpl::Instance().OpenWindow( self.view.bounds.size.height*2, self.view.bounds.size.width*2, true, 1, true );
    RendererImpl::Instance().SetClearColor( 1, 1, 1 );
    RendererImpl::Instance().SetBloom( true, 0.85f, 2 );
    RendererImpl::Instance().SetSSAO( false, 2, 1.6f );
    
    renderTexture = new TextureImpl();
    renderTexture->CreateRenderTexture2D( RendererImpl::Instance().Height(), RendererImpl::Instance().Width(), TextureImpl::Format::UnsignedByte, TextureImpl::Filter::Linear, TextureImpl::Tiling::Repeat, "render texture" );

    renderTextureCube = new TextureImpl();
    renderTextureCube->CreateRenderTextureCube( 512 );

    //camera.SetProjection( 55, (float)renderer->Height() / (float)renderer->Width(), 1, 200 );
    camera.SetProjection( 55, (float)self.view.bounds.size.height / (float)self.view.bounds.size.width, 1, 200 );
    camera.LookAt( Vec3( 0, 0, -20 ), Vec3( 0, 0, 2000 ), Vec3( 0, 1, 0 ) );
    scene.Add( &camera );
    
    cubeCamera.SetProjection( 45, 1, 1, 200 );
    cubeCamera.SetPosition( Vec3( 0, 0, -6 ) );
    cubeCamera.SetRenderTarget( renderTextureCube );
    scene.Add( &cubeCamera );
    
    rtCamera.SetProjection( 55, 1, 1, 200 );
    rtCamera.LookAt( Vec3( 0, 0, -20 ), Vec3( 0, 0, 2000 ), Vec3( 0, 1, 0 ) );
    rtCamera.SetRenderTarget( renderTexture );
    scene.Add( &rtCamera );
    
    texture = new TextureImpl( "/Assets/textures/checker.pvr", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear );
    
    Error::Enum error = model.LoadFromFile( "/skinnedMesh.ae3d" );
    
    if (error != Error::NO_ERRORS)
    {
        NSLog( @"Could not load model!" );
    }
    
    model.SetPosition( Vec3( 0, 6, 0 ) );

    material.SetShaders( "unlit", "unlit", "", "" );
    material.SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    material.SetTexture( "textureMap", texture );

    rtMaterial.SetShaders( "unlit", "unlit", "", "" );
    rtMaterial.SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    rtMaterial.SetTexture( "textureMap", renderTexture );
    
    cubeMaterial.SetShaders( "envmap", "envmap", "", "" );
    cubeMaterial.SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    cubeMaterial.SetTexture( "textureMap", renderTextureCube );
    
    model.SetMaterialForMesh( &material, "body" );
    scene.Add( &model );
    
    //model2.LoadFromFile( "Assets/Models/skinnedMesh.ae3d" );
    //model3.LoadFromFile( "Assets/Models/skinnedMesh.ae3d" );
}

- (void)tearDownGL
{
    [EAGLContext setCurrentContext:self.context];
}

#pragma mark - GLKView and GLKViewController delegate methods

- (void)update
{
}

- (void)glkView:(GLKView *)view drawInRect:(CGRect)rect
{
    RendererImpl& renderer = RendererImpl::Instance();
    
    renderer.ClearScreen();
    
    renderer.Draw( &scene );
    renderer.Draw( "hello world", "aether3d_font", 1, 10, 280, Vec4( 1, 1, 1, 1 ) );
    renderer.DrawStatistics();
    
    renderer.FinishFrame();
    renderer.SwapBuffers();
}

@end
