#include <cassert>
#include <iostream>
#include <vector>
#include <cmath>
#include "../common.hpp"
#include "MathUtil.hpp"

const int gWidth = 1280;
const int gHeight = 720;

int main()
{
    Aether3D_API dll;
    Error::Enum error = dll.LoadDLL( "../Renderers/libaether3d_gl32" );

    if (error == Error::DLL_NOT_FOUND)
    {
        std::cerr << "Could not find engine DLL!" << std::endl;
        return (int)error;
    }
    else if (error == Error::FACTORY_NOT_FOUND)
    {
        std::cerr << "No entry point in DLL!" << std::endl;
        return (int)error;
    }
    else if (error == Error::DLL_VERSION_MISMATCH)
    {
        std::cerr << "DLL version does not match header version!" << std::endl;
        return (int)error;
    }

    std::string assetPath = "../Assets/builtin/";
    Renderer* renderer = dll.MakeRenderer( assetPath );
    error = renderer->OpenWindow( gWidth, gHeight, false, 1, true );
    //error = renderer->OpenWindow( 0, 0, true, 1, true );

    if (error != Error::NO_ERRORS)
    {
        std::cerr << "Could not create rendering context!" << std::endl;
        renderer->Deinit();
        return (int)error;
    }

    std::unique_ptr< KeyboardInput > keyboardInput( dll.MakeKeyboardInput() );
    std::unique_ptr< MouseInput > mouseInput( dll.MakeMouseInput() );
    mouseInput->SetRelative( true );

    std::unique_ptr< Texture > asphalt( dll.MakeTexture( "Assets/Textures/asphalt.jpg", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Texture > white( dll.MakeTexture( "../Assets/builtin/textures/white.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Texture > black( dll.MakeTexture( "../Assets/builtin/textures/black.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Texture > normalTexture( dll.MakeTexture( "Assets/Textures/table_n.png", true, false, Texture::Tiling::Repeat, Texture::Filter::Linear ) );

    std::unique_ptr< Texture > skybox( dll.MakeTexture( "../Assets/textures/skybox/skybox_left.jpg",
        "../Assets/textures/skybox/skybox_right.jpg",
        "../Assets/textures/skybox/skybox_top.jpg",
        "../Assets/textures/skybox/skybox_bottom.jpg",
        "../Assets/textures/skybox/skybox_front.jpg",
        "../Assets/textures/skybox/skybox_back.jpg" ) );

    std::unique_ptr< Model > redSphere( dll.MakeModel( "Assets/Models/sphere.ae3d", error ) );
    redSphere->SetPosition( Vec3( 12, 0, 0 ) );
    redSphere->SetScale( 0.5f );

    std::unique_ptr< Model > redSphere2( dll.MakeModel( "Assets/Models/sphere.ae3d", error ) );
    redSphere2->SetPosition( Vec3( 12, 0, 0 ) );
    redSphere2->SetScale( 0.5f );

    std::unique_ptr< Material > redMaterial( dll.MakeMaterial() );
    redMaterial->SetShaders( "unlit", "unlit", "", "" );
    redMaterial->SetTexture( "textureMap", white.get() );
    redMaterial->SetVector( "tint", Vec4( 1, 0, 0, 1 ) );

    redSphere->SetMaterialForMesh( redMaterial.get(), "Sphere" );
    redSphere2->SetMaterialForMesh( redMaterial.get(), "Sphere" );

    std::unique_ptr< Model > greenSphere( dll.MakeModel( "Assets/Models/sphere.ae3d", error ) );
    greenSphere->SetPosition( Vec3( 16, 4, 0 ) );
    greenSphere->SetScale( 0.5f );

    std::unique_ptr< Model > greenSphere2( dll.MakeModel( "Assets/Models/sphere.ae3d", error ) );
    greenSphere2->SetPosition( Vec3( 16, 4, 0 ) );
    greenSphere2->SetScale( 0.5f );

    std::unique_ptr< Material > greenMaterial( dll.MakeMaterial() );
    greenMaterial->SetShaders( "unlit", "unlit", "", "" );
    greenMaterial->SetTexture( "textureMap", white.get() );
    greenMaterial->SetVector( "tint", Vec4( 0, 1, 0, 1 ) );

    greenSphere->SetMaterialForMesh( greenMaterial.get(), "Sphere" );
    greenSphere2->SetMaterialForMesh( greenMaterial.get(), "Sphere" );

    std::unique_ptr< Model > reflectiveSphere( dll.MakeModel( "Assets/Models/sphere.ae3d", error ) );
    reflectiveSphere->SetPosition( Vec3( 16, 0, 0 ) );
    reflectiveSphere->SetScale( 2 );

    std::unique_ptr< Texture > renderTextureCube( dll.MakeRenderTextureCube( 512 ) );

    std::unique_ptr< Camera > cubeCamera( dll.MakeCamera() );
    cubeCamera->SetProjection( 45, 1, 1, 200 );
    cubeCamera->SetPosition( reflectiveSphere->GetPosition() );
    cubeCamera->SetRenderTarget( renderTextureCube.get() );

    std::unique_ptr< Material > reflMaterial( dll.MakeMaterial() );
    reflMaterial->SetShaders( "cook_torrance", "cook_torrance", "", "ENVIRONMENT_MAP" );
    reflMaterial->SetVector( "diffuseMap_ST", Vec4( 2, 2, 0, 0 ) );
    reflMaterial->SetTexture( "diffuseMap", white.get() );
    reflMaterial->SetTexture( "normalMap", normalTexture.get() );
    reflMaterial->SetTexture( "envMap", renderTextureCube.get() );
    reflMaterial->SetVector( "materialSpecular", Vec3( 0.8f, 0.8f, 0.8f ) );
    reflMaterial->SetFloat( "roughness", 0.1f );
    reflMaterial->SetFloat( "reflectivity", 0.9f );

    reflectiveSphere->SetMaterialForMesh( reflMaterial.get(), "Sphere" );

    std::unique_ptr< Material > cookTorranceMaterial( dll.MakeMaterial() );
    cookTorranceMaterial->SetShaders( "cook_torrance", "cook_torrance", "", "" );
    cookTorranceMaterial->SetTexture( "diffuseMap", black.get() );
    cookTorranceMaterial->SetTexture( "normalMap", normalTexture.get() );
    cookTorranceMaterial->SetVector( "materialSpecular", Vec3( 0.3f, 0.3f, 0.3f ) );
    cookTorranceMaterial->SetFloat( "roughness", 0.2f );

    std::unique_ptr< Scene > scene( dll.MakeScene() );
    scene->SetSkybox( skybox.get() );
    scene->Add( redSphere.get() );
    scene->Add( redSphere2.get() );
    scene->Add( greenSphere.get() );
    scene->Add( greenSphere2.get() );
    scene->Add( reflectiveSphere.get() );
    scene->Add( cubeCamera.get() );

    // Spheres
    const int sphereCount = 10;
    std::vector< std::unique_ptr< Model > > spheres( sphereCount );
    std::vector< std::unique_ptr< Material > > materials( sphereCount );

    for (std::size_t s = 0; s < spheres.size(); ++s)
    {
        spheres[ s ].reset( dll.MakeModel( "Assets/Models/sphere.ae3d", error ) );
        spheres[ s ]->SetPosition( Vec3( 10 + (float)s * 5, -5, s == 5 ? -4 : 0 ) );
        spheres[ s ]->SetScale( 2 );

        materials[ s ].reset( dll.MakeMaterial() );
        materials[ s ]->SetShaders( "cook_torrance", "cook_torrance", "", "" );
        materials[ s ]->SetTexture( "diffuseMap", asphalt.get() );
        materials[ s ]->SetTexture( "normalMap", normalTexture.get() );
        materials[ s ]->SetVector( "materialSpecular", Vec3( 0.3f, 0.3f, 0.3f ) );
        materials[ s ]->SetFloat( "roughness", 0.05f * s );

        spheres[ s ]->SetMaterialForMesh( materials[ s ].get(), "Sphere" );
        scene->Add( spheres[ s ].get() );
    }

    std::unique_ptr< Camera > camera( dll.MakeCamera() );
    camera->SetProjection( 45, (float)renderer->Width() / (float)renderer->Height(), 1, 200 );
    camera->LookAt( Vec3( 0, 0, -20 ), Vec3( 2000, 2000, 0 ), Vec3( 0, 1, 0 ) );
    scene->Add( camera.get() );

    std::unique_ptr< Light > redLight( dll.MakeLight( Light::Type::Point ) );
    redLight->SetPosition( Vec3( 10, -5, -2 ) );
    redLight->SetRadius( 10 );
    redLight->SetColor( Vec3( 1, 0, 0 ) );
    //scene->Add( redLight.get() );

    std::unique_ptr< Light > blueLight( dll.MakeLight( Light::Type::Point ) );
    blueLight->SetPosition( Vec3( 20, 0, -2 ) );
    blueLight->SetRadius( 10 );
    blueLight->SetColor( Vec3( 0, 0, 1 ) );
    //scene->Add( blueLight.get() );

    std::unique_ptr< Light > spotLight( dll.MakeLight( Light::Type::Spot ) );
    spotLight->SetPosition( Vec3( 20, -4, -2 ) );
    redSphere->SetPosition( Vec3( 20, -4, -2 ) );
    greenSphere->SetPosition( Vec3( 40, -4, -2 ) );
    spotLight->SetDirection( Vec3( 1, -0.1f, 0 ).Normalized() );
    spotLight->SetConeAngle( 30 );
    spotLight->SetRadius( 7 );
    spotLight->SetColor( Vec3( 0, 0, 1 ) );
    spotLight->SetCastShadow( true, 1024 );
    scene->Add( spotLight.get() );

    std::unique_ptr< Texture > gridTexture( dll.MakeTexture( "Assets/Textures/metal_grid.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );
    std::unique_ptr< Texture > gridTextureN( dll.MakeTexture( "Assets/Textures/metal_grid_n2.png", true, false, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );
    std::unique_ptr< Texture > gridTextureS( dll.MakeTexture( "Assets/Textures/metal_grid_s.png", true, false, Texture::Tiling::Repeat, Texture::Filter::Anisotropic8 ) );

    std::unique_ptr< Material > gridMaterial( dll.MakeMaterial() );
    gridMaterial->SetShaders( "cook_torrance", "cook_torrance", "", "SPECULAR_MAP" );
    gridMaterial->SetVector( "diffuseMap_ST", Vec4( 10, 10, 0, 0 ) );
    gridMaterial->SetTexture( "diffuseMap", gridTexture.get() );
    gridMaterial->SetTexture( "normalMap", gridTextureN.get() );
    gridMaterial->SetTexture( "specularMap", gridTextureS.get() );
    gridMaterial->SetVector( "materialSpecular", Vec3( 0.2f, 0.2f, 0.2f ) );
    gridMaterial->SetFloat( "roughness", 0.3f );

    std::unique_ptr< Model > floor( dll.MakeModel( "Assets/Models/plane.ae3d", error ) );
    floor->SetScale( 10 );
    floor->SetPosition( Vec3( 0, 1, 0 ) );
    floor->SetMaterialForMesh( gridMaterial.get(), "Cube" );
    scene->Add( floor.get() );

    renderer->SetClearColor( 0, 0, 0 );
    renderer->SetWindowTitle( "Aether3D Example: Lighting" );
    renderer->SetBloom( false, 0.55f, 2 );
    renderer->SetSSAO( false, 10, 1.4f );

    int frame = 0;
    float angle = 0;
    
    while (!keyboardInput->IsDown( ae3d::keyboardinput::KEY_ESC ))
    {
        mouseInput->SetRelative(renderer->HasFocus());

        HandleInput( keyboardInput.get(), camera.get() );
        HandleInput( mouseInput.get(), camera.get() );

        const float x = std::sin( frame * MathUtil::deg2rad ) * 10;
        const float z = std::cos( frame * MathUtil::deg2rad ) * 10;

        //blueLight->SetPosition( Vec3( x + 30, 0, z ) );
        /*++angle;
        Quaternion rotation;
        const Vec3 axis( 0, 1, 0 );
        rotation.FromAxisAngle( axis, angle );
        spotLight->SetRotation( rotation );
*/
        renderer->Draw( scene.get() );
        renderer->DrawStatistics();
        renderer->FinishFrame();
        renderer->SwapBuffers();

        ++frame;
    }

    renderer->Deinit();
    return 0;
}
