#include <cassert>
#include <iostream>
#include "../common.hpp"
#include "MathUtil.hpp"

const int gWidth = 1280;
const int gHeight = 720;

//const int gWidth = 1920;
//const int gHeight = 1080;

int main()
{
    Aether3D_API dll;
    // Note: This sample uses so many lights that it should be run with libaether3d_gl43 or libaether3d_d3d11.
    Error::Enum error = dll.LoadDLL( "../Renderers/libaether3d_gl43" );

    if (error == Error::DLL_NOT_FOUND)
    {
        assert( !"Could not find engine DLL!" );
        return (int)error;
    }
    else if (error == Error::FACTORY_NOT_FOUND)
    {
        assert( !"No entry point in DLL!" );
        return (int)error;
    }
    else if (error == Error::DLL_VERSION_MISMATCH)
    {
        assert( !"DLL header version does not match client header version!" );
        return (int)error;
    }

    Renderer* renderer = dll.MakeRenderer( "../Assets/builtin/" );
    assert( renderer );

    error = renderer->OpenWindow( gWidth, gHeight, false /* fullscreen */, 1, true );
    //error = renderer->OpenWindow( 0, 0, true, 1, true );

    if (error != Error::NO_ERRORS)
    {
        std::cerr << "Could not create rendering context!" << std::endl;
        renderer->Deinit();
        return (int)error;
    }

    std::unique_ptr< KeyboardInput > keyboardInput( dll.MakeKeyboardInput() );
    std::unique_ptr< MouseInput > mouseInput( dll.MakeMouseInput() );
    mouseInput->SetRelative( true );

    renderer->SetClearColor( 0.6f, 0.6f, 1 );
    renderer->SetWindowTitle( "Aether3D Example: Game" );
    renderer->SetClearColor( 0.2f, 0.2f, 0.5f );
    renderer->SetBloom( false, 0.5f, 2 );
    renderer->SetSSAO( false, 2.2f, 1.4f );

    std::unique_ptr< Scene > scene( dll.MakeScene() );
    scene->LoadFromFile( "scene.scene" );

    std::unique_ptr< Camera > camera( dll.MakeCamera() );
    camera->SetProjection( 55, (float)renderer->Width() / (float)renderer->Height(), 1, 340 );
    //camera->LookAt( Vec3( -100, 30, 0 ), Vec3( 2000, 0, 0 ), Vec3( 0, 1, 0 ) );
    camera->LookAt( Vec3( 100, 20, 0 ), Vec3( -2000, 0, 0 ), Vec3( 0, 1, 0 ) );
    scene->Add( camera.get() );
     
    /*std::unique_ptr< Light > pointLight1( dll.MakeLight( Light::Type::Point ) );
    pointLight1->SetPosition( Vec3( 20, -5, 0 ) );
    pointLight1->SetRadius( 10 );
    pointLight1->SetColor( Vec3( 1, 0, 0 ) );
    scene->Add( pointLight1.get() );
    */
    const int rows = 17;
    const int columns = 12;
    const int levels = 10;
    std::vector< Light* > pointLights( rows * columns * levels + 8 );
    int i = 0;
    const int radius = 8;

    for (int r = 0; r < rows; ++r)
    {
        for (int c = 0; c < columns; ++c)
        {
            for (int y = 0; y < levels; ++y)
            {
                pointLights[ i ] = dll.MakeLight( Light::Type::Point );
                pointLights[ i ]->SetPosition( Vec3( -100 + r * radius * 1.6f, y * radius * 1.6f + 2, -100 + c * radius * 1.6f ) );
                pointLights[ i ]->SetRadius( 10 );
                pointLights[ i ]->SetColor( Vec3( MathUtil::Random( 0, 1 ), MathUtil::Random( 0, 1 ), MathUtil::Random( 0, 1 ) ) );

                scene->Add( pointLights[ i ] );
                ++i;
            }
        }
    }

    for (int k = 0; k < 8; ++k)
    {
        pointLights[ i ] = dll.MakeLight( Light::Type::Point );
        pointLights[ i ]->SetPosition( Vec3( -110 + i * radius * 1.6f, 0 * radius * 1.6f + 2, -110 + i * radius * 1.6f ) );
        pointLights[ i ]->SetRadius( 10 );
        pointLights[ i ]->SetColor( Vec3( MathUtil::Random( 0, 1 ), MathUtil::Random( 0, 1 ), MathUtil::Random( 0, 1 ) ) );

        scene->Add( pointLights[ i ] );
        ++i;
        std::cout << "i: " << i << std::endl;
    }

    while (!keyboardInput->IsDown( ae3d::keyboardinput::KEY_ESC ))
    {
        mouseInput->SetRelative(renderer->HasFocus());

        HandleInput( keyboardInput.get(), camera.get() );
        HandleInput( mouseInput.get(), camera.get() );

        renderer->Draw( scene.get() );        
        renderer->DrawStatistics();

        std::string lightText = "Lights: ";
        lightText += std::to_string( pointLights.size() );
        renderer->Draw( lightText.c_str(), "aether3d_font", 1, 5, 150, Vec4( 1, 1, 1, 1 ) );
        
        renderer->FinishFrame();
        renderer->SwapBuffers();
    }

    for (std::size_t l = 0; l < pointLights.size(); ++l)
    {
        delete pointLights[ l ];
    }

    renderer->Deinit();
    return 0;
}

