#include <algorithm>
#include <iostream>
#include <memory>
#include <cassert>
#include "Aether3D.hpp"
#include "Errors.hpp"
#include "Vec3.hpp"
#ifdef AETHER3D_UNIX
#    include <unistd.h>
#endif

Aether3D_API dll;

static void TestDeledDAE( Error::Enum& outError )
{
    const int convertResult = system( "./convert_dae ../Tests/assets/models/deled.dae" );
    
    if (convertResult != 0)
    {
        std::cerr << "Could not convert file." << std::endl;
        outError = Error::CONTRACT_VIOLATION;
        return;
    }
    
    Model* model = dll.MakeModel( "../Tests/assets/models/deled.ae3d", outError );
    
    if (outError != Error::NO_ERRORS)
    {
        std::cerr << "deled.ae3d not found!" << std::endl;
        outError = Error::FILE_NOT_FOUND;
        return;
    }
    
    if (model->GetMeshNames().size() != 152)
    {
        std::cerr << "Mesh count isn't ok for deled.ae3d!" << std::endl;
        outError = Error::CORRUPTED_DATA;
        return;
    }
    
    if (std::find( model->GetMeshNames().begin(), model->GetMeshNames().end(), "cylinder172" ) == model->GetMeshNames().end())
    {
        std::cerr << "Mesh cylinder172 not found in deled.ae3d!" << std::endl;
        outError = Error::CORRUPTED_DATA;
        return;
    }
}

static void TestBlenderSkinnedDAE( Error::Enum& outError )
{
    const int convertResult = system( "./convert_dae ../Tests/assets/models/human_anim_test2_sl.dae" );
    
    if (convertResult != 0)
    {
        std::cerr << "Could not convert file." << std::endl;
        outError = Error::CONTRACT_VIOLATION;
        return;
    }
    
    Model* model = dll.MakeModel( "../Tests/assets/models/human_anim_test2_sl.ae3d", outError );
    
    if (outError != Error::NO_ERRORS)
    {
        std::cerr << "deled.ae3d not found!" << std::endl;
        outError = Error::FILE_NOT_FOUND;
        return;
    }
    
    auto jointNames = model->GetMeshJointNames( "body" );
    
    if (std::find( jointNames.begin(), jointNames.end(), "Bone_001" ) == jointNames.end())
    {
        std::cerr << "Mesh body doesn't contain joint Bone_001." << std::endl;
        outError = Error::CORRUPTED_DATA;
        return;
    }
}

static Error::Enum TestDAEConverter()
{
    const int makeResult = system( "make -f Makefile_dae" );
    
    if (makeResult != 0)
    {
        std::cerr << "Could not compile convert_dae." << std::endl;
        return Error::CONTRACT_VIOLATION;
    }
    
    Error::Enum error = Error::NO_ERRORS;
    
    TestDeledDAE( error );
    assert( error == Error::NO_ERRORS );

    TestBlenderSkinnedDAE( error );
    assert( error == Error::NO_ERRORS );

    return error;
}

static Error::Enum TestOBJConverter()
{
    const int makeResult = system( "make -f Makefile_obj" );
    
    if (makeResult != 0)
    {
        std::cerr << "Could not compile convert_obj." << std::endl;
        return Error::CONTRACT_VIOLATION;
    }
    
    const int convertResult = system( "./convert_obj ../Tests/assets/models/textured_cube.obj" );
    
    if (convertResult != 0)
    {
        std::cerr << "Could not convert file." << std::endl;
        return Error::CONTRACT_VIOLATION;
    }
    
    Error::Enum error = Error::NO_ERRORS;
    dll.MakeModel( "../Tests/assets/models/textured_cube.ae3d", error );
    
    if (error != Error::NO_ERRORS)
    {
        std::cerr << "Could not load textured_cube.ae3d" << std::endl;
        return error;
    }
    
    return error;
}

static Error::Enum TestFBXConverter()
{
    const int makeResult = system( "make -f Makefile_fbx" );
    
    if (makeResult != 0)
    {
        std::cerr << "Could not compile convert_fbx." << std::endl;
        return Error::CONTRACT_VIOLATION;
    }

    const int convertResult = system( "./convert_fbx ../Tests/assets/models/blender_2.71_default.fbx" );
    
    if (convertResult != 0)
    {
        std::cerr << "Could not convert file." << std::endl;
        return Error::CONTRACT_VIOLATION;
    }
    
    Error::Enum error = Error::NO_ERRORS;
    dll.MakeModel( "../Tests/assets/models/blender_2.71_default.ae3d", error );
    
    if (error != Error::NO_ERRORS)
    {
        std::cerr << "Could not load blender_2.71_default.ae3d" << std::endl;
        return error;
    }
    
    return error;
}

int main( int paramCount, char* params[] )
{
    Error::Enum error = dll.LoadDLL( paramCount > 1 ? std::string( params[ 1 ] ) : "" );
    
    if (error != Error::NO_ERRORS)
    {
        return (int)error;
    }
    
    Renderer* renderer = dll.MakeRenderer( "../../Assets/builtin/" );
    
    if (!renderer)
    {
        return (int)Error::FACTORY_NOT_FOUND;
    }
    
    int expectedWidth = 320;
    int expectedHeight = 200;
    error = renderer->OpenWindow( expectedWidth, expectedHeight, false, 1, true );
    
    if (error != Error::NO_ERRORS)
    {
        return (int)error;
    }

    chdir( "../../Tools" );

    error = TestOBJConverter();

    if (error != Error::NO_ERRORS)
    {
        return (int)error;
    }

    error = TestFBXConverter();

    if (error != Error::NO_ERRORS)
    {
        return (int)error;
    }

    error = TestDAEConverter();
    
    return (int)error;
}
