#include <cassert>
#include <iostream>
#include <vector>
#include "../../../Samples/common.hpp"

#define TEST_SKINNING "SKIN"

const int gWidth = 1280;
const int gHeight = 720;

#if AETHER3D_WINDOWS
#include <Windows.h>
INT WINAPI WinMain( HINSTANCE, HINSTANCE, LPSTR, INT )
#else
int main()
#endif
{
    Aether3D_API dll;
    Error::Enum error = dll.LoadDLL( "../../../Renderers/libaether3d_gl32.dylib" );

    if (error == Error::DLL_NOT_FOUND)
    {
        std::cerr << "Could not find engine DLL!" << std::endl;
        return (int)error;
    }
    else if (error == Error::FACTORY_NOT_FOUND)
    {
        std::cerr << "No entry point in DLL!" << std::endl;
        return (int)error;
    }
    else if (error == Error::DLL_VERSION_MISMATCH)
    {
        std::cerr << "DLL version does not match header version!" << std::endl;
        return (int)error;
    }

    Renderer* renderer = dll.MakeRenderer( "../../../Assets/builtin/" );
    assert( renderer );

    error = renderer->OpenWindow( gWidth, gHeight, false /* fullscreen */, 1, true );
    //error = renderer->OpenWindow( 0, 0, true, 1 );

    if (error != Error::NO_ERRORS)
    {
        std::cerr << "Could not create rendering context!" << std::endl;
        renderer->Deinit();
        return (int)error;
    }

    std::unique_ptr< KeyboardInput > keyboardInput( dll.MakeKeyboardInput() );
    std::unique_ptr< MouseInput > mouseInput( dll.MakeMouseInput() );
    mouseInput->SetRelative( true );
    
    renderer->SetClearColor( 0.6f, 0.6f, 1 );
    renderer->SetWindowTitle( "Aether3D Example: Model Loading" );
    renderer->AddShaderSearchPath( "../../../Assets/builtin/shaders/glsl150/" );

    std::unique_ptr< Texture > white( dll.MakeTexture( "../../../Assets/builtin/textures/white.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Texture > checker( dll.MakeTexture( "../../../Assets/builtin/textures/checker.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Texture > fontTexture( dll.MakeTexture( "../../../Assets/builtin/textures/font.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );

    std::unique_ptr< Material > material( dll.MakeMaterial() );
    material->SetShaders( "unlit", "unlit", "", "" );
    material->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    material->SetTexture( "textureMap", white.get() );
    //material->SetTexture( "textureMap", fontTexture.get() );
    //material->SetOpaque( false );
    
    std::unique_ptr< Material > skinMaterial( dll.MakeMaterial() );
    skinMaterial->SetShaders( "unlit", "unlit", "", TEST_SKINNING );
    skinMaterial->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    skinMaterial->SetTexture( "textureMap", checker.get() );

    std::unique_ptr< Scene > scene( dll.MakeScene() );

    // Large-scale demo asset loading begins.
    int rows = 5;
    int columns = 5;
    
    std::vector< Model* > skinModels( rows * columns );
    int i = 0;

    for (int r = 0; r < rows; ++r)
    {
        for (int c = 0; c < columns; ++c)
        {
            skinModels[ i ] = dll.MakeModel( "../../../Samples/Assets/Models/skinnedMesh.ae3d", error );
            skinModels[ i ]->SetPosition( Vec3( r * 10, -1.5f, c * 10 ) );
            skinModels[ i ]->SetMaterialForMesh( skinMaterial.get( ), "body" );
            
            scene->Add( skinModels[ i ] );
            ++i;
        }
    }
    
    std::unique_ptr< Texture > asphaltTexture( dll.MakeTexture( "../../../Samples/Assets/Textures/asphalt.jpg", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Material > asphaltMaterial( dll.MakeMaterial() );
    asphaltMaterial->SetShaders( "unlit", "unlit", "", "" );
    asphaltMaterial->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    asphaltMaterial->SetVector( "textureMap_ST", Vec4( 6, 6, 0, 0 ) );
    //asphaltMaterial->SetTexture( "textureMap", white/*asphaltTexture*/.get() );
    asphaltMaterial->SetTexture( "textureMap", asphaltTexture.get() );

    std::unique_ptr< Texture > skybox( dll.MakeTexture( "../../../Assets/textures/skybox/skybox_left.jpg",
                                                       "../../../Assets/textures/skybox/skybox_right.jpg",
                                                       "../../../Assets/textures/skybox/skybox_top.jpg",
                                                       "../../../Assets/textures/skybox/skybox_bottom.jpg",
                                                       "../../../Assets/textures/skybox/skybox_front.jpg",
                                                       "../../../Assets/textures/skybox/skybox_back.jpg") );

    std::unique_ptr< Model > floor( dll.MakeModel( "../../../Samples/Assets/Models/plane.ae3d", error ) );
    floor->SetScale( 10 );
    floor->SetMaterialForMesh( asphaltMaterial.get(), "Cube" );
    scene->Add( floor.get() );
    //scene->SetSkybox( skybox.get() );
    
    // Large-scale demo asset loading ends.
    
    std::unique_ptr< Camera > camera( dll.MakeCamera() );
    camera->SetProjection( 55, (float)renderer->Width() / (float)renderer->Height(), 1, 200 );
    //camera->SetProjection( -10, 10, -10, 10, 1, 200 );
    camera->LookAt( Vec3( 0, 0, -20 ), Vec3( 0, 0, 2000 ), Vec3( 0, 1, 0 ) );
    scene->Add( camera.get() );
    
    renderer->SetClearColor( 0.2f, 0.2f, 0.5f );
    
    float angle = 0;
    //double gpuTime = 0;

    while (!keyboardInput->IsDown( ae3d::keyboardinput::KEY_ESC ))
    {
        // Timer Query broken in OS X since upgrading to Mavericks.
        //renderer->BeginQuery();
        
        ++angle;
        Quaternion rotation;
        Vec3 axis( 1, 1, 1 );
        axis = axis.Normalized();
        rotation.FromAxisAngle( axis, angle );
        
        for (auto skinModel : skinModels)
        {
            skinModel->SetRotation( rotation );
        }
        
        HandleInput( keyboardInput.get(), camera.get() );
        HandleInput( mouseInput.get(), camera.get() );

        renderer->Draw( scene.get() );
        
        //std::string gpuText = "GPU: ";
        //gpuText += std::to_string( gpuTime ) + " ms";
        //renderer->Draw( gpuText, "aether3d_font", 1, 5, 50, 1, 1, 1, 1 );

        renderer->DrawStatistics();
        renderer->FinishFrame();

        renderer->SwapBuffers();
        //gpuTime = renderer->EndQuery();
    }

    renderer->Deinit();
    return 0;
}

