// This test should be run in Tests folder.
#include <cassert>
#include <iostream>
#include <string>
#include "../../../Samples/common.hpp"

const int gWidth = 1920;
const int gHeight = 1080;

#if AETHER3D_WINDOWS
#include <Windows.h>
INT WINAPI WinMain( HINSTANCE, HINSTANCE, LPSTR, INT )
#else
int main()
#endif
{
    Aether3D_API dll;
    Error::Enum error = dll.LoadDLL( "../Renderers/libaether3d_gl32" );

    if (error == Error::DLL_NOT_FOUND)
    {
        std::cerr << "Could not find engine DLL!" << std::endl;
        return (int)error;
    }
    else if (error == Error::FACTORY_NOT_FOUND)
    {
        std::cerr << "No entry point in DLL!" << std::endl;
        return (int)error;
    }
    else if (error == Error::DLL_VERSION_MISMATCH)
    {
        std::cerr << "DLL version does not match header version!" << std::endl;
        return (int)error;
    }

    Renderer* renderer = dll.MakeRenderer( "../Assets/builtin" );
    assert( renderer );

    error = renderer->OpenWindow( gWidth, gHeight, false /* fullscreen */, 1, true );
    //error = renderer->OpenWindow( 0, 0, true, 1 );

    if (error != Error::NO_ERRORS)
    {
        std::cerr << "Could not create rendering context!" << std::endl;
        renderer->Deinit();
        return (int)error;
    }

    std::unique_ptr< KeyboardInput > keyboardInput( dll.MakeKeyboardInput() );
    std::unique_ptr< MouseInput > mouseInput( dll.MakeMouseInput() );
    mouseInput->SetRelative( true );
    
    renderer->SetClearColor( 0.6f, 0.6f, 1 );
    renderer->SetWindowTitle( "Aether3D Example: SSAO" );

    std::unique_ptr< Texture > white( dll.MakeTexture( "../Assets/builtin/textures/white.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Texture > checker( dll.MakeTexture( "../Assets/builtin/textures/checker.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Texture > headTexture( dll.MakeTexture( "../Samples/Assets/textures/panel.jpg", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Texture > fontTexture( dll.MakeTexture( "../Assets/builtin/textures/font.png", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );

    std::unique_ptr< Model > sphere( dll.MakeModel( "../Samples/Assets/Models/lighting_test.ae3d", error ) );
    sphere->SetPosition( Vec3( 10, -8, 0 ) );
    sphere->SetScale( 2 );

    std::unique_ptr< Model > skinnedModel( dll.MakeModel( "../Samples/Assets/Models/skinnedMesh.ae3d", error ) );
    skinnedModel->SetPosition( Vec3( 0, -1.5f, 0 ) );

    std::unique_ptr< Model > morphTest( dll.MakeModel( "../Samples/Assets/Models/morph.ae3d", error ) );
    morphTest->SetPosition( Vec3( 10, 0, 5 ) );

    std::unique_ptr< Material > material( dll.MakeMaterial() );
    material->SetShaders( "unlit", "unlit", "", "" );
    material->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    material->SetTexture( "textureMap", white.get() );
    //material->SetTexture( "textureMap", fontTexture.get() );
    //material->SetOpaque( false );
    
    std::unique_ptr< Material > skinMaterial( dll.MakeMaterial() );
    skinMaterial->SetShaders( "unlit", "unlit", "", "SKIN" );
    skinMaterial->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    skinMaterial->SetTexture( "textureMap", white.get() );

    std::unique_ptr< Material > morphMaterial( dll.MakeMaterial() );
    morphMaterial->SetShaders( "morph_test", "unlit", "", "" );
    morphMaterial->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    morphMaterial->SetTexture( "textureMap", headTexture.get() );

    skinnedModel->SetMaterialForMesh( skinMaterial.get( ), "body" );
    morphTest->SetMaterialForMesh( morphMaterial.get(), "polySurfaceShape1" );

    sphere->SetMaterialForMesh( material.get( ), "Sphere" );

    std::unique_ptr< Scene > scene( dll.MakeScene() );
    scene->Add( skinnedModel.get( ) );
    scene->Add( morphTest.get( ) );
    scene->Add( sphere.get( ) );

    // Large-scale demo asset loading begins.
    std::unique_ptr< Texture > asphaltTexture( dll.MakeTexture( "../Samples/Assets/Textures/asphalt.jpg", true, true, Texture::Tiling::Repeat, Texture::Filter::Linear ) );
    std::unique_ptr< Material > asphaltMaterial( dll.MakeMaterial() );
    asphaltMaterial->SetShaders( "unlit", "unlit", "", "" );
    asphaltMaterial->SetVector( "tint", Vec4( 1, 1, 1, 1 ) );
    asphaltMaterial->SetVector( "textureMap_ST", Vec4( 6, 6, 0, 0 ) );
    //asphaltMaterial->SetTexture( "textureMap", white/*asphaltTexture*/.get() );
    asphaltMaterial->SetTexture( "textureMap", asphaltTexture.get() );

    std::unique_ptr< Texture > skybox( dll.MakeTexture( "../Assets/textures/skybox/skybox_left.jpg",
                                                       "../Assets/textures/skybox/skybox_right.jpg",
                                                       "../Assets/textures/skybox/skybox_top.jpg",
                                                       "../Assets/textures/skybox/skybox_bottom.jpg",
                                                       "../Assets/textures/skybox/skybox_front.jpg",
                                                       "../Assets/textures/skybox/skybox_back.jpg") );

    std::unique_ptr< Model > floor( dll.MakeModel( "../Samples/Assets/Models/plane.ae3d", error ) );
    floor->SetScale( 10 );
    floor->SetMaterialForMesh( asphaltMaterial.get(), "Cube" );
    scene->Add( floor.get() );
    //scene->SetSkybox( skybox.get() );
    
    // Large-scale demo asset loading ends.
    
    std::unique_ptr< Camera > camera( dll.MakeCamera() );
    camera->SetProjection( 55, (float)renderer->Width() / (float)renderer->Height(), 1, 200 );
    //camera->SetProjection( -10, 10, -10, 10, 1, 200 );
    camera->LookAt( Vec3( 0, 0, -20 ), Vec3( 0, 0, 2000 ), Vec3( 0, 1, 0 ) );
    scene->Add( camera.get() );
    
    skinnedModel->ApplyAnimation( 0, 0 );

    renderer->SetClearColor( 0.2f, 0.2f, 0.5f );
    renderer->SetBloom( false, 0.45f, 2 );
    renderer->SetSSAO( true, 2.2f, 2.4f );
     
    float angle = 0;
    bool morphPlaying = false;
    double gpuTime = 0;

    Quaternion morphRotation;
    morphRotation.FromAxisAngle( Vec3( 0, 1, 0 ), 180 );
    morphTest->SetRotation( morphRotation );
    
    while (!keyboardInput->IsDown( ae3d::keyboardinput::KEY_ESC ))
    {
        // Timer Query broken in OS X since upgrading to Mavericks.
        renderer->BeginQuery();

        if (!morphPlaying && keyboardInput->IsDown( ae3d::keyboardinput::KEY_r ))
        {
            morphTest->ApplyAnimation( 0, 0 );
            morphPlaying = true;
        }
        else if (morphPlaying)
        {
            morphPlaying = false;
        }
        
        ++angle;
        Quaternion rotation;
        Vec3 axis = Vec3( 1, 1, 1 ).Normalized();
        rotation.FromAxisAngle( axis, angle );
        sphere->SetRotation( rotation );
        
        HandleInput( keyboardInput.get(), camera.get() );
        HandleInput( mouseInput.get(), camera.get() );

        Vec3 p = camera->ScreenPoint( skinnedModel->GetPosition() );
        renderer->Draw( "model", "aether3d_font", 1, p.x, p.y + 120, Vec4( 0, 0, 0, 1 ) );

        renderer->Draw( scene.get() );
        
        std::string gpuText = "GPU: ";
        gpuText += std::to_string( gpuTime ) + " ms";
        renderer->Draw( gpuText.c_str(), "aether3d_font", 1, 5, 50, Vec4( 1, 1, 1, 1 ) );
        renderer->Draw( "press r for blend shape animation.", "aether3d_font", 1, 5, 80, Vec4( 1, 1, 1, 1 ) );

        renderer->DrawStatistics();
        renderer->FinishFrame();

        renderer->SwapBuffers();
        gpuTime = renderer->EndQuery();
    }

    renderer->Deinit();
}

