/**
   \file convert_dae.cpp
 
   \author Timo Wiren
   \date 2014-07-24
   Licence: zlib Licence
 
   Note: Matrices in COLLADA are column matrices.
 
   Only one skeleton per mesh is supported.
   Skinning tested only with Blender.
   Blender: Meshes must be exported with "Second Life" or "OpenSim" flag.
*/
#include <fstream>
#include <iostream>
#include <string>
#include <sstream>
#include <cassert>
#include "pugixml.hpp"
#include "common.hpp"
#include "Matrix.hpp"

void Take4mostSignificantWeights( const std::vector< float >& weights, Influence& outInfluence );

bool gZup = false;

static void AddJointPointer( Joint* joint, const std::string& name, Mesh& outMesh )
{
    if (name == joint->name)
    {
        outMesh.joints.push_back( *joint );
        return;
    }

    for (std::size_t c = 0; c < joint->children.size(); ++c)
    {
        AddJointPointer( &joint->children[ c ], name, outMesh );
    }
}


/**
   \param node <input>
   \param outMesh Mesh where joint names are read into.
 */
static void ReadJointNames( const pugi::xml_node& node, Mesh& outMesh )
{
    std::string source = node.attribute( "source" ).as_string();
    source = source.substr( 1, source.length() );

    for (auto sourceNode : node.parent().parent().children( "source" ))
    {
        if (source == sourceNode.attribute( "id" ).as_string())
        {
            const std::string names = sourceNode.child( "Name_array" ).text().as_string();
            std::stringstream stm( names );

            while (!stm.eof())
            {
                std::string name;
                stm >> name;
                outMesh.jointNames.push_back( name );
            }

            // Finds joint pointers and stores them to jointArray.
            // We do this so we don't have to find a joint from the skeleton
            // for each influence when deforming the mesh, because that's really slow.
            for (std::size_t i = 0; i < outMesh.jointNames.size(); ++i)
            {
                AddJointPointer( &outMesh.skeletonRoot, outMesh.jointNames[ i ], outMesh );
            }

            break;
        }
    }
}


/**
   \param node <input>
   \param outMesh Mesh.
 */
static void ReadWeights( const pugi::xml_node& node, Mesh& outMesh )
{
    std::string source = node.attribute( "source" ).as_string();
    source = source.substr( 1, source.length() );

    for (auto sourceNode : node.parent().parent().children( "source" ))
    {
        if (source == sourceNode.attribute( "id" ).as_string())
        {
            const std::string weights = sourceNode.child( "float_array" ).text().as_string();
            std::stringstream stm( weights );
            int count = sourceNode.child( "float_array" ).attribute( "count" ).as_int();
            assert( count > 0 );
            outMesh.weights.resize( count );

            for (int i = 0; i < count; ++i)
            {
                stm >> outMesh.weights[ i ];
            }

            break;
        }
    }
}


/**
   \param node <vertex_weights>
   \param outMesh Mesh.
 */
static void ReadInfluences( const pugi::xml_node& node, Mesh& outMesh )
{
    int jointOffset = -1;
    int weightOffset = -1;

    for (auto input : node.children( "input" ))
    {
        if (std::string( "JOINT" ) == input.attribute( "semantic" ).as_string())
        {
            jointOffset = input.attribute( "offset" ).as_int();
            ReadJointNames( input, outMesh );
        }
        else if (std::string( "WEIGHT" ) == input.attribute( "semantic" ).as_string())
        {
            weightOffset = input.attribute( "offset" ).as_int();
            ReadWeights( input, outMesh );
        }
    }

    std::stringstream vcountStm( node.child( "vcount" ).text().as_string() );

    const int count = node.attribute( "count" ).as_int();
    std::vector< unsigned > influences( count );
    bool moreThan4weights = false;

    for (int i = 0; i < count; ++i)
    {
        vcountStm >> influences[ i ];

        if (influences[ i ] > 4)
        {
            moreThan4weights = true;
        }
    }

    if (moreThan4weights)
    {
        std::cerr << "Warning!  More than 4 bones affect a vertex!  Selecting 4 most weighted bones." << std::endl;
    }

    std::stringstream vStm( node.child( "v" ).text().as_string() );
    outMesh.influences.resize( influences.size() );

    for (std::size_t v = 0; v < influences.size(); ++v)
    {
        Influence influence;

        for (unsigned i = 0; i < influences[ v ]; ++i)
        {
            int jointIndex = 0;
            int weightIndex = 0;

            if (jointOffset == 0)
            {
                vStm >> jointIndex;
                vStm >> weightIndex;
            }
            else if (weightOffset == 0)
            {
                vStm >> weightIndex;
                vStm >> jointIndex;
            }

            influence.jointIndices.push_back(  jointIndex  );
            influence.weightIndices.push_back( weightIndex );

            if (influence.weightIndices.size() > 4)
            {
                Take4mostSignificantWeights( outMesh.weights, influence );
            }

            outMesh.influences[ v ] = influence;
        }
    }
}


/**
 \param node <skin> node.
 \param outMesh Mesh.
 */
static void ReadSkin( const pugi::xml_node& node, Mesh& outMesh )
{
    auto matrix = node.child( "bind_shape_matrix" );
    std::stringstream matrixDataStm( matrix.text().as_string() );

    for (int i = 0; i < 16; ++i)
    {
        matrixDataStm >> outMesh.bindShapeMatrix.m[ i ];
    }

    outMesh.bindShapeMatrix.Transpose( outMesh.bindShapeMatrix );

    // Applies BSM to vertices.
    for (std::size_t v = 0; v < outMesh.vertex.size(); ++v)
    {
        Matrix44::TransformPoint( outMesh.vertex[ v ], outMesh.bindShapeMatrix, &outMesh.vertex[ v ] );
    }

    ReadInfluences( node.child( "vertex_weights" ), outMesh );

    // Reads inverse bind matrices.
    for (auto inputNode : node.child( "joints" ).children( "input" ))
    {
        if (std::string( "INV_BIND_MATRIX" ) != inputNode.attribute( "semantic" ).as_string())
        {
            continue;
        }

        std::string source = inputNode.attribute( "source" ).value();
        source = source.substr( 1, source.length() );

        // Finds the <source with inverse bind matrices.
        for (auto sourceNode : node.children( "source" ))
        {
            if (source != sourceNode.attribute( "id" ).value())
            {
                continue;
            }

            std::string matrixData = sourceNode.child( "float_array" ).text().as_string();
            std::stringstream stm( matrixData );
            Matrix44 invBindMatrix;

            for (std::size_t jointNameIndex = 0; jointNameIndex < outMesh.jointNames.size(); ++jointNameIndex)
            {
                for (int i = 0; i < 16; ++i)
                {
                    stm >> invBindMatrix.m[ i ];
                }

                invBindMatrix.Transpose( invBindMatrix );

                for (std::size_t jointIndex = 0; jointIndex < outMesh.joints.size(); ++jointIndex )
                {
                    if (outMesh.jointNames[ jointNameIndex ] == outMesh.joints[ jointIndex ].name)
                    {
                        outMesh.joints[ jointIndex ].invBindMatrix = invBindMatrix;

                        Matrix44::Multiply( outMesh.joints[ jointIndex ].invBindMatrix, outMesh.joints[ jointIndex ].transform, outMesh.joints[ jointIndex ].transform );
                    }
                }
            }

            break;
        }

        break;
    }
}


static void ReadVertices( const pugi::xml_node& meshNode, Mesh& outMesh )
{
    // Assumes that mesh has only one <vertices> and it has only one <input>.
    std::string inputUrl = meshNode.child( "vertices" ).child( "input" ).attribute( "source" ).value();
    inputUrl = inputUrl.substr( 1, inputUrl.length() );

    for (auto sourceNode : meshNode.children( "source" ))
    {
        if (sourceNode.attribute( "id" ).value() != inputUrl)
        {
            continue;
        }

        pugi::xml_node floatArrayNode = sourceNode.child( "float_array" );
        const int count = floatArrayNode.attribute( "count" ).as_int();
        outMesh.vertex.resize( count / 3 );

        const std::string vertices = floatArrayNode.text().as_string();
        std::stringstream stm( vertices );

        if (gZup)
        {
            for (int v = 0; v < count / 3; ++v)
            {
                stm >> outMesh.vertex[v].x;
                stm >> outMesh.vertex[v].z;
                stm >> outMesh.vertex[v].y;
                outMesh.vertex[v].z *= -1.0f;
            }
        }
        else
        {
            for (int v = 0; v < count / 3; ++v)
            {
                stm >> outMesh.vertex[v].x;
                stm >> outMesh.vertex[v].y;
                stm >> outMesh.vertex[v].z;
            }
        }

        break;
    }
}


static void ReadNormals( const pugi::xml_node& meshNode, const std::string& url, Mesh& outMesh )
{
    for (auto sourceNode : meshNode.children( "source" ))
    {
        if (sourceNode.attribute( "id" ).value() != url)
        {
            continue;
        }

        pugi::xml_node floatArrayNode = sourceNode.child( "float_array" );
        const int count = floatArrayNode.attribute( "count" ).as_int();
        outMesh.vnormal.resize( count / 3 );

        const std::string normals = floatArrayNode.text().as_string();
        std::stringstream stm( normals );

        if (gZup)
        {
            for (int v = 0; v < count / 3; ++v)
            {
                stm >> outMesh.vnormal[v].x;
                stm >> outMesh.vnormal[v].z;
                stm >> outMesh.vnormal[v].y;
                outMesh.vnormal[v].z *= -1;
            }
        }
        else
        {
            for (int v = 0; v < count / 3; ++v)
            {
                stm >> outMesh.vnormal[v].x;
                stm >> outMesh.vnormal[v].y;
                stm >> outMesh.vnormal[v].z;
            }
        }

        break;
    }
}


static void ReadColors( const pugi::xml_node& meshNode, const std::string& url, Mesh& outMesh )
{
    for (auto sourceNode : meshNode.children( "source" ))
    {
        if (sourceNode.attribute( "id" ).value() != url)
        {
            continue;
        }
        
        pugi::xml_node floatArrayNode = sourceNode.child( "float_array" );
        const int count = floatArrayNode.attribute( "count" ).as_int();
        outMesh.colors.resize( count / 3 );
        
        const std::string colors = floatArrayNode.text().as_string();
        std::stringstream stm( colors );
        
        if (gZup)
        {
            for (int v = 0; v < count / 3; ++v)
            {
                stm >> outMesh.colors[ v ].x;
                stm >> outMesh.colors[ v ].z;
                stm >> outMesh.colors[ v ].y;
            }
        }
        else
        {
            for (int v = 0; v < count / 3; ++v)
            {
                stm >> outMesh.colors[ v ].x;
                stm >> outMesh.colors[ v ].y;
                stm >> outMesh.colors[ v ].z;
            }
        }
        
        break;
    }
}


static void ReadTexCoords( const pugi::xml_node& meshNode, const std::string& url, Mesh& outMesh )
{
    for (auto sourceNode : meshNode.children( "source" ))
    {
        if (sourceNode.attribute( "id" ).value() != url)
        {
            continue;
        }

        pugi::xml_node floatArrayNode = sourceNode.child( "float_array" );
        const int count = floatArrayNode.attribute( "count" ).as_int();
        outMesh.tcoord.resize( count / 2 );

        const std::string texCoords = floatArrayNode.text().as_string();
        std::stringstream stm( texCoords );
        int texCoordInd = 0;

        while (!stm.eof())
        {
            stm >> outMesh.tcoord[ texCoordInd ].u;
            stm >> outMesh.tcoord[ texCoordInd ].v;
            ++texCoordInd;
        }

        break;
    }
}


static void ReadMesh( const pugi::xml_node& meshNode, Mesh& outMesh )
{
    // \todo Seems like <mesh> can contain many of these depending of material,
    //       see Astroboy sample.
    pugi::xml_node polyList = meshNode.child( "polylist" );
    pugi::xml_node triangleList = meshNode.child( "trianglelist" );
    pugi::xml_node triangles = meshNode.child( "triangles" );

    // Reads indices.
    if (meshNode.child( "polygons" ))
    {
        std::cerr << "Mesh is not triangulated!  Check your export settings." << std::endl;
        exit( 1 );
    }
    else if (!polyList && !triangleList && !triangles)
    {
       assert( !"Unhandled indices node type." );
    }

    pugi::xml_node indicesNode;

    if (polyList)
    {
        indicesNode = polyList;
    }
    else if (triangleList)
    {
        indicesNode = triangleList;
    }
    else if (triangles)
    {
        indicesNode = triangles;
    }
    else
    {
        assert( !"Unhandled indices node type." );
    }

    int verOffset = -1;
    int norOffset = -1;
    int texOffset = -1;
    int colOffset = -1;

    // Iterates through triangle's or polylist's children, looking for
    // <input> to find vertices, normals and texcoords and their offsets.
    for (auto inputNode : indicesNode.children( "input" ))
    {
        const std::string semantic = inputNode.attribute( "semantic" ).value();
        std::string url = inputNode.attribute( "source" ).value();
        url = url.substr( 1, url.length() );

        if (semantic == "VERTEX")
        {
            verOffset = inputNode.attribute( "offset" ).as_int();
            ReadVertices( meshNode, outMesh );
        }
        else if (semantic == "NORMAL")
        {
            norOffset = inputNode.attribute( "offset" ).as_int();
            ReadNormals( meshNode, url, outMesh );
        }
        else if (semantic == "COLOR")
        {
            colOffset = inputNode.attribute( "offset" ).as_int();
            ReadColors( meshNode, url, outMesh );
        }
        else if (semantic == "TEXCOORD")
        {
            texOffset = inputNode.attribute( "offset" ).as_int();
            ReadTexCoords( meshNode, url, outMesh );
        }
    }

    const int indicesCount = indicesNode.attribute( "count" ).as_int();

    if (indicesCount > 65535)
    {
        std::cerr << "Error: Mesh contains more than 65535 triangles!" << std::endl;
        exit(1);
    }

    std::vector< int > vCounts( indicesCount );

    const std::string vCountList = indicesNode.child( "vcount" ).text().as_string();
    std::stringstream stm( vCountList );

    for (int f = 0; f < indicesCount; ++f)
    {
        stm >> vCounts[ f ];

        if (vCounts[ f ] > 4)
        {
            std::cerr << "Error!  Polygon has more than 4 vertices." << std::endl;
            exit( 1 );
        }
    }

    const std::string vData = indicesNode.child( "p" ).text().as_string();
    std::stringstream vDataStm( vData );
    const int elemCount = (verOffset == -1 ? 0 : 1) +
                          (norOffset == -1 ? 0 : 1) +
                          (texOffset == -1 ? 0 : 1) + 
                          (colOffset == -1 ? 0 : 1);

    for (int f = 0; f < indicesCount; ++f)
    {
        outMesh.face.push_back( Face() );
        unsigned short i;

        // Reads the first three vertices.
        for (int v = 0; v < 3; ++v)
        {
            vDataStm >> i;
            if (verOffset == 0) { outMesh.face.back().vInd[v] = i; }
            if (norOffset == 0) { outMesh.face.back().vnInd[v] = i; }
            if (texOffset == 0) { outMesh.face.back().uvInd[v] = i; }
            if (colOffset == 0) { outMesh.face.back().colInd[v] = i; }

            if (elemCount > 1)
            {
                vDataStm >> i;
                if (verOffset == 1) { outMesh.face.back().vInd[v] = i; }
                if (norOffset == 1) { outMesh.face.back().vnInd[v] = i; }
                if (texOffset == 1) { outMesh.face.back().uvInd[v] = i; }
                if (colOffset == 1) { outMesh.face.back().colInd[v] = i; }
            }

            if (elemCount > 2 && norOffset != texOffset && texOffset != colOffset)
            {
                vDataStm >> i;
                if (verOffset == 2) { outMesh.face.back().vInd[v] = i; }
                if (norOffset == 2) { outMesh.face.back().vnInd[v] = i; }
                if (texOffset == 2) { outMesh.face.back().uvInd[v] = i; }
                if (colOffset == 2) { outMesh.face.back().colInd[v] = i; }
            }

            if (elemCount > 3 && norOffset != texOffset && texOffset != colOffset)
            {
                vDataStm >> i;
                if (verOffset == 3) { outMesh.face.back().vInd[v] = i; }
                if (norOffset == 3) { outMesh.face.back().vnInd[v] = i; }
                if (texOffset == 3) { outMesh.face.back().uvInd[v] = i; }
                if (colOffset == 3) { outMesh.face.back().colInd[v] = i; }
            }

            // \todo maybe a new if () because of color?
        }
        
        // Reads fourth vertex if it exists.
        if (vCounts[ f ] != 4) { continue; }

        Face face2;

        face2.vInd[0]   = outMesh.face.back().vInd[2];
        face2.uvInd[0]  = outMesh.face.back().uvInd[2];
        face2.vnInd[0]  = outMesh.face.back().vnInd[2];
        face2.colInd[0] = outMesh.face.back().colInd[2];

        face2.vInd[2]   = outMesh.face.back().vInd[0];
        face2.uvInd[2]  = outMesh.face.back().uvInd[0];
        face2.vnInd[2]  = outMesh.face.back().vnInd[0];
        face2.colInd[2] = outMesh.face.back().colInd[0];

        vDataStm >> i;
        if (verOffset == 0) { face2.vInd[1] = i; }
        if (norOffset == 0) { face2.vnInd[1] = i; }
        if (texOffset == 0) { face2.uvInd[1] = i; }
        if (colOffset == 0) { face2.colInd[1] = i; }

        if (elemCount > 1)
        {
            vDataStm >> i;
            if (verOffset == 1) { face2.vInd[1] = i; }
            if (norOffset == 1) { face2.vnInd[1] = i; }
            if (texOffset == 1) { face2.uvInd[1] = i; }
            if (colOffset == 1) { face2.colInd[1] = i; }
        }

        if (elemCount > 2 && norOffset != texOffset && texOffset != colOffset)
        {
            vDataStm >> i;
            if (verOffset == 2) { face2.vInd[1] = i; }
            if (norOffset == 2) { face2.vnInd[1] = i; }
            if (texOffset == 2) { face2.uvInd[1] = i; }
            if (colOffset == 2) { face2.colInd[1] = i; }
        }
        
        if (elemCount > 3 && norOffset != texOffset && texOffset != colOffset)
        {
            vDataStm >> i;
            if (verOffset == 3) { face2.vInd[1] = i; }
            if (norOffset == 3) { face2.vnInd[1] = i; }
            if (texOffset == 3) { face2.uvInd[1] = i; }
            if (colOffset == 3) { face2.colInd[1] = i; }
        }
            // \todo maybe a new if () because of color?
        
        outMesh.face.push_back( face2 );
    }
}


static void ReadNonMatrixTransform( const pugi::xml_node& node, Matrix44& outTransform )
{
    Vec3 translate;
    auto translateNode = node.child( "translate" );
    std::stringstream translateStm( translateNode.value() );
    translateStm >> translate.x >> translate.y >> translate.z;

    Vec3 rotate;
    for (auto rotateNode : node.children( "rotate" ))
    {
        std::stringstream stm( rotateNode.value() );

        const std::string axis = rotateNode.attribute( "sid" ).value();
        // We are interested only in the last value.
        if (axis == "rotationX" || axis == "rotateX" || axis == "rotX")
        {
            stm >> rotate.x >> rotate.x >> rotate.x >> rotate.x;
        }
        else if (axis == "rotationY" || axis == "rotateY" || axis == "rotY")
        {
            stm >> rotate.y >> rotate.y >> rotate.y >> rotate.y;
        }
        else if (axis == "rotationZ" || axis == "rotateZ" || axis == "rotZ")
        {
            stm >> rotate.z >> rotate.z >> rotate.z >> rotate.z;
        }
    }

    Vec3 scale( 1, 1, 1 );

    if (node.child( "scale" ))
    {
        std::stringstream scaleStm( node.child( "scale" ).text().as_string() );
        scaleStm >> scale.x >> scale.y >> scale.z;
    }

    outTransform.MakeRotationXYZ( rotate.x, rotate.y, rotate.z );
    outTransform.Scale( scale.x, scale.y, scale.z );

    // \fixme is this correct or should I use Matrix44::Translate()?
    outTransform.m[ 12 ] = translate.x;
    outTransform.m[ 13 ] = translate.y;
    outTransform.m[ 14 ] = translate.z;
}


/**
   Reads joint nodes to mesh's skeleton.  Should begin from the root.

   \param node XML node that will be read to outJoint.
   \param outJoint Joint that will contain node's data.
 */
static void ReadJointNode( const pugi::xml_node& node, Joint& outJoint )
{
    outJoint.name = node.attribute( "id" ).value();

    for (auto child : node.children())
    {
        std::string nodeName = child.name();

        if (nodeName == "matrix")
        {
            std::stringstream stm( child.value() );

            for (std::size_t i = 0; i < 16; ++i)
            {
                stm >> outJoint.transform.m[ i ];
            }

            outJoint.transform.Transpose( outJoint.transform );
        }
        else if (nodeName == "node")
        {
            outJoint.children.push_back( Joint() );
            ReadJointNode( child, outJoint.children.back() );
        }
    }
}


/**
   \param node <instance_controller>
   \param outMesh Mesh.
 */
static void ReadAnimations( const pugi::xml_node node, Mesh& outMesh )
{
    auto libAnimations = node.root().first_element_by_path( "COLLADA/library_animations" );

    if (!libAnimations)
    {
        return;
    }

    for (auto animNode : libAnimations.children( "animation" ))
    {
        const std::string id = animNode.attribute( "id" ).value();
        outMesh.animations[ id ] = Animation();
        outMesh.animations[ id ].name = id;

        // Gets INPUT and OUTPUT.
        for (auto inputNode : animNode.child( "sampler" ).children( "input" ))
        {
            std::string sourceId = inputNode.attribute( "source" ).value();
            sourceId = sourceId.substr( 1, sourceId.length() );

            const bool isInput = std::string( "INPUT" ) == inputNode.attribute( "semantic" ).value();
            const bool isOutput = std::string( "OUTPUT" ) == inputNode.attribute( "semantic" ).value();

            if (isInput || isOutput)
            {
                for (auto sourceNode : animNode.children( "source" ))
                {
                    if (sourceId == sourceNode.attribute( "id" ).value())
                    {
                        std::stringstream stm( sourceNode.child( "float_array" ).text().as_string() );
                        const int count = sourceNode.child( "float_array" ).attribute( "count" ).as_int();

                        Animation& animation = outMesh.animations[ id ];
                        std::vector< float >* array = isInput ? &animation.input : &animation.output;

                        array->resize( count );

                        for (int i = 0; i < count; ++i)
                        {
                            stm >> (*array)[ i ];
                        }

                        break;
                    }
                }
            }
            else if (std::string( "INTERPOLATION" ) == inputNode.attribute( "semantic" ).value())
            {
                for (auto sourceNode : animNode.children( "source" ))
                {
                    if (sourceId == sourceNode.attribute( "id" ).value())
                    {
                        std::stringstream stm( sourceNode.child( "Name_array" ).text().as_string() );
                        const int count = sourceNode.child( "Name_array" ).attribute( "count" ).as_int();
                        Animation& animation = outMesh.animations[ id ];

                        animation.interpolation.resize( count );

                        for (int i = 0; i < count; ++i)
                        {
                            std::string interp;
                            stm >> interp;

                            if (interp == "LINEAR")
                            {
                                animation.interpolation[ i ] = Animation::Linear;
                            }
                            else if (interp == "BEZIER")
                            {
                                animation.interpolation[ i ] = Animation::Bezier;
                            }
                            else
                            {
                                assert( !"Unhandled interpolation type." );
                            }
                        }
                    }
                }
            }
            // \todo merge with OUT_TANGENT
            else if (std::string( "IN_TANGENT" ) == inputNode.attribute( "semantic" ).value())
            {
                for (auto sourceNode : animNode.children( "source" ))
                {
                    if (sourceId != sourceNode.attribute( "id" ).value())
                    {
                        continue;
                    }

                    std::stringstream stm( sourceNode.child( "float_array" ).text().as_string() );
                    const int count = sourceNode.child( "float_array" ).attribute( "count" ).as_int();
                    Animation& animation = outMesh.animations[ id ];
                    animation.inTangents.resize( count );

                    for (int i = 0; i < count; ++i)
                    {
                        stm >> animation.inTangents[ i ];
                    }
                }
            }
            else if (std::string( "OUT_TANGENT" ) == inputNode.attribute( "semantic" ).value())
            {
                for (auto sourceNode : animNode.children( "source" ))
                {
                    if (sourceId != sourceNode.attribute( "id" ).value())
                    {
                        continue;
                    }

                    std::stringstream stm( sourceNode.child( "float_array" ).text().as_string() );
                    const int count = sourceNode.child( "float_array" ).attribute( "count" ).as_int();
                    Animation& animation = outMesh.animations[ id ];
                    animation.outTangents.resize( count );

                    for (int i = 0; i < count; ++i)
                    {
                        stm >> animation.outTangents[ i ];
                    }
                }
            }
        }

        std::string channelTarget = animNode.child( "channel" ).attribute( "target" ).value();
        // Parses target joint name.  \fixme Tested only on Blender.
        const std::string jointName = channelTarget.substr( 0, channelTarget.find( '/' ) );

        // Parses target location/rotation/transform.
        outMesh.animations[ id ].operation = channelTarget.substr( channelTarget.find( '/' ) + 1, channelTarget.length() );

        for (std::size_t j = 0; j < outMesh.joints.size(); ++j)
        {
            if (outMesh.joints[ j ].name == jointName)
            {
                outMesh.animations[ id ].joint = (int)j;
                break;
            }
        }
    }
}


/**
 Reads skeleton instance controller for a mesh.
 Assumes that the node pointed by <skeleton> is at most two
 levels down from <visual_scene>.

 \param parentNode <instance_controller>'s parent node.
 \param outMesh Mesh that will store the geometry.
 */
static void ReadSkeleton( const pugi::xml_node& parentNode, Mesh& outMesh )
{
    std::string skeletonId;
    auto skeleton = parentNode.first_element_by_path( "instance_controller/skeleton" );
    if (skeleton)
    {
        skeletonId = skeleton.text().as_string();
        skeletonId = skeletonId.substr( 1, skeletonId.length() );
    }

    auto sceneNode = parentNode.root().first_element_by_path( "COLLADA/library_visual_scenes/visual_scene" );

    Matrix44 transform;

    // Blender stores skeleton root one level down, so iterate children.
    for (auto node : sceneNode.children( "node" ))
    {
        for (auto innerNode : node.children( "node" ))
        {
            if (std::string( innerNode.attribute( "id" ).as_string() ) != skeletonId)
            {
                continue;
            }
            
            // Reads parent's transform.
            auto matrix = node.child( "matrix" );
            if (matrix)
            {
                std::stringstream matrixDataStm( matrix.text().as_string() );

                for (int i = 0; i < 16; ++i)
                {
                    matrixDataStm >> transform.m[ i ];
                }

                transform.Transpose( transform );
            }
            else
            {
                ReadNonMatrixTransform( node, transform );
            }

            ReadJointNode( innerNode, outMesh.skeletonRoot );
            // Applies skeleton root parent's transform to skeleton root.
            Matrix44::Multiply( outMesh.skeletonRoot.transform, transform, outMesh.skeletonRoot.transform );

            break;
        }
    }

    std::string skinUrl = parentNode.child( "instance_controller" ).attribute( "url" ).value();
    skinUrl = skinUrl.substr( 1, skinUrl.length() );
    // Finds the skin in <library_controllers>.
    auto libCtrlNode = parentNode.root().first_element_by_path( "COLLADA/library_controllers" );

    pugi::xml_node skin;

    // \todo flatten
    // \fixme VS 2012 warning: unreachable code?
    for (auto ctrlNode : libCtrlNode.children())
    {
        if (skinUrl != ctrlNode.attribute( "id" ).as_string())
        {
            break;
        }

        std::string geomUrl = ctrlNode.child( "skin" ).attribute( "source" ).as_string();
        geomUrl = geomUrl.substr( 1, geomUrl.length() );

        auto libGeomNode = parentNode.root().first_element_by_path( "COLLADA/library_geometries" );

        for (auto geomNode : libGeomNode.children())
        {
            if (geomUrl == geomNode.attribute( "id" ).as_string())
            {
                // Finds the mesh from <library_geometries>
                auto geometries = geomNode.root().first_element_by_path( "COLLADA/library_geometries" );
                
                for (auto geometry : geometries.children( "geometry" ))
                {
                    if (geometry.attribute( "id" ).value() == geomUrl)
                    {
                        ReadMesh( geometry.child( "mesh" ), outMesh );

                        for (std::size_t v = 0; v < outMesh.vertex.size(); ++v)
                        {
                            Matrix44::TransformPoint( outMesh.vertex[ v ], transform, &outMesh.vertex[ v ] );
                        }

                        break;
                    }
                }

                break;
            }
        }

        ReadSkin( ctrlNode.child( "skin" ), outMesh );
        ReadAnimations( ctrlNode, outMesh );

        for (std::size_t j = 0; j < outMesh.joints.size(); ++j)
        {
            Matrix44::Multiply( transform,
                             outMesh.joints[ j ].transform,
                             outMesh.joints[ j ].transform );
        }

        break;
    }
}


/**
 Reads instance geometry (Collada term for a non-skinned mesh).

 \param geomNode <instance_geometry>
 \param outMesh Mesh where geometry is read into
 */
static void ReadInstanceGeometry( const pugi::xml_node& geomNode, Mesh& outMesh )
{
    std::string url = geomNode.attribute( "url" ).value();
    url = url.substr( 1, url.length() );

    bool foundMesh = false;

    // Finds the mesh from <library_geometries>
    auto geometries = geomNode.root().first_element_by_path( "COLLADA/library_geometries" );

    for (auto geometry : geometries.children( "geometry" ))
    {
        if (geometry.attribute( "id" ).value() == url ||
            geometry.attribute( "name" ).value() == url)
        {
            foundMesh = true;
            outMesh.name = geometry.attribute("name").value();
            ReadMesh( geometry.child( "mesh" ), outMesh );
            break;
        }
    }

    assert( foundMesh );

    auto matrix = geomNode.parent().child( "matrix" );

    if (matrix)
    {
        Matrix44 transform;
        std::stringstream matrixDataStm( matrix.text().as_string() );

        for (int i = 0; i < 16; ++i)
        {
            matrixDataStm >> transform.m[ i ];
        }

        transform.Transpose( transform );

        for (std::size_t v = 0; v < outMesh.vertex.size(); ++v)
        {
            Vec3 pos = outMesh.vertex[ v ];

            // Undo z-up -> y-up, because transform matrix is z-up.
            if (gZup)
            {
                pos.x =  outMesh.vertex[ v ].x;
                pos.y = -outMesh.vertex[ v ].z;
                pos.z =  outMesh.vertex[ v ].y;
            }

            Matrix44::TransformPoint( pos, transform, &outMesh.vertex[ v ] );

            if (gZup)
            {
                pos.x =  outMesh.vertex[ v ].x;
                pos.y =  outMesh.vertex[ v ].z;
                pos.z = -outMesh.vertex[ v ].y;

                outMesh.vertex[ v ] = pos;
            }
        }
        
        transform.Transpose( transform );

        for (std::size_t n = 0; n < outMesh.vnormal.size(); ++n)
        {
            Matrix44::TransformDirection( outMesh.vnormal[ n ], transform, &outMesh.vnormal[ n ] );
        }
    }
    else
    {
        // TODO: Use ReadNonMatrixTransform(). Its result is not usable in this case. (why?) -2014-03-02

        auto translate = geomNode.parent().child( "translate" );
        std::stringstream translateStm( translate.text().as_string() );
        Vec3 translation;
        translateStm >> translation.x >> translation.y >> translation.z;

        Vec3 rotation;

        for (auto rotationElem : geomNode.parent().children( "rotate" ))
        {
            std::stringstream rotateStm( rotationElem.text().as_string() );

            if (std::string( rotationElem.attribute( "sid" ).value() ) == "rotationX" ||
                std::string( rotationElem.attribute( "sid" ).value() ) == "rotateX" ||
                std::string( rotationElem.attribute( "sid" ).value() ) == "RotX")
            {
                // Ignores axis because we know it's x-axis and only gets the angle.
                rotateStm >> rotation.x >> rotation.x >> rotation.x >> rotation.x;
            }
            else if (std::string( rotationElem.attribute( "sid" ).value() ) == "rotationY" ||
                     std::string( rotationElem.attribute( "sid" ).value() ) == "rotateY" ||
                     std::string( rotationElem.attribute( "sid" ).value() ) == "RotY")
            {
                rotateStm >> rotation.y >> rotation.y >> rotation.y >> rotation.y;
            }
            else if (std::string( rotationElem.attribute( "sid" ).value() ) == "rotationZ" ||
                     std::string( rotationElem.attribute( "sid" ).value() ) == "rotateZ" ||
                     std::string( rotationElem.attribute( "sid" ).value() ) == "RotZ")
            {
                rotateStm >> rotation.z >> rotation.z >> rotation.z >> rotation.z;
            }
        }

        auto scaleElem = geomNode.parent().child( "scale" );
        std::stringstream scaleStm( scaleElem.text().as_string() );
        Vec3 scale;
        scaleStm >> scale.x >> scale.y >> scale.z;

        Matrix44 transform;
        transform.MakeRotationXYZ( rotation.x, -rotation.z, rotation.y );
        transform.Translate( Vec3( translation.x, translation.z, -translation.y ) );

        for (std::size_t v = 0; v < outMesh.vertex.size(); ++v)
        {
            Vec3 pos = outMesh.vertex[ v ] * Vec3( scale.x, scale.z, scale.y );
            Matrix44::TransformPoint( pos, transform, &outMesh.vertex[ v ] );
        }

        for (std::size_t v = 0; v < outMesh.vnormal.size(); ++v)
        {
            Matrix44::TransformDirection( outMesh.vnormal[ v ], transform, &outMesh.vnormal[ v ] );
        }
    }
}


/**
   \param parentNode <morph>?
   \param outMesh Mesh.
 */
static void ReadMorph( const pugi::xml_node& parentNode, Mesh& outMesh )
{
    std::string url = parentNode.child( "instance_controller" ).attribute( "url" ).value();
    url = url.substr( 1, url.length() );

    auto controllers = parentNode.root().first_element_by_path( "COLLADA/library_controllers" );

    for (auto controller : controllers.children( "controller" ))
    {
        if (url != controller.attribute( "id" ).value())
        {
            continue;
        }

        auto morphNode = controller.child( "morph" );

        // Morph source seems to be the base mesh for morphing.
        std::string morphSource = morphNode.attribute( "source" ).value();
        morphSource = morphSource.substr( 1, morphSource.length() );

        // Finds and reads base mesh geometry.
        outMesh.name = morphSource;

        auto sceneNode = parentNode.root().first_element_by_path( "COLLADA/library_visual_scenes/visual_scene" );

        for (auto node : sceneNode.children( "node" ) )
        {
            auto ig = node.child( "instance_geometry" );

            if (!ig)
            {
                continue;
            }

            std::string geomURL = ig.attribute( "url" ).value();
            geomURL = geomURL.substr( 1, geomURL.length() );

            if (geomURL == morphSource)
            {
                ReadInstanceGeometry( ig, outMesh );
                break;
            }
        }

        std::string targetSource;
        std::string weightSource;

        for (auto input : morphNode.child( "targets" ).children( "input" ))
        {
            if (std::string( "MORPH_TARGET" ) == input.attribute( "semantic" ).value())
            {
                targetSource = input.attribute( "source" ).value();
                targetSource = targetSource.substr( 1, targetSource.length() );
            }
            else if (std::string( "MORPH_WEIGHT" ) == input.attribute( "semantic" ).value())
            {
                weightSource = input.attribute( "source" ).value();
                weightSource = weightSource.substr( 1, weightSource.length() );
            }
        }

        for (auto source : morphNode.children( "source" ))
        {
            if (targetSource == source.attribute( "id" ).value())
            {
                std::string targets = source.child( "IDREF_array" ).text().as_string();
                int targetCount = source.child( "IDREF_array" ).attribute( "count" ).as_int();
                outMesh.morphTargets.resize( targetCount );

                std::stringstream stm( targets );

                for (int i = 0; i < targetCount; ++i)
                {
                    stm >> outMesh.morphTargets[ i ];
                }
            }
            else if (weightSource == source.attribute( "id" ).value())
            {
                std::string weights = source.child( "float_array" ).text().as_string();
                int weightCount = source.child( "float_array" ).attribute( "count" ).as_int();

                outMesh.morphWeights.resize( weightCount );

                std::stringstream stm( weights );

                for (int i = 0; i < weightCount; ++i)
                {
                    stm >> outMesh.morphWeights[ i ];
                }
            }
        }
    }
}


static void StoreMorphTargetNodeNames( const pugi::xml_node& sceneNode, std::vector< std::string >& outTargetNodeNames )
{
    std::string url;

    for (auto node : sceneNode.children( "node" ))
    {
        if (node.child( "instance_controller" ))
        {
            url = node.child( "instance_controller" ).attribute( "url" ).value();
            url = url.substr( 1, url.length() );
            break;
        }
    }

    auto libCtrls = sceneNode.root().first_element_by_path( "COLLADA/library_controllers" );

    for (const auto& ctrlNode : libCtrls)
    {
        if (!ctrlNode.child( "morph" ))
        {
            continue;
        }

        auto morphNode = ctrlNode.child( "morph" );

        for (const auto& morphChild : morphNode.children( "source" ))
        {
            auto idRefNode = morphChild.child( "IDREF_array" );

            if (!idRefNode)
            {
                continue;
            }

            std::string targetNodes = idRefNode.text().as_string();
            std::stringstream stm( targetNodes );
            outTargetNodeNames.resize( idRefNode.attribute( "count" ).as_int() );

            for (std::size_t n = 0; n < outTargetNodeNames.size(); ++n)
            {
                stm >> outTargetNodeNames[ n ];
            }

            break;
        }

        break;
    }
}


bool IsMorphTarget(const std::string& nodeName, const std::vector< std::string >& morphTargetNodeNames )
{
    for (const auto& name : morphTargetNodeNames)
    {
        if (nodeName == name)
        {
            return true;
        }
    }

    return false;
}


static std::string GetMorphBaseMeshName( const pugi::xml_node& sceneNode )
{
    for (auto node : sceneNode.children( "node" ))
    {
        auto ctrl = node.child( "instance_controller" );

        if (ctrl && !ctrl.child( "skeleton" ))
        {
            std::string url = ctrl.attribute( "url" ).value();
            url = url.substr( 1, url.length() );

            auto controllers = ctrl.root().first_element_by_path( "COLLADA/library_controllers" );

            for (auto controller : controllers.children( "controller" ))
            {
                if (url != controller.attribute( "id" ).value())
                {
                    continue;
                }

                auto morphNode = controller.child( "morph" );

                // Morph source seems to be the base mesh for morphing.
                std::string morphSource = morphNode.attribute( "source" ).value();
                morphSource = morphSource.substr( 1, morphSource.length() );
                return morphSource;
            }
        }
    }

    return "no_base_mesh";
}

static void LoadDAE( const char* path, std::vector< Mesh >& outMeshes )
{
    pugi::xml_document doc;
    if (!doc.load_file( path ))
    {
        std::cerr << "Could not parse file." << std::endl;
        exit( 1 );
    }

    const std::string upAxis = doc.first_element_by_path( "COLLADA/asset/up_axis" ).text().as_string();

    if (upAxis == "Z_UP")
    {
        gZup = true;
    }

    auto sceneNode = doc.first_element_by_path( "COLLADA/library_visual_scenes/visual_scene" );

    if (!sceneNode)
    {
        std::cerr << "Not a valid .dae file!" << std::endl;
        exit( 1 );
    }

    // Finds morph target node, if any, and stores names of its targets so that
    // they are not processed as normal meshes.
    std::vector< std::string > morphTargetNodeNames;
    StoreMorphTargetNodeNames( sceneNode, morphTargetNodeNames );

    // Finds morph base mesh name so it will not be processed like other meshes.
    std::string baseMeshName = GetMorphBaseMeshName( sceneNode );

    // Processes meshes that are not morph targets.
    for (auto node : sceneNode.children( "node" ))
    {
        if (node.child( "instance_controller" ))
        {
            outMeshes.push_back( Mesh() );

            if (node.child( "instance_controller" ).child( "skeleton" ))
            {
                outMeshes.back().name = node.attribute( "name" ).value();
                ReadSkeleton( node, outMeshes.back() );
            }
            else // Assumes that the controller controls morph targets.
            {
                ReadMorph( node, outMeshes.back() );
                ReadAnimations( node.child( "instance_controller" ), outMeshes.back() );
            }
        }

        for (auto geomNode : node.children( "instance_geometry" ))
        {
            std::string url = geomNode.attribute( "url" ).value();
            url = url.substr( 1, url.length() );

            if (!IsMorphTarget( url, morphTargetNodeNames ) && baseMeshName != url)
            {
                outMeshes.push_back( Mesh() );
                ReadInstanceGeometry( geomNode, outMeshes.back() );
            }
            else if (baseMeshName != url)
            {
                outMeshes.push_back( Mesh() );
                outMeshes.back().morphTarget = true;
                ReadInstanceGeometry( geomNode, outMeshes.back() );
            }
        }

        // Supports one level of hierarchy. Some DCCs put a root node above others.
        // \todo Support recursion.
        for (auto innerNode : node.children( "node" ))
        {
            if (innerNode.child( "instance_controller" ))
            {
                outMeshes.push_back( Mesh() );
                outMeshes.back().name = innerNode.attribute( "name" ).value();
                ReadSkeleton( innerNode, outMeshes.back() );
            }

            if (innerNode.child( "instance_geometry" ))
            {
                outMeshes.push_back( Mesh() );
                outMeshes.back().name = innerNode.attribute( "name" ).value();
                ReadInstanceGeometry( innerNode.child( "instance_geometry" ), outMeshes.back() );
            }
        }
    }
}


int main( int paramCount, char** paramStrings )
{
    if (paramCount != 2)
    {
        std::cerr << "Usage: ./convert_dae file.dae" << std::endl;
        return 1;
    }
    
    {
        std::ifstream ifs( paramStrings[ 1 ], std::ios_base::binary );
    
        if (!ifs)
        {
            std::cerr << "Error: Couldn't open file " << paramStrings[ 1 ] << std::endl;
            return 1;
        }
    }

    std::cerr << "Converting... " << std::endl;
    
    LoadDAE( paramStrings[ 1 ], gMeshes );

    // Creates a new file name by removing 'dae' and appending 'ae3d'.
    std::string outFile = std::string( paramStrings[ 1 ] );
    outFile = outFile.substr( 0, outFile.length() - 3 );
    outFile.append( "ae3d" );
    
    WriteAe3d( outFile );
}
